@@ -44,6 +44,7 @@
 
 #include <openssl/ssl.h>
 #include <openssl/err.h>
+#include <openssl/conf.h>
 
 enum {
   IO_NONE,
@@ -148,11 +149,11 @@ static ssize_t send_callback(nghttp2_session *session,
                              void *user_data)
 {
   struct Connection *connection;
-  ssize_t rv;
+  int rv;
   connection = (struct Connection*)user_data;
   connection->want_io = IO_NONE;
   ERR_clear_error();
-  rv = SSL_write(connection->ssl, data, length);
+  rv = SSL_write(connection->ssl, data, (int)length);
   if(rv < 0) {
     int err = SSL_get_error(connection->ssl, rv);
     if(err == SSL_ERROR_WANT_WRITE || err == SSL_ERROR_WANT_READ) {
@@ -177,11 +178,11 @@ static ssize_t recv_callback(nghttp2_session *session,
                              void *user_data)
 {
   struct Connection *connection;
-  ssize_t rv;
+  int rv;
   connection = (struct Connection*)user_data;
   connection->want_io = IO_NONE;
   ERR_clear_error();
-  rv = SSL_read(connection->ssl, buf, length);
+  rv = SSL_read(connection->ssl, buf, (int)length);
   if(rv < 0) {
     int err = SSL_get_error(connection->ssl, rv);
     if(err == SSL_ERROR_WANT_WRITE || err == SSL_ERROR_WANT_READ) {
@@ -270,10 +271,10 @@ static int on_stream_close_callback(nghttp2_session *session,
   req = nghttp2_session_get_stream_user_data(session, stream_id);
   if(req) {
     int rv;
-    rv = nghttp2_submit_goaway(session, NGHTTP2_FLAG_NONE, NGHTTP2_NO_ERROR,
-                               NULL, 0);
+    rv = nghttp2_session_terminate_session(session, NGHTTP2_NO_ERROR);
+
     if(rv != 0) {
-      diec("nghttp2_submit_goaway", rv);
+      diec("nghttp2_session_terminate_session", rv);
     }
   }
   return 0;
@@ -309,13 +310,21 @@ static int on_data_chunk_recv_callback(nghttp2_session *session, uint8_t flags,
  */
 static void setup_nghttp2_callbacks(nghttp2_session_callbacks *callbacks)
 {
-  memset(callbacks, 0, sizeof(nghttp2_session_callbacks));
-  callbacks->send_callback = send_callback;
-  callbacks->recv_callback = recv_callback;
-  callbacks->on_frame_send_callback = on_frame_send_callback;
-  callbacks->on_frame_recv_callback = on_frame_recv_callback;
-  callbacks->on_stream_close_callback = on_stream_close_callback;
-  callbacks->on_data_chunk_recv_callback = on_data_chunk_recv_callback;
+  nghttp2_session_callbacks_set_send_callback(callbacks, send_callback);
+
+  nghttp2_session_callbacks_set_recv_callback(callbacks, recv_callback);
+
+  nghttp2_session_callbacks_set_on_frame_send_callback
+    (callbacks, on_frame_send_callback);
+
+  nghttp2_session_callbacks_set_on_frame_recv_callback
+    (callbacks, on_frame_recv_callback);
+
+  nghttp2_session_callbacks_set_on_stream_close_callback
+    (callbacks, on_stream_close_callback);
+
+  nghttp2_session_callbacks_set_on_data_chunk_recv_callback
+    (callbacks, on_data_chunk_recv_callback);
 }
 
 /*
@@ -506,7 +515,7 @@ static void request_free(struct Request *req)
  */
 static void fetch_uri(const struct URI *uri)
 {
-  nghttp2_session_callbacks callbacks;
+  nghttp2_session_callbacks *callbacks;
   int fd;
   SSL_CTX *ssl_ctx;
   SSL *ssl;
@@ -518,8 +527,6 @@ static void fetch_uri(const struct URI *uri)
 
   request_init(&req, uri);
 
-  setup_nghttp2_callbacks(&callbacks);
-
   /* Establish connection and setup SSL */
   fd = connect_to(req.host, req.port);
   if(fd == -1) {
@@ -550,8 +557,20 @@ static void fetch_uri(const struct URI *uri)
   set_tcp_nodelay(fd);
 
   printf("[INFO] SSL/TLS handshake completed\n");
-  rv = nghttp2_session_client_new(&connection.session, &callbacks,
+
+  rv = nghttp2_session_callbacks_new(&callbacks);
+
+  if(rv != 0) {
+    diec("nghttp2_session_callbacks_new", rv);
+  }
+
+  setup_nghttp2_callbacks(callbacks);
+
+  rv = nghttp2_session_client_new(&connection.session, callbacks,
                                   &connection);
+
+  nghttp2_session_callbacks_del(callbacks);
+
   if(rv != 0) {
     diec("nghttp2_session_client_new", rv);
   }
@@ -685,6 +704,8 @@ int main(int argc, char **argv)
   act.sa_handler = SIG_IGN;
   sigaction(SIGPIPE, &act, 0);
 
+  OPENSSL_config(NULL);
+  OpenSSL_add_all_algorithms();
   SSL_load_error_strings();
   SSL_library_init();
 
