@@ -32,6 +32,7 @@
 
 #include <openssl/ssl.h>
 #include <openssl/err.h>
+#include <openssl/conf.h>
 
 #include <event.h>
 #include <event2/event.h>
@@ -330,17 +331,30 @@ static SSL* create_ssl(SSL_CTX *ssl_ctx)
 
 static void initialize_nghttp2_session(http2_session_data *session_data)
 {
-  nghttp2_session_callbacks callbacks;
+  nghttp2_session_callbacks *callbacks;
 
-  memset(&callbacks, 0, sizeof(callbacks));
+  nghttp2_session_callbacks_new(&callbacks);
 
-  callbacks.send_callback = send_callback;
-  callbacks.on_frame_recv_callback = on_frame_recv_callback;
-  callbacks.on_data_chunk_recv_callback = on_data_chunk_recv_callback;
-  callbacks.on_stream_close_callback = on_stream_close_callback;
-  callbacks.on_header_callback = on_header_callback;
-  callbacks.on_begin_headers_callback = on_begin_headers_callback;
-  nghttp2_session_client_new(&session_data->session, &callbacks, session_data);
+  nghttp2_session_callbacks_set_send_callback(callbacks, send_callback);
+
+  nghttp2_session_callbacks_set_on_frame_recv_callback
+    (callbacks, on_frame_recv_callback);
+
+  nghttp2_session_callbacks_set_on_data_chunk_recv_callback
+    (callbacks, on_data_chunk_recv_callback);
+
+  nghttp2_session_callbacks_set_on_stream_close_callback
+    (callbacks, on_stream_close_callback);
+
+  nghttp2_session_callbacks_set_on_header_callback
+    (callbacks, on_header_callback);
+
+  nghttp2_session_callbacks_set_on_begin_headers_callback
+    (callbacks, on_begin_headers_callback);
+
+  nghttp2_session_client_new(&session_data->session, callbacks, session_data);
+
+  nghttp2_session_callbacks_del(callbacks);
 }
 
 static void send_client_connection_header(http2_session_data *session_data)
@@ -414,17 +428,22 @@ static int session_send(http2_session_data *session_data)
 static void readcb(struct bufferevent *bev, void *ptr)
 {
   http2_session_data *session_data = (http2_session_data*)ptr;
-  int rv;
+  ssize_t readlen;
   struct evbuffer *input = bufferevent_get_input(bev);
   size_t datalen = evbuffer_get_length(input);
   unsigned char *data = evbuffer_pullup(input, -1);
-  rv = nghttp2_session_mem_recv(session_data->session, data, datalen);
-  if(rv < 0) {
-    warnx("Fatal error: %s", nghttp2_strerror(rv));
+
+  readlen = nghttp2_session_mem_recv(session_data->session, data, datalen);
+  if(readlen < 0) {
+    warnx("Fatal error: %s", nghttp2_strerror((int)readlen));
+    delete_http2_session_data(session_data);
+    return;
+  }
+  if(evbuffer_drain(input, readlen) != 0) {
+    warnx("Fatal error: evbuffer_drain failed");
     delete_http2_session_data(session_data);
     return;
   }
-  evbuffer_drain(input, rv);
   if(session_send(session_data) != 0) {
     delete_http2_session_data(session_data);
     return;
@@ -555,6 +574,8 @@ int main(int argc, char **argv)
   act.sa_handler = SIG_IGN;
   sigaction(SIGPIPE, &act, NULL);
 
+  OPENSSL_config(NULL);
+  OpenSSL_add_all_algorithms();
   SSL_load_error_strings();
   SSL_library_init();
 
