@@ -36,6 +36,7 @@
 
 #include <openssl/ssl.h>
 #include <openssl/err.h>
+#include <openssl/conf.h>
 
 #include <event.h>
 #include <event2/event.h>
@@ -68,7 +69,6 @@ typedef struct http2_session_data {
   app_context *app_ctx;
   nghttp2_session *session;
   char *client_addr;
-  size_t handshake_leftlen;
 } http2_session_data;
 
 struct app_context {
@@ -83,7 +83,7 @@ static int next_proto_cb(SSL *s, const unsigned char **data, unsigned int *len,
                          void *arg)
 {
   *data = next_proto_list;
-  *len = next_proto_list_len;
+  *len = (unsigned int)next_proto_list_len;
   return SSL_TLSEXT_ERR_OK;
 }
 
@@ -191,7 +191,6 @@ static http2_session_data* create_http2_session_data(app_context *app_ctx,
     (app_ctx->evbase, fd, ssl,
      BUFFEREVENT_SSL_ACCEPTING,
      BEV_OPT_CLOSE_ON_FREE | BEV_OPT_DEFER_CALLBACKS);
-  session_data->handshake_leftlen = NGHTTP2_CLIENT_CONNECTION_PREFACE_LEN;
   rv = getnameinfo(addr, addrlen, host, sizeof(host), NULL, 0, NI_NUMERICHOST);
   if(rv != 0) {
     session_data->client_addr = strdup("(unknown)");
@@ -240,16 +239,20 @@ static int session_send(http2_session_data *session_data)
    function. */
 static int session_recv(http2_session_data *session_data)
 {
-  int rv;
+  ssize_t readlen;
   struct evbuffer *input = bufferevent_get_input(session_data->bev);
   size_t datalen = evbuffer_get_length(input);
   unsigned char *data = evbuffer_pullup(input, -1);
-  rv = nghttp2_session_mem_recv(session_data->session, data, datalen);
-  if(rv < 0) {
-    warnx("Fatal error: %s", nghttp2_strerror(rv));
+
+  readlen = nghttp2_session_mem_recv(session_data->session, data, datalen);
+  if(readlen < 0) {
+    warnx("Fatal error: %s", nghttp2_strerror((int)readlen));
+    return -1;
+  }
+  if(evbuffer_drain(input, readlen) != 0) {
+    warnx("Fatal error: evbuffer_drain failed");
     return -1;
   }
-  evbuffer_drain(input, rv);
   if(session_send(session_data) != 0) {
     return -1;
   }
@@ -366,6 +369,7 @@ static int error_reply(nghttp2_session *session,
                        http2_stream_data *stream_data)
 {
   int rv;
+  ssize_t writelen;
   int pipefd[2];
   nghttp2_nv hdrs[] = {
     MAKE_NV(":status", "404")
@@ -384,10 +388,10 @@ static int error_reply(nghttp2_session *session,
     return 0;
   }
 
-  rv = write(pipefd[1], ERROR_HTML, sizeof(ERROR_HTML) - 1);
+  writelen = write(pipefd[1], ERROR_HTML, sizeof(ERROR_HTML) - 1);
   close(pipefd[1]);
 
-  if(rv != sizeof(ERROR_HTML)) {
+  if(writelen != sizeof(ERROR_HTML) - 1) {
     close(pipefd[0]);
     return -1;
   }
@@ -549,16 +553,38 @@ static int on_stream_close_callback(nghttp2_session *session,
 
 static void initialize_nghttp2_session(http2_session_data *session_data)
 {
-  nghttp2_session_callbacks callbacks;
+  nghttp2_option *option;
+  nghttp2_session_callbacks *callbacks;
 
-  memset(&callbacks, 0, sizeof(callbacks));
+  nghttp2_option_new(&option);
 
-  callbacks.send_callback = send_callback;
-  callbacks.on_frame_recv_callback = on_frame_recv_callback;
-  callbacks.on_stream_close_callback = on_stream_close_callback;
-  callbacks.on_header_callback = on_header_callback;
-  callbacks.on_begin_headers_callback = on_begin_headers_callback;
-  nghttp2_session_server_new(&session_data->session, &callbacks, session_data);
+  /* Tells nghttp2_session object that it handles client connection
+     preface */
+  nghttp2_option_set_recv_client_preface(option, 1);
+
+  nghttp2_session_callbacks_new(&callbacks);
+
+  nghttp2_session_callbacks_set_send_callback(callbacks, send_callback);
+
+  nghttp2_session_callbacks_set_on_frame_recv_callback
+    (callbacks, on_frame_recv_callback);
+
+  nghttp2_session_callbacks_set_on_stream_close_callback
+    (callbacks, on_stream_close_callback);
+
+  nghttp2_session_callbacks_set_on_header_callback
+    (callbacks, on_header_callback);
+
+  nghttp2_session_callbacks_set_on_begin_headers_callback
+    (callbacks, on_begin_headers_callback);
+
+
+
+  nghttp2_session_server_new2(&session_data->session, callbacks, session_data,
+                              option);
+
+  nghttp2_session_callbacks_del(callbacks);
+  nghttp2_option_del(option);
 }
 
 /* Send HTTP/2 client connection header, which includes 24 bytes
@@ -621,6 +647,14 @@ static void eventcb(struct bufferevent *bev, short events, void *ptr)
   http2_session_data *session_data = (http2_session_data*)ptr;
   if(events & BEV_EVENT_CONNECTED) {
     fprintf(stderr, "%s connected\n", session_data->client_addr);
+
+    initialize_nghttp2_session(session_data);
+
+    if(send_server_connection_header(session_data) != 0) {
+      delete_http2_session_data(session_data);
+      return;
+    }
+
     return;
   }
   if(events & BEV_EVENT_EOF) {
@@ -633,38 +667,6 @@ static void eventcb(struct bufferevent *bev, short events, void *ptr)
   delete_http2_session_data(session_data);
 }
 
-/* readcb for bufferevent to check first 24 bytes client connection
-   header. */
-static void handshake_readcb(struct bufferevent *bev, void *ptr)
-{
-  http2_session_data *session_data = (http2_session_data*)ptr;
-  uint8_t data[24];
-  struct evbuffer *input = bufferevent_get_input(session_data->bev);
-  int readlen = evbuffer_remove(input, data, session_data->handshake_leftlen);
-  const char *conhead = NGHTTP2_CLIENT_CONNECTION_PREFACE;
-
-  if(memcmp(conhead + NGHTTP2_CLIENT_CONNECTION_PREFACE_LEN
-            - session_data->handshake_leftlen, data, readlen) != 0) {
-    delete_http2_session_data(session_data);
-    return;
-  }
-  session_data->handshake_leftlen -= readlen;
-  if(session_data->handshake_leftlen == 0) {
-    bufferevent_setcb(session_data->bev, readcb, writecb, eventcb, ptr);
-    /* Process pending data in buffer since they are not notified
-       further */
-    initialize_nghttp2_session(session_data);
-    if(send_server_connection_header(session_data) != 0) {
-      delete_http2_session_data(session_data);
-      return;
-    }
-    if(session_recv(session_data) != 0) {
-      delete_http2_session_data(session_data);
-      return;
-    }
-  }
-}
-
 /* callback for evconnlistener */
 static void acceptcb(struct evconnlistener *listener, int fd,
                      struct sockaddr *addr, int addrlen, void *arg)
@@ -673,8 +675,8 @@ static void acceptcb(struct evconnlistener *listener, int fd,
   http2_session_data *session_data;
 
   session_data = create_http2_session_data(app_ctx, fd, addr, addrlen);
-  bufferevent_setcb(session_data->bev, handshake_readcb, NULL, eventcb,
-                    session_data);
+
+  bufferevent_setcb(session_data->bev, readcb, writecb, eventcb, session_data);
 }
 
 static void start_listen(struct event_base *evbase, const char *service,
@@ -703,6 +705,8 @@ static void start_listen(struct event_base *evbase, const char *service,
                                        LEV_OPT_REUSEABLE, 16,
                                        rp->ai_addr, rp->ai_addrlen);
     if(listener) {
+      freeaddrinfo(res);
+
       return;
     }
   }
@@ -748,6 +752,8 @@ int main(int argc, char **argv)
   act.sa_handler = SIG_IGN;
   sigaction(SIGPIPE, &act, NULL);
 
+  OPENSSL_config(NULL);
+  OpenSSL_add_all_algorithms();
   SSL_load_error_strings();
   SSL_library_init();
 
