@@ -33,35 +33,33 @@
 #include "nghttp2_net.h"
 #include "nghttp2_priority_spec.h"
 
-int nghttp2_frame_is_data_frame(uint8_t *head)
-{
-  return head[2] == 0;
-}
-
 void nghttp2_frame_pack_frame_hd(uint8_t* buf, const nghttp2_frame_hd *hd)
 {
-  nghttp2_put_uint16be(&buf[0], hd->length);
-  buf[2]=  hd->type;
-  buf[3] = hd->flags;
-  nghttp2_put_uint32be(&buf[4], hd->stream_id);
+  nghttp2_put_uint32be(&buf[0], (uint32_t)(hd->length << 8));
+  buf[3]=  hd->type;
+  buf[4] = hd->flags;
+  nghttp2_put_uint32be(&buf[5], hd->stream_id);
+  /* ignore hd->reserved for now */
 }
 
 void nghttp2_frame_unpack_frame_hd(nghttp2_frame_hd *hd, const uint8_t* buf)
 {
-  hd->length = nghttp2_get_uint16(&buf[0]) & NGHTTP2_FRAME_LENGTH_MASK;
-  hd->type = buf[2];
-  hd->flags = buf[3];
-  hd->stream_id = nghttp2_get_uint32(&buf[4]) & NGHTTP2_STREAM_ID_MASK;
+  hd->length = nghttp2_get_uint32(&buf[0]) >> 8;
+  hd->type = buf[3];
+  hd->flags = buf[4];
+  hd->stream_id = nghttp2_get_uint32(&buf[5]) & NGHTTP2_STREAM_ID_MASK;
+  hd->reserved = 0;
 }
 
-static void frame_set_hd(nghttp2_frame_hd *hd, uint16_t length,
-                         uint8_t type, uint8_t flags,
-                         int32_t stream_id)
+void nghttp2_frame_hd_init(nghttp2_frame_hd *hd, size_t length,
+                           uint8_t type, uint8_t flags,
+                           int32_t stream_id)
 {
   hd->length = length;
   hd->type = type;
   hd->flags = flags;
   hd->stream_id = stream_id;
+  hd->reserved = 0;
 }
 
 void nghttp2_frame_headers_init(nghttp2_headers *frame,
@@ -70,7 +68,7 @@ void nghttp2_frame_headers_init(nghttp2_headers *frame,
                                 const nghttp2_priority_spec *pri_spec,
                                 nghttp2_nv *nva, size_t nvlen)
 {
-  frame_set_hd(&frame->hd, 0, NGHTTP2_HEADERS, flags, stream_id);
+  nghttp2_frame_hd_init(&frame->hd, 0, NGHTTP2_HEADERS, flags, stream_id);
   frame->padlen = 0;
   frame->nva = nva;
   frame->nvlen = nvlen;
@@ -91,8 +89,8 @@ void nghttp2_frame_headers_free(nghttp2_headers *frame)
 void nghttp2_frame_priority_init(nghttp2_priority *frame, int32_t stream_id,
                                  const nghttp2_priority_spec *pri_spec)
 {
-  frame_set_hd(&frame->hd, NGHTTP2_PRIORITY_SPECLEN, NGHTTP2_PRIORITY,
-               NGHTTP2_FLAG_NONE, stream_id);
+  nghttp2_frame_hd_init(&frame->hd, NGHTTP2_PRIORITY_SPECLEN, NGHTTP2_PRIORITY,
+                        NGHTTP2_FLAG_NONE, stream_id);
   frame->pri_spec = *pri_spec;
 }
 
@@ -101,10 +99,10 @@ void nghttp2_frame_priority_free(nghttp2_priority *frame)
 
 void nghttp2_frame_rst_stream_init(nghttp2_rst_stream *frame,
                                    int32_t stream_id,
-                                   nghttp2_error_code error_code)
+                                   uint32_t error_code)
 {
-  frame_set_hd(&frame->hd, 4, NGHTTP2_RST_STREAM, NGHTTP2_FLAG_NONE,
-               stream_id);
+  nghttp2_frame_hd_init(&frame->hd, 4, NGHTTP2_RST_STREAM, NGHTTP2_FLAG_NONE,
+                        stream_id);
   frame->error_code = error_code;
 }
 
@@ -115,8 +113,8 @@ void nghttp2_frame_rst_stream_free(nghttp2_rst_stream *frame)
 void nghttp2_frame_settings_init(nghttp2_settings *frame, uint8_t flags,
                                  nghttp2_settings_entry *iv, size_t niv)
 {
-  frame_set_hd(&frame->hd, niv * NGHTTP2_FRAME_SETTINGS_ENTRY_LENGTH,
-               NGHTTP2_SETTINGS, flags, 0);
+  nghttp2_frame_hd_init(&frame->hd, niv * NGHTTP2_FRAME_SETTINGS_ENTRY_LENGTH,
+                        NGHTTP2_SETTINGS, flags, 0);
   frame->niv = niv;
   frame->iv = iv;
 }
@@ -131,11 +129,12 @@ void nghttp2_frame_push_promise_init(nghttp2_push_promise *frame,
                                      int32_t promised_stream_id,
                                      nghttp2_nv *nva, size_t nvlen)
 {
-  frame_set_hd(&frame->hd, 0, NGHTTP2_PUSH_PROMISE, flags, stream_id);
+  nghttp2_frame_hd_init(&frame->hd, 0, NGHTTP2_PUSH_PROMISE, flags, stream_id);
   frame->padlen = 0;
   frame->nva = nva;
   frame->nvlen = nvlen;
   frame->promised_stream_id = promised_stream_id;
+  frame->reserved = 0;
 }
 
 void nghttp2_frame_push_promise_free(nghttp2_push_promise *frame)
@@ -146,7 +145,7 @@ void nghttp2_frame_push_promise_free(nghttp2_push_promise *frame)
 void nghttp2_frame_ping_init(nghttp2_ping *frame, uint8_t flags,
                              const uint8_t *opaque_data)
 {
-  frame_set_hd(&frame->hd, 8, NGHTTP2_PING, flags, 0);
+  nghttp2_frame_hd_init(&frame->hd, 8, NGHTTP2_PING, flags, 0);
   if(opaque_data) {
     memcpy(frame->opaque_data, opaque_data, sizeof(frame->opaque_data));
   } else {
@@ -158,15 +157,16 @@ void nghttp2_frame_ping_free(nghttp2_ping *frame)
 {}
 
 void nghttp2_frame_goaway_init(nghttp2_goaway *frame, int32_t last_stream_id,
-                               nghttp2_error_code error_code,
+                               uint32_t error_code,
                                uint8_t *opaque_data, size_t opaque_data_len)
 {
-  frame_set_hd(&frame->hd, 8+opaque_data_len, NGHTTP2_GOAWAY,
-               NGHTTP2_FLAG_NONE, 0);
+  nghttp2_frame_hd_init(&frame->hd, 8+opaque_data_len, NGHTTP2_GOAWAY,
+                        NGHTTP2_FLAG_NONE, 0);
   frame->last_stream_id = last_stream_id;
   frame->error_code = error_code;
   frame->opaque_data = opaque_data;
   frame->opaque_data_len = opaque_data_len;
+  frame->reserved = 0;
 }
 
 void nghttp2_frame_goaway_free(nghttp2_goaway *frame)
@@ -179,14 +179,16 @@ void nghttp2_frame_window_update_init(nghttp2_window_update *frame,
                                       int32_t stream_id,
                                       int32_t window_size_increment)
 {
-  frame_set_hd(&frame->hd, 4, NGHTTP2_WINDOW_UPDATE, flags, stream_id);
+  nghttp2_frame_hd_init(&frame->hd, 4, NGHTTP2_WINDOW_UPDATE, flags,
+                        stream_id);
   frame->window_size_increment = window_size_increment;
+  frame->reserved = 0;
 }
 
 void nghttp2_frame_window_update_free(nghttp2_window_update *frame)
 {}
 
-void nghttp2_frame_altsvc_init(nghttp2_altsvc *frame, int32_t stream_id,
+void nghttp2_frame_altsvc_init(nghttp2_extension *frame, int32_t stream_id,
                                uint32_t max_age,
                                uint16_t port,
                                uint8_t *protocol_id,
@@ -195,68 +197,52 @@ void nghttp2_frame_altsvc_init(nghttp2_altsvc *frame, int32_t stream_id,
                                uint8_t *origin, size_t origin_len)
 {
   size_t payloadlen;
+  nghttp2_ext_altsvc *altsvc;
+
+  altsvc = frame->payload;
 
   payloadlen = NGHTTP2_ALTSVC_MINLEN + protocol_id_len + host_len + origin_len;
 
-  frame_set_hd(&frame->hd, payloadlen, NGHTTP2_ALTSVC, NGHTTP2_FLAG_NONE,
-               stream_id);
+  nghttp2_frame_hd_init(&frame->hd, payloadlen, NGHTTP2_EXT_ALTSVC,
+                        NGHTTP2_FLAG_NONE, stream_id);
 
-  frame->max_age = max_age;
-  frame->port = port;
-  frame->protocol_id = protocol_id;
-  frame->protocol_id_len = protocol_id_len;
-  frame->host = host;
-  frame->host_len = host_len;
-  frame->origin = origin;
-  frame->origin_len = origin_len;
+  altsvc->max_age = max_age;
+  altsvc->port = port;
+  altsvc->protocol_id = protocol_id;
+  altsvc->protocol_id_len = protocol_id_len;
+  altsvc->host = host;
+  altsvc->host_len = host_len;
+  altsvc->origin = origin;
+  altsvc->origin_len = origin_len;
 }
 
-void nghttp2_frame_altsvc_free(nghttp2_altsvc *frame)
+void nghttp2_frame_altsvc_free(nghttp2_extension *frame)
 {
-  free(frame->protocol_id);
-}
+  nghttp2_ext_altsvc *altsvc;
 
-void nghttp2_frame_blocked_init(nghttp2_blocked *frame, int32_t stream_id)
-{
-  frame_set_hd(&frame->hd, 0, NGHTTP2_BLOCKED, NGHTTP2_FLAG_NONE,
-               stream_id);
-}
+  altsvc = frame->payload;
 
-void nghttp2_frame_blocked_free(nghttp2_blocked *frame)
-{}
-
-void nghttp2_frame_data_init(nghttp2_data *frame, nghttp2_private_data *pdata)
-{
-  frame->hd = pdata->hd;
-  frame->padlen = pdata->padlen;
-  /* flags may have NGHTTP2_FLAG_END_STREAM or
-     NGHTTP2_FLAG_END_SEGMENT even if the sent chunk is not the end of
-     the stream */
-  if(!pdata->eof) {
-    frame->hd.flags &= ~(NGHTTP2_FLAG_END_STREAM | NGHTTP2_FLAG_END_SEGMENT);
+  if(altsvc == NULL) {
+    return;
   }
+
+  free(altsvc->protocol_id);
 }
 
 size_t nghttp2_frame_trail_padlen(nghttp2_frame *frame, size_t padlen)
 {
-  return padlen
-    - ((frame->hd.flags & NGHTTP2_FLAG_PAD_HIGH) > 0)
-    - ((frame->hd.flags & NGHTTP2_FLAG_PAD_LOW) > 0);
+  return padlen - ((frame->hd.flags & NGHTTP2_FLAG_PADDED) > 0);
 }
 
-void nghttp2_frame_private_data_init(nghttp2_private_data *frame,
-                                     uint8_t flags,
-                                     int32_t stream_id,
-                                     const nghttp2_data_provider *data_prd)
+void nghttp2_frame_data_init(nghttp2_data *frame, uint8_t flags,
+                             int32_t stream_id)
 {
   /* At this moment, the length of DATA frame is unknown */
-  frame_set_hd(&frame->hd, 0, NGHTTP2_DATA, flags, stream_id);
-  frame->data_prd = *data_prd;
+  nghttp2_frame_hd_init(&frame->hd, 0, NGHTTP2_DATA, flags, stream_id);
   frame->padlen = 0;
-  frame->eof = 0;
 }
 
-void nghttp2_frame_private_data_free(nghttp2_private_data *frame)
+void nghttp2_frame_data_free(nghttp2_data *frame)
 {}
 
 size_t nghttp2_frame_priority_len(uint8_t flags)
@@ -350,7 +336,7 @@ int nghttp2_frame_pack_headers(nghttp2_bufs *bufs,
                                nghttp2_hd_deflater *deflater)
 {
   size_t nv_offset;
-  ssize_t rv;
+  int rv;
   nghttp2_buf *buf;
 
   assert(bufs->head == bufs->cur);
@@ -513,8 +499,8 @@ size_t nghttp2_frame_pack_settings_payload(uint8_t *buf,
 {
   size_t i;
   for(i = 0; i < niv; ++i, buf += NGHTTP2_FRAME_SETTINGS_ENTRY_LENGTH) {
-    buf[0] = iv[i].settings_id;
-    nghttp2_put_uint32be(buf + 1, iv[i].value);
+    nghttp2_put_uint16be(buf, iv[i].settings_id);
+    nghttp2_put_uint32be(buf + 2, iv[i].value);
   }
   return NGHTTP2_FRAME_SETTINGS_ENTRY_LENGTH * niv;
 }
@@ -544,8 +530,8 @@ int nghttp2_frame_unpack_settings_payload(nghttp2_settings *frame,
 void nghttp2_frame_unpack_settings_entry(nghttp2_settings_entry *iv,
                                          const uint8_t *payload)
 {
-  iv->settings_id = payload[0];
-  iv->value = nghttp2_get_uint32(&payload[1]);
+  iv->settings_id = nghttp2_get_uint16(&payload[0]);
+  iv->value = nghttp2_get_uint32(&payload[2]);
 }
 
 int nghttp2_frame_unpack_settings_payload2(nghttp2_settings_entry **iv_ptr,
@@ -583,7 +569,7 @@ int nghttp2_frame_pack_push_promise(nghttp2_bufs *bufs,
                                     nghttp2_hd_deflater *deflater)
 {
   size_t nv_offset = 4;
-  ssize_t rv;
+  int rv;
   nghttp2_buf *buf;
 
   assert(bufs->head == bufs->cur);
@@ -691,7 +677,7 @@ void nghttp2_frame_unpack_goaway_payload(nghttp2_goaway *frame,
                                          size_t var_gift_payloadlen)
 {
   frame->last_stream_id = nghttp2_get_uint32(payload) & NGHTTP2_STREAM_ID_MASK;
-  frame->error_code = nghttp2_get_uint32(payload+4);
+  frame->error_code = nghttp2_get_uint32(payload + 4);
 
   frame->opaque_data = var_gift_payload;
   frame->opaque_data_len = var_gift_payloadlen;
@@ -759,48 +745,47 @@ void nghttp2_frame_unpack_window_update_payload(nghttp2_window_update *frame,
     NGHTTP2_WINDOW_SIZE_INCREMENT_MASK;
 }
 
-int nghttp2_frame_pack_altsvc(nghttp2_bufs *bufs, nghttp2_altsvc *frame)
+int nghttp2_frame_pack_altsvc(nghttp2_bufs *bufs, nghttp2_extension *frame)
 {
   int rv;
   nghttp2_buf *buf;
+  nghttp2_ext_altsvc *altsvc;
 
   assert(bufs->head == bufs->cur);
 
+  altsvc = frame->payload;
+
   buf = &bufs->head->buf;
 
   buf->pos -= NGHTTP2_FRAME_HDLEN;
 
   nghttp2_frame_pack_frame_hd(buf->pos, &frame->hd);
 
-  nghttp2_put_uint32be(buf->last, frame->max_age);
+  nghttp2_put_uint32be(buf->last, altsvc->max_age);
   buf->last += 4;
 
-  nghttp2_put_uint16be(buf->last, frame->port);
+  nghttp2_put_uint16be(buf->last, altsvc->port);
   buf->last += 2;
 
-  /* Reserved */
-  buf->last[0] = 0;
+  buf->last[0] = altsvc->protocol_id_len;
   ++buf->last;
 
-  buf->last[0] = frame->protocol_id_len;
-  ++buf->last;
-
-  rv = nghttp2_bufs_add(bufs, frame->protocol_id, frame->protocol_id_len);
+  rv = nghttp2_bufs_add(bufs, altsvc->protocol_id, altsvc->protocol_id_len);
   if(rv != 0) {
     goto fail;
   }
 
-  rv = nghttp2_bufs_addb(bufs, frame->host_len);
+  rv = nghttp2_bufs_addb(bufs, altsvc->host_len);
   if(rv != 0) {
     goto fail;
   }
 
-  rv = nghttp2_bufs_add(bufs, frame->host, frame->host_len);
+  rv = nghttp2_bufs_add(bufs, altsvc->host, altsvc->host_len);
   if(rv != 0) {
     goto fail;
   }
 
-  rv = nghttp2_bufs_add(bufs, frame->origin, frame->origin_len);
+  rv = nghttp2_bufs_add(bufs, altsvc->origin, altsvc->origin_len);
   if(rv != 0) {
     goto fail;
   }
@@ -816,62 +801,48 @@ int nghttp2_frame_pack_altsvc(nghttp2_bufs *bufs, nghttp2_altsvc *frame)
   return rv;
 }
 
-int nghttp2_frame_unpack_altsvc_payload(nghttp2_altsvc *frame,
+int nghttp2_frame_unpack_altsvc_payload(nghttp2_extension *frame,
                                         const uint8_t *payload,
                                         size_t payloadlen,
                                         uint8_t *var_gift_payload,
                                         size_t var_gift_payloadlen)
 {
   nghttp2_buf buf;
+  nghttp2_ext_altsvc *altsvc;
 
-  frame->max_age = nghttp2_get_uint32(payload);
+  altsvc = frame->payload;
+
+  altsvc->max_age = nghttp2_get_uint32(payload);
   payload += 4;
 
-  frame->port = nghttp2_get_uint16(payload);
+  altsvc->port = nghttp2_get_uint16(payload);
   payload += 2;
 
-  /* Skip Reserved */
-  ++payload;
-
-  frame->protocol_id_len = *payload;
+  altsvc->protocol_id_len = *payload;
 
   nghttp2_buf_wrap_init(&buf, var_gift_payload, var_gift_payloadlen);
   buf.last += var_gift_payloadlen;
 
-  /* 1 for HOST_LEN */
-  if(nghttp2_buf_len(&buf) < 1 + (ssize_t)frame->protocol_id_len) {
+  /* 1 for Host-Len */
+  if(nghttp2_buf_len(&buf) < 1 + (ssize_t)altsvc->protocol_id_len) {
     return NGHTTP2_ERR_FRAME_SIZE_ERROR;
   }
 
-  frame->protocol_id = buf.pos;
-  buf.pos += frame->protocol_id_len;
+  altsvc->protocol_id = buf.pos;
+  buf.pos += altsvc->protocol_id_len;
 
-  frame->host_len = *buf.pos;
+  altsvc->host_len = *buf.pos;
   ++buf.pos;
 
-  if(nghttp2_buf_len(&buf) < (ssize_t)frame->host_len) {
+  if(nghttp2_buf_len(&buf) < (ssize_t)altsvc->host_len) {
     return NGHTTP2_ERR_FRAME_SIZE_ERROR;
   }
 
-  frame->host = buf.pos;
-  buf.pos += frame->host_len;
+  altsvc->host = buf.pos;
+  buf.pos += altsvc->host_len;
 
-  frame->origin = buf.pos;
-  frame->origin_len = nghttp2_buf_len(&buf);
-
-  return 0;
-}
-
-int nghttp2_frame_pack_blocked(nghttp2_bufs *bufs, nghttp2_blocked *frame)
-{
-  nghttp2_buf *buf;
-
-  assert(bufs->head == bufs->cur);
-
-  buf = &bufs->head->buf;
-  buf->pos -= NGHTTP2_FRAME_HDLEN;
-
-  nghttp2_frame_pack_frame_hd(buf->pos, &frame->hd);
+  altsvc->origin = buf.pos;
+  altsvc->origin_len = nghttp2_buf_len(&buf);
 
   return 0;
 }
@@ -962,8 +933,8 @@ void nghttp2_nv_array_sort(nghttp2_nv *nva, size_t nvlen)
   qsort(nva, nvlen, sizeof(nghttp2_nv), nv_compar);
 }
 
-ssize_t nghttp2_nv_array_copy(nghttp2_nv **nva_ptr,
-                              const nghttp2_nv *nva, size_t nvlen)
+int nghttp2_nv_array_copy(nghttp2_nv **nva_ptr,
+                          const nghttp2_nv *nva, size_t nvlen)
 {
   size_t i;
   uint8_t *data;
@@ -974,8 +945,7 @@ ssize_t nghttp2_nv_array_copy(nghttp2_nv **nva_ptr,
     buflen += nva[i].namelen + nva[i].valuelen;
   }
 
-  /* If all name/value pair is 0-length, remove them */
-  if(nvlen == 0 || buflen == 0) {
+  if(nvlen == 0) {
     *nva_ptr = NULL;
 
     return 0;
@@ -1006,7 +976,7 @@ ssize_t nghttp2_nv_array_copy(nghttp2_nv **nva_ptr,
     data += nva[i].valuelen;
     ++p;
   }
-  return nvlen;
+  return 0;
 }
 
 int nghttp2_iv_check(const nghttp2_settings_entry *iv, size_t niv)
@@ -1022,7 +992,6 @@ int nghttp2_iv_check(const nghttp2_settings_entry *iv, size_t niv)
     case NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS:
       break;
     case NGHTTP2_SETTINGS_ENABLE_PUSH:
-    case NGHTTP2_SETTINGS_COMPRESS_DATA:
       if(iv[i].value != 0 && iv[i].value != 1) {
         return 0;
       }
@@ -1032,8 +1001,14 @@ int nghttp2_iv_check(const nghttp2_settings_entry *iv, size_t niv)
         return 0;
       }
       break;
-    default:
-      return 0;
+    case NGHTTP2_SETTINGS_MAX_FRAME_SIZE:
+      if(iv[i].value < NGHTTP2_MAX_FRAME_SIZE_MIN ||
+         iv[i].value > NGHTTP2_MAX_FRAME_SIZE_MAX) {
+        return 0;
+      }
+      break;
+    case NGHTTP2_SETTINGS_MAX_HEADER_LIST_SIZE:
+      break;
     }
   }
   return 1;
@@ -1042,41 +1017,21 @@ int nghttp2_iv_check(const nghttp2_settings_entry *iv, size_t niv)
 static void frame_set_pad(nghttp2_buf *buf, size_t padlen)
 {
   size_t trail_padlen;
+  size_t newlen;
 
-  if(padlen > 256) {
-    DEBUGF(fprintf(stderr, "send: padlen=%zu, shift left 2 bytes\n", padlen));
+  DEBUGF(fprintf(stderr, "send: padlen=%zu, shift left 1 bytes\n", padlen));
 
-    memmove(buf->pos - 2, buf->pos, NGHTTP2_FRAME_HDLEN);
+  memmove(buf->pos - 1, buf->pos, NGHTTP2_FRAME_HDLEN);
 
-    buf->pos -= 2;
+  --buf->pos;
 
-    buf->pos[3] |= NGHTTP2_FLAG_PAD_HIGH | NGHTTP2_FLAG_PAD_LOW;
+  buf->pos[4] |= NGHTTP2_FLAG_PADDED;
 
-    nghttp2_put_uint16be(buf->pos, nghttp2_get_uint16(buf->pos) + padlen);
+  newlen = (nghttp2_get_uint32(buf->pos) >> 8) + padlen;
+  nghttp2_put_uint32be(buf->pos, (uint32_t)((newlen << 8) + buf->pos[3]));
 
-    trail_padlen = padlen - 2;
-    buf->pos[NGHTTP2_FRAME_HDLEN] = trail_padlen >> 8;
-    buf->pos[NGHTTP2_FRAME_HDLEN + 1] = trail_padlen & 0xff;
-
-  } else if(padlen > 0) {
-    DEBUGF(fprintf(stderr, "send: padlen=%zu, shift left 1 bytes\n", padlen));
-
-    memmove(buf->pos - 1, buf->pos, NGHTTP2_FRAME_HDLEN);
-
-    --buf->pos;
-
-    buf->pos[3] |= NGHTTP2_FLAG_PAD_LOW;
-
-    nghttp2_put_uint16be(buf->pos, nghttp2_get_uint16(buf->pos) + padlen);
-
-    trail_padlen = padlen - 1;
-    buf->pos[NGHTTP2_FRAME_HDLEN] = trail_padlen;
-
-  } else {
-    DEBUGF(fprintf(stderr, "send: padlen=0, no shift left was made\n"));
-
-    return;
-  }
+  trail_padlen = padlen - 1;
+  buf->pos[NGHTTP2_FRAME_HDLEN] = trail_padlen;
 
   /* zero out padding */
   memset(buf->last, 0, trail_padlen);
@@ -1088,13 +1043,9 @@ static void frame_set_pad(nghttp2_buf *buf, size_t padlen)
 }
 
 int nghttp2_frame_add_pad(nghttp2_bufs *bufs, nghttp2_frame_hd *hd,
-                          size_t padlen, nghttp2_frame_type type)
+                          size_t padlen)
 {
-  int rv;
-  size_t trail_padlen;
-  size_t last_avail;
   nghttp2_buf *buf;
-  nghttp2_frame_hd last_hd;
 
   if(padlen == 0) {
     DEBUGF(fprintf(stderr, "send: padlen = 0, nothing to do\n"));
@@ -1108,87 +1059,28 @@ int nghttp2_frame_add_pad(nghttp2_bufs *bufs, nghttp2_frame_hd *hd,
    *  0                   1                   2                   3
    *  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
-   * | | |Frame header   | Frame payload...                          |
-   * +-+-+---------------+-------------------------------------------+
-   * | | |Frame header   | Frame payload...                          |
-   * +-+-+---------------+-------------------------------------------+
-   * | | |Frame header   | Frame payload...                          |
-   * +-+-+---------------+-------------------------------------------+
+   * | |Frame header     | Frame payload...                          :
+   * +-+-----------------+-------------------------------------------+
+   * | |Frame header     | Frame payload...                          :
+   * +-+-----------------+-------------------------------------------+
+   * | |Frame header     | Frame payload...                          :
+   * +-+-----------------+-------------------------------------------+
    *
-   * Since we limit the length of the padding bytes, they are
-   * completely included in one frame payload or across 2 frames.  We
-   * are going to adjust buf->pos of frame which includes part of
-   * padding. And serialize (memmove) frame header in the correct
-   * position. Also extends buf->last to include padding.
+   * We arranged padding so that it is included in the first frame
+   * completely.  For padded frame, we are going to adjust buf->pos of
+   * frame which includes padding and serialize (memmove) frame header
+   * in the correct position.  Also extends buf->last to include
+   * padding.
    */
 
-  nghttp2_bufs_seek_last_present(bufs);
+  buf = &bufs->head->buf;
 
-  buf = &bufs->cur->buf;
+  assert(nghttp2_buf_avail(buf) >= (ssize_t)padlen - 1);
 
-  last_avail = nghttp2_buf_avail(buf);
-
-  if(last_avail >= padlen) {
-    /* Last frame can include all paddings bytes */
-    DEBUGF(fprintf(stderr, "send: last frame includes all paddings\n"));
-
-    frame_set_pad(buf, padlen);
-
-  } else {
-    /* padding across 2 frames */
-
-    /* type = DATA must not be here */
-    assert(type == NGHTTP2_CONTINUATION);
-
-    /* This will seek to the last chain */
-    rv = nghttp2_bufs_advance(bufs);
-    if(rv == NGHTTP2_ERR_BUFFER_ERROR) {
-      return NGHTTP2_ERR_FRAME_SIZE_ERROR;
-    }
-
-    if(rv != 0) {
-      return rv;
-    }
-
-    if(type == NGHTTP2_CONTINUATION) {
-      /* former last frame has END_HEADERS flag set. Clear it. */
-      buf->pos[3] &= ~NGHTTP2_FLAG_END_HEADERS;
-    }
-
-    trail_padlen = nghttp2_buf_avail(buf);
-
-    /* former last frame may have zero buffer available */
-    if(trail_padlen == 0) {
-      DEBUGF(fprintf(stderr,
-                     "send: last frame has no space to include padding\n"));
-    } else {
-      DEBUGF(fprintf(stderr, "send: padding across 2 frames\n"));
-
-      frame_set_pad(buf, trail_padlen);
-    }
-
-    /* This buffer does not have frame header serialized */
-    buf = &bufs->cur->buf;
-
-    trail_padlen = padlen - trail_padlen;
-
-    last_hd.length = 0;
-    last_hd.type = type;
-    last_hd.stream_id = hd->stream_id;
-
-    if(type == NGHTTP2_CONTINUATION) {
-      last_hd.flags = NGHTTP2_FLAG_END_HEADERS;
-    } else {
-      last_hd.flags = NGHTTP2_FLAG_NONE;
-    }
-
-    buf->pos -= NGHTTP2_FRAME_HDLEN;
-    nghttp2_frame_pack_frame_hd(buf->pos, &last_hd);
-
-    frame_set_pad(buf, trail_padlen);
-  }
+  frame_set_pad(buf, padlen);
 
   hd->length += padlen;
+  hd->flags |= NGHTTP2_FLAG_PADDED;
 
   DEBUGF(fprintf(stderr, "send: final payloadlen=%zu, padlen=%zu\n",
                  hd->length, padlen));
