@@ -33,7 +33,6 @@
 #include "nghttp2_hd.h"
 #include "nghttp2_buf.h"
 
-#define NGHTTP2_FRAME_LENGTH_MASK ((1 << 14) - 1)
 #define NGHTTP2_STREAM_ID_MASK ((1u << 31) - 1)
 #define NGHTTP2_PRI_GROUP_ID_MASK ((1u << 31) - 1)
 #define NGHTTP2_PRIORITY_MASK ((1u << 31) - 1)
@@ -41,21 +40,31 @@
 #define NGHTTP2_SETTINGS_ID_MASK ((1 << 24) - 1)
 
 /* The number of bytes of frame header. */
-#define NGHTTP2_FRAME_HDLEN 8
+#define NGHTTP2_FRAME_HDLEN 9
 
-#define NGHTTP2_MAX_PAYLOADLEN 16383
+#define NGHTTP2_MAX_FRAME_SIZE_MAX ((1 << 24) - 1)
+#define NGHTTP2_MAX_FRAME_SIZE_MIN (1 << 14)
+
+#define NGHTTP2_MAX_PAYLOADLEN 16384
 /* The one frame buffer length for tranmission.  We may use several of
-   them to support CONTINUATION.  To account for padding specifiers
-   (PAD_HIGH and PAD_LOW), we allocate extra 2 bytes, which saves
-   extra large memcopying. */
+   them to support CONTINUATION.  To account for Pad Length field, we
+   allocate extra 1 byte, which saves extra large memcopying. */
 #define NGHTTP2_FRAMEBUF_CHUNKLEN \
-  (NGHTTP2_FRAME_HDLEN + 2 + NGHTTP2_MAX_PAYLOADLEN)
+  (NGHTTP2_FRAME_HDLEN + 1 + NGHTTP2_MAX_PAYLOADLEN)
 
-/* The maximum length of DATA frame payload. */
-#define NGHTTP2_DATA_PAYLOADLEN 4096
+/* Number of inbound buffer */
+#define NGHTTP2_FRAMEBUF_MAX_NUM 5
+
+/* The default length of DATA frame payload. This should be small enough
+ * for the data payload and the header to fit into 1 TLS record */
+#define NGHTTP2_DATA_PAYLOADLEN ((NGHTTP2_MAX_FRAME_SIZE_MIN) - (NGHTTP2_FRAME_HDLEN))
+
+/* Maximum headers payload length, calculated in compressed form.
+   This applies to transmission only. */
+#define NGHTTP2_MAX_HEADERSLEN 65536
 
 /* The number of bytes for each SETTINGS entry */
-#define NGHTTP2_FRAME_SETTINGS_ENTRY_LENGTH 5
+#define NGHTTP2_FRAME_SETTINGS_ENTRY_LENGTH 6
 
 /* The maximum header table size in SETTINGS_HEADER_TABLE_SIZE */
 #define NGHTTP2_MAX_HEADER_TABLE_SIZE ((1u << 31) - 1)
@@ -64,53 +73,33 @@
 #define NGHTTP2_PRIORITY_SPECLEN 5
 
 /* Length of fixed part in ALTSVC frame, that is the sum of fields of
-   Max-Age, Port, Reserved and PID_LEN. */
-#define NGHTTP2_ALTSVC_FIXED_PARTLEN 8
+   Max-Age, Port and Proto-Len. */
+#define NGHTTP2_ALTSVC_FIXED_PARTLEN 7
 
-/* Minimum length of ALTSVC frame.  NGHTTP2_ALTSVC_FIXED_PARTLEN +
-   HOST_LEN. */
-#define NGHTTP2_ALTSVC_MINLEN 9
+/* Minimum length of ALTSVC extension frame payload.
+   NGHTTP2_ALTSVC_FIXED_PARTLEN + Host-Len. */
+#define NGHTTP2_ALTSVC_MINLEN 8
 
+/* Maximum length of padding in bytes. */
+#define NGHTTP2_MAX_PADLEN 256
 
-/* Category of frames. */
-typedef enum {
-  /* non-DATA frame */
-  NGHTTP2_CAT_CTRL,
-  /* DATA frame */
-  NGHTTP2_CAT_DATA
-} nghttp2_frame_category;
-
-/**
- * @struct
- *
- * The DATA frame used in the library privately. It has the following
- * members:
- */
-typedef struct {
-  nghttp2_frame_hd hd;
-  /**
-   * The data to be sent for this DATA frame.
-   */
-  nghttp2_data_provider data_prd;
-  /**
-   * The number of bytes added as padding. This includes PAD_HIGH and
-   * PAD_LOW.
-   */
-  size_t padlen;
-  /**
-   * The flag to indicate whether EOF was reached or not. Initially
-   * |eof| is 0. It becomes 1 after all data were read. This is used
-   * exclusively by nghttp2 library and not in the spec.
-   */
-  uint8_t eof;
-} nghttp2_private_data;
-
-int nghttp2_frame_is_data_frame(uint8_t *head);
+/* Union of extension frame payload */
+typedef union {
+  nghttp2_ext_altsvc altsvc;
+} nghttp2_ext_frame_payload;
 
 void nghttp2_frame_pack_frame_hd(uint8_t *buf, const nghttp2_frame_hd *hd);
 
 void nghttp2_frame_unpack_frame_hd(nghttp2_frame_hd *hd, const uint8_t* buf);
 
+/**
+ * Initializes frame header |hd| with given parameters.  Reserved bit
+ * is set to 0.
+ */
+void nghttp2_frame_hd_init(nghttp2_frame_hd *hd, size_t length,
+                           uint8_t type, uint8_t flags,
+                           int32_t stream_id);
+
 /**
  * Returns the number of priority field depending on the |flags|.  If
  * |flags| has neither NGHTTP2_FLAG_PRIORITY_GROUP nor
@@ -169,7 +158,7 @@ int nghttp2_frame_pack_headers(nghttp2_bufs *bufs,
 /*
  * Unpacks HEADERS frame byte sequence into |frame|.  This function
  * only unapcks bytes that come before name/value header block and
- * after PAD_HIGH and PAD_LOW.
+ * after possible Pad Length field.
  *
  * This function always succeeds and returns 0.
  */
@@ -301,7 +290,7 @@ int nghttp2_frame_pack_push_promise(nghttp2_bufs *bufs,
 /*
  * Unpacks PUSH_PROMISE frame byte sequence into |frame|.  This
  * function only unapcks bytes that come before name/value header
- * block and after PAD_HIGH and PAD_LOW.
+ * block and after possible Pad Length field.
  *
  * This function returns 0 if it succeeds or one of the following
  * negative error codes:
@@ -403,6 +392,9 @@ void nghttp2_frame_unpack_window_update_payload(nghttp2_window_update *frame,
  * The caller must make sure that nghttp2_bufs_reset(bufs) is called
  * before calling this function.
  *
+ * The caller must make sure that frame->payload points to
+ * nghttp2_ext_altsvc object.
+ *
  * This function returns 0 if it succeeds or one of the following
  * negative error codes:
  *
@@ -411,7 +403,7 @@ void nghttp2_frame_unpack_window_update_payload(nghttp2_window_update *frame,
  * NGHTTP2_ERR_FRAME_SIZE_ERROR
  *     The length of the frame is too large.
  */
-int nghttp2_frame_pack_altsvc(nghttp2_bufs *bufs, nghttp2_altsvc *frame);
+int nghttp2_frame_pack_altsvc(nghttp2_bufs *bufs, nghttp2_extension *frame);
 
 
 /*
@@ -422,28 +414,21 @@ int nghttp2_frame_pack_altsvc(nghttp2_bufs *bufs, nghttp2_altsvc *frame);
  * and then |frame|.  The |var_gift_payloadlen| must be freed by
  * nghttp2_frame_altsvc_free().
  *
+ * The caller must make sure that frame->payload points to
+ * nghttp2_ext_altsvc object.
+ *
  * This function returns 0 if it succeeds or one of the following
  * negative error codes:
  *
  * NGHTTP2_ERR_FRAME_SIZE_ERROR
  *   The |var_gift_payload| does not contain required data.
  */
-int nghttp2_frame_unpack_altsvc_payload(nghttp2_altsvc *frame,
+int nghttp2_frame_unpack_altsvc_payload(nghttp2_extension *frame,
                                         const uint8_t *payload,
                                         size_t payloadlen,
                                         uint8_t *var_gift_payload,
                                         size_t var_gift_payloadlen);
 
-/*
- * Packs BLOCKED frame |frame| in wire format and store it in |bufs|.
- *
- * The caller must make sure that nghttp2_bufs_reset(bufs) is called
- * before calling this function.
- *
- * This function always returns 0.
- */
-int nghttp2_frame_pack_blocked(nghttp2_bufs *bufs, nghttp2_blocked *frame);
-
 /*
  * Initializes HEADERS frame |frame| with given values.  |frame| takes
  * ownership of |nva|, so caller must not free it. If |stream_id| is
@@ -465,7 +450,7 @@ void nghttp2_frame_priority_free(nghttp2_priority *frame);
 
 void nghttp2_frame_rst_stream_init(nghttp2_rst_stream *frame,
                                    int32_t stream_id,
-                                   nghttp2_error_code error_code);
+                                   uint32_t error_code);
 
 void nghttp2_frame_rst_stream_free(nghttp2_rst_stream *frame);
 
@@ -507,7 +492,7 @@ void nghttp2_frame_ping_free(nghttp2_ping *frame);
  * free it. If the |opaque_data_len| is 0, opaque_data could be NULL.
  */
 void nghttp2_frame_goaway_init(nghttp2_goaway *frame, int32_t last_stream_id,
-                               nghttp2_error_code error_code,
+                               uint32_t error_code,
                                uint8_t *opaque_data, size_t opaque_data_len);
 
 void nghttp2_frame_goaway_free(nghttp2_goaway *frame);
@@ -525,7 +510,7 @@ void nghttp2_frame_window_update_free(nghttp2_window_update *frame);
    |protocol_id_len| == 0 and |host_len| + |origin_len| > 0.  If
    |protocol_id_len|, |host_len| and |origin_len| are all zero,
    |protocol_id| can be NULL. */
-void nghttp2_frame_altsvc_init(nghttp2_altsvc *frame, int32_t stream_id,
+void nghttp2_frame_altsvc_init(nghttp2_extension *frame, int32_t stream_id,
                                uint32_t max_age,
                                uint16_t port,
                                uint8_t *protocol_id,
@@ -533,27 +518,23 @@ void nghttp2_frame_altsvc_init(nghttp2_altsvc *frame, int32_t stream_id,
                                uint8_t *host, size_t host_len,
                                uint8_t *origin, size_t origin_len);
 
-void nghttp2_frame_altsvc_free(nghttp2_altsvc *frame);
-
-void nghttp2_frame_blocked_init(nghttp2_blocked *frame, int32_t stream_id);
-
-void nghttp2_frame_blocked_free(nghttp2_blocked *frame);
-
-void nghttp2_frame_data_init(nghttp2_data *frame, nghttp2_private_data *pdata);
+/*
+ *  Frees resources used by |frame|.  This function does not free
+ *  frame->payload itself.
+ */
+void nghttp2_frame_altsvc_free(nghttp2_extension *frame);
 
 /*
- * Returns the number of padding bytes after payload. The total
- * padding length is given in the |padlen|. The returned value does
- * not include the PAD_HIGH and PAD_LOW.
+ * Returns the number of padding bytes after payload.  The total
+ * padding length is given in the |padlen|.  The returned value does
+ * not include the Pad Length field.
  */
 size_t nghttp2_frame_trail_padlen(nghttp2_frame *frame, size_t padlen);
 
-void nghttp2_frame_private_data_init(nghttp2_private_data *frame,
-                                     uint8_t flags,
-                                     int32_t stream_id,
-                                     const nghttp2_data_provider *data_prd);
+void nghttp2_frame_data_init(nghttp2_data *frame, uint8_t flags,
+                             int32_t stream_id);
 
-void nghttp2_frame_private_data_free(nghttp2_private_data *frame);
+void nghttp2_frame_data_free(nghttp2_data *frame);
 
 /*
  * Makes copy of |iv| and return the copy. The |niv| is the number of
@@ -576,14 +557,14 @@ void nghttp2_nv_array_sort(nghttp2_nv *nva, size_t nvlen);
  *
  * The |*nva_ptr| must be freed using nghttp2_nv_array_del().
  *
- * This function returns the number of name/value pairs in |*nva_ptr|,
- * or one of the following negative error codes:
+ * This function returns 0 if it succeeds or one of the following
+ * negative error codes:
  *
  * NGHTTP2_ERR_NOMEM
  *     Out of memory.
  */
-ssize_t nghttp2_nv_array_copy(nghttp2_nv **nva_ptr,
-                              const nghttp2_nv *nva, size_t nvlen);
+int nghttp2_nv_array_copy(nghttp2_nv **nva_ptr,
+                          const nghttp2_nv *nva, size_t nvlen);
 
 /*
  * Returns nonzero if the name/value pair |a| equals to |b|. The name
@@ -606,11 +587,10 @@ void nghttp2_nv_array_del(nghttp2_nv *nva);
 int nghttp2_iv_check(const nghttp2_settings_entry *iv, size_t niv);
 
 /*
- * Sets PAD_HIGH and PAD_LOW fields, flags and adjust frame header
- * position of each buffers in |bufs|.  The padding is given in the
- * |padlen|. The |hd| is the frame header for the serialized data.
- * The |type| is used as a frame type when padding requires additional
- * buffers.
+ * Sets Pad Length field and flags and adjusts frame header position
+ * of each buffers in |bufs|.  The number of padding is given in the
+ * |padlen| including Pad Length field.  The |hd| is the frame header
+ * for the serialized data.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -621,6 +601,6 @@ int nghttp2_iv_check(const nghttp2_settings_entry *iv, size_t niv);
  *     The length of the resulting frame is too large.
  */
 int nghttp2_frame_add_pad(nghttp2_bufs *bufs, nghttp2_frame_hd *hd,
-                          size_t padlen, nghttp2_frame_type type);
+                          size_t padlen);
 
 #endif /* NGHTTP2_FRAME_H */
