@@ -38,13 +38,14 @@
 #include "nghttp2_outbound_item.h"
 #include "nghttp2_int.h"
 #include "nghttp2_buf.h"
+#include "nghttp2_callbacks.h"
 
 /*
  * Option flags.
  */
 typedef enum {
-  NGHTTP2_OPTMASK_NO_AUTO_STREAM_WINDOW_UPDATE = 1 << 0,
-  NGHTTP2_OPTMASK_NO_AUTO_CONNECTION_WINDOW_UPDATE = 1 << 1
+  NGHTTP2_OPTMASK_NO_AUTO_WINDOW_UPDATE = 1 << 0,
+  NGHTTP2_OPTMASK_RECV_CLIENT_PREFACE = 1 << 1,
 } nghttp2_optmask;
 
 typedef enum {
@@ -65,6 +66,8 @@ typedef struct {
 /* Internal state when receiving incoming frame */
 typedef enum {
   /* Receiving frame header */
+  NGHTTP2_IB_READ_CLIENT_PREFACE,
+  NGHTTP2_IB_READ_FIRST_SETTINGS,
   NGHTTP2_IB_READ_HEAD,
   NGHTTP2_IB_READ_NBYTE,
   NGHTTP2_IB_READ_HEADER_BLOCK,
@@ -76,19 +79,23 @@ typedef enum {
   NGHTTP2_IB_READ_ALTSVC,
   NGHTTP2_IB_EXPECT_CONTINUATION,
   NGHTTP2_IB_IGN_CONTINUATION,
-  NGHTTP2_IB_READ_PAD_CONTINUATION,
-  NGHTTP2_IB_IGN_PAD_CONTINUATION,
   NGHTTP2_IB_READ_PAD_DATA,
   NGHTTP2_IB_READ_DATA,
   NGHTTP2_IB_IGN_DATA
 } nghttp2_inbound_state;
 
+#define NGHTTP2_INBOUND_NUM_IV 7
+
 typedef struct {
   nghttp2_frame frame;
+  /* Storage for extension frame payload.  frame->ext.payload points
+     to this structure to avoid frequent memory allocation. */
+  nghttp2_ext_frame_payload ext_frame_payload;
   /* The received SETTINGS entry. The protocol says that we only cares
      about the defined settings ID. If unknown ID is received, it is
-     subject to connection error */
-  nghttp2_settings_entry iv[5];
+     ignored.  We use last entry to hold minimum header table size if
+     same settings are multiple times. */
+  nghttp2_settings_entry iv[NGHTTP2_INBOUND_NUM_IV];
   /* buffer pointers to small buffer, raw_sbuf */
   nghttp2_buf sbuf;
   /* buffer pointers to large buffer, raw_lbuf */
@@ -102,9 +109,21 @@ typedef struct {
   /* padding length for the current frame */
   size_t padlen;
   nghttp2_inbound_state state;
-  uint8_t raw_sbuf[8];
+  /* Small buffer.  Currently the largest contiguous chunk to buffer
+     is frame header.  We buffer part of payload, but they are smaller
+     than frame header. */
+  uint8_t raw_sbuf[NGHTTP2_FRAME_HDLEN];
 } nghttp2_inbound_frame;
 
+typedef struct {
+  uint32_t header_table_size;
+  uint32_t enable_push;
+  uint32_t max_concurrent_streams;
+  uint32_t initial_window_size;
+  uint32_t max_frame_size;
+  uint32_t max_header_set_size;
+} nghttp2_settings_storage;
+
 typedef enum {
   NGHTTP2_GOAWAY_NONE = 0,
   /* Flag means GOAWAY frame is sent to the remote peer. */
@@ -118,10 +137,13 @@ typedef enum {
 struct nghttp2_session {
   nghttp2_map /* <nghttp2_stream*> */ streams;
   nghttp2_stream_roots roots;
-  /* Queue for outbound frames other than stream-creating HEADERS */
+  /* Queue for outbound frames other than stream-creating HEADERS and
+     DATA */
   nghttp2_pq /* <nghttp2_outbound_item*> */ ob_pq;
   /* Queue for outbound stream-creating HEADERS frame */
   nghttp2_pq /* <nghttp2_outbound_item*> */ ob_ss_pq;
+  /* QUeue for DATA frame */
+  nghttp2_pq /* <nghttp2_outbound_item*> */ ob_da_pq;
   nghttp2_active_outbound_item aob;
   nghttp2_inbound_frame iframe;
   nghttp2_hd_deflater hd_deflater;
@@ -158,10 +180,10 @@ struct nghttp2_session {
      in-flight SETTINGS. */
   ssize_t inflight_niv;
   /* The number of outgoing streams. This will be capped by
-     remote_settings[NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS]. */
+     remote_settings.max_concurrent_streams. */
   size_t num_outgoing_streams;
   /* The number of incoming streams. This will be capped by
-     local_settings[NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS]. */
+     local_settings.max_concurrent_streams. */
   size_t num_incoming_streams;
   /* The number of closed streams still kept in |streams| hash.  The
      closed streams can be accessed through single linked list
@@ -181,8 +203,10 @@ struct nghttp2_session {
   /* Counter of unique ID of PING. Wraps when it exceeds
      NGHTTP2_MAX_UNIQUE_ID */
   uint32_t next_unique_id;
+  /* This is the last-stream-ID we have sent in GOAWAY */
+  int32_t local_last_stream_id;
   /* This is the value in GOAWAY frame received from remote endpoint. */
-  int32_t last_stream_id;
+  int32_t remote_last_stream_id;
   /* Current sender window size. This value is computed against the
      current initial window size of remote endpoint. */
   int32_t remote_window_size;
@@ -190,6 +214,10 @@ struct nghttp2_session {
      WINDOW_UPDATE. This could be negative after submitting negative
      value to WINDOW_UPDATE. */
   int32_t recv_window_size;
+  /* The number of bytes consumed by the application and now is
+     subject to WINDOW_UPDATE.  This is only used when auto
+     WINDOW_UPDATE is turned off. */
+  int32_t consumed_size;
   /* The amount of recv_window_size cut using submitting negative
      value to WINDOW_UPDATE */
   int32_t recv_reduction;
@@ -200,9 +228,9 @@ struct nghttp2_session {
   int32_t local_window_size;
   /* Settings value received from the remote endpoint. We just use ID
      as index. The index = 0 is unused. */
-  uint32_t remote_settings[NGHTTP2_SETTINGS_MAX+1];
+  nghttp2_settings_storage remote_settings;
   /* Settings value of the local endpoint. */
-  uint32_t local_settings[NGHTTP2_SETTINGS_MAX+1];
+  nghttp2_settings_storage local_settings;
   /* Option flags. This is bitwise-OR of 0 or more of nghttp2_optmask. */
   uint32_t opt_flags;
   /* Unacked local SETTINGS_MAX_CONCURRENT_STREAMS value. We use this
@@ -213,9 +241,6 @@ struct nghttp2_session {
   /* Flags indicating GOAWAY is sent and/or recieved. The flags are
      composed by bitwise OR-ing nghttp2_goaway_flag. */
   uint8_t goaway_flags;
-  /* nonzero if blocked was sent and remote_window_size is still 0 or
-     negative */
-  uint8_t blocked_sent;
 };
 
 /* Struct used when updating initial window size of each active
@@ -235,24 +260,28 @@ int nghttp2_session_is_my_stream_id(nghttp2_session *session,
                                     int32_t stream_id);
 
 /*
- * Adds frame |frame| to the outbound queue in |session|. The
- * |frame_cat| must be either NGHTTP2_CTRL or NGHTTP2_DATA. If the
- * |frame_cat| is NGHTTP2_CTRL, the |frame| must be a pointer to
- * nghttp2_frame. If the |frame_cat| is NGHTTP2_DATA, it must be a
- * pointer to nghttp2_private_data. |aux_data| is a pointer to the arbitrary
- * data. Its interpretation is defined per the type of the frame. When
- * this function succeeds, it takes ownership of |frame| and
- * |aux_data|, so caller must not free them on success.
+ * Initializes |item|.  No memory allocation is done in this function.
+ * Don't call nghttp2_outbound_item_free() until frame member is
+ * initialized.
+ */
+void nghttp2_session_outbound_item_init(nghttp2_session* session,
+                                        nghttp2_outbound_item *item);
+
+/*
+ * Adds |item| to the outbound queue in |session|.  When this function
+ * succeeds, it takes ownership of |item|. So caller must not free it
+ * on success.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
  *
  * NGHTTP2_ERR_NOMEM
  *     Out of memory.
+ * NGHTTP2_ERR_STREAM_CLOSED
+ *     Stream already closed (DATA frame only)
  */
-int nghttp2_session_add_frame(nghttp2_session *session,
-                              nghttp2_frame_category frame_cat,
-                              void *abs_frame, void *aux_data);
+int nghttp2_session_add_item(nghttp2_session *session,
+                             nghttp2_outbound_item *item);
 
 /*
  * Adds RST_STREAM frame for the stream |stream_id| with the error
@@ -271,7 +300,7 @@ int nghttp2_session_add_frame(nghttp2_session *session,
  */
 int nghttp2_session_add_rst_stream(nghttp2_session *session,
                                    int32_t stream_id,
-                                   nghttp2_error_code error_code);
+                                   uint32_t error_code);
 
 /*
  * Adds PING frame. This is a convenient functin built on top of
@@ -288,7 +317,7 @@ int nghttp2_session_add_rst_stream(nghttp2_session *session,
  *     Out of memory.
  */
 int nghttp2_session_add_ping(nghttp2_session *session, uint8_t flags,
-                             uint8_t *opaque_data);
+                             const uint8_t *opaque_data);
 
 /*
  * Adds GOAWAY frame with the last-stream-ID |last_stream_id| and the
@@ -305,7 +334,7 @@ int nghttp2_session_add_ping(nghttp2_session *session, uint8_t flags,
  */
 int nghttp2_session_add_goaway(nghttp2_session *session,
                                int32_t last_stream_id,
-                               nghttp2_error_code error_code,
+                               uint32_t error_code,
                                const uint8_t *opaque_data,
                                size_t opaque_data_len);
 
@@ -377,7 +406,7 @@ nghttp2_stream* nghttp2_session_open_stream(nghttp2_session *session,
  *     The callback function failed.
  */
 int nghttp2_session_close_stream(nghttp2_session *session, int32_t stream_id,
-                                 nghttp2_error_code error_code);
+                                 uint32_t error_code);
 
 /*
  * Deletes |stream| from memory.  After this function returns, stream
@@ -590,19 +619,6 @@ int nghttp2_session_on_window_update_received(nghttp2_session *session,
 int nghttp2_session_on_altsvc_received(nghttp2_session *session,
                                        nghttp2_frame *frame);
 
-/*
- * Called when BLOCKED is received, assuming |frame| is properly
- * initialized.
- *
- * This function returns 0 if it succeeds, or one of the following
- * negative error codes:
- *
- * NGHTTP2_ERR_CALLBACK_FAILURE
- *   The callback function failed.
- */
-int nghttp2_session_on_blocked_received(nghttp2_session *session,
-                                        nghttp2_frame *frame);
-
 /*
  * Called when DATA is received, assuming |frame| is properly
  * initialized.
@@ -635,13 +651,8 @@ nghttp2_stream* nghttp2_session_get_stream_raw(nghttp2_session *session,
 
 /*
  * Packs DATA frame |frame| in wire frame format and stores it in
- * |*buf_ptr|.  The capacity of |*buf_ptr| is |*buflen_ptr|
- * length. This function expands |*buf_ptr| as necessary to store
- * given |frame|. It packs header in first 8 bytes starting
- * |*bufoff_ptr| offset. The |*bufoff_ptr| is calculated based on
- * usage of padding. Remaining bytes are the DATA apyload and are
- * filled using |frame->data_prd|. The length of payload is at most
- * |datamax| bytes.
+ * |bufs|.  Payload will be read using |aux_data->data_prd|.  The
+ * length of payload is at most |datamax| bytes.
  *
  * This function returns 0 if it succeeds, or one of the following
  * negative error codes:
@@ -658,7 +669,8 @@ nghttp2_stream* nghttp2_session_get_stream_raw(nghttp2_session *session,
 int nghttp2_session_pack_data(nghttp2_session *session,
                               nghttp2_bufs *bufs,
                               size_t datamax,
-                              nghttp2_private_data *frame);
+                              nghttp2_frame *frame,
+                              nghttp2_data_aux_data *aux_data);
 
 /*
  * Returns top of outbound frame queue. This function returns NULL if
@@ -720,4 +732,19 @@ int nghttp2_session_reprioritize_stream
 (nghttp2_session *session, nghttp2_stream *stream,
  const nghttp2_priority_spec *pri_spec);
 
+/*
+ * Terminates current |session| with the |error_code|.  The |reason|
+ * is NULL-terminated debug string.
+ *
+ * This function returns 0 if it succeeds, or one of the following
+ * negative error codes:
+ *
+ * NGHTTP2_ERR_NOMEM
+ *     Out of memory.
+ * NGHTTP2_ERR_INVALID_ARGUMENT
+ *     The |reason| is too long.
+ */
+int nghttp2_session_terminate_session_with_reason
+(nghttp2_session *session, uint32_t error_code, const char *reason);
+
 #endif /* NGHTTP2_SESSION_H */
