@@ -43,7 +43,7 @@
 
 #include <event.h>
 #include <event2/listener.h>
-#include <event2/bufferevent.h>
+#include <event2/bufferevent_ssl.h>
 
 #ifdef  __cplusplus
 extern "C" {
@@ -58,6 +58,8 @@ extern "C" {
 #include "app_helper.h"
 #include "http2.h"
 #include "util.h"
+#include "libevent_util.h"
+#include "ssl.h"
 
 #ifndef O_BINARY
 # define O_BINARY (0)
@@ -93,9 +95,9 @@ namespace {
 void append_nv(Stream *stream, const std::vector<nghttp2_nv>& nva)
 {
   for(auto& nv : nva) {
-    http2::split_add_header(stream->headers,
-                            nv.name, nv.namelen, nv.value, nv.valuelen,
-                            nv.flags & NGHTTP2_NV_FLAG_NO_INDEX);
+    http2::add_header(stream->headers,
+                      nv.name, nv.namelen, nv.value, nv.valuelen,
+                      nv.flags & NGHTTP2_NV_FLAG_NO_INDEX);
   }
 }
 } // namespace
@@ -103,6 +105,7 @@ void append_nv(Stream *stream, const std::vector<nghttp2_nv>& nva)
 Config::Config()
   : stream_read_timeout{60, 0},
     stream_write_timeout{60, 0},
+    session_option(nullptr),
     data_ptr(nullptr),
     padding(0),
     num_worker(1),
@@ -112,16 +115,24 @@ Config::Config()
     daemon(false),
     verify_client(false),
     no_tls(false),
-    error_gzip(false)
-{}
+    error_gzip(false),
+    early_response(false)
+{
+  nghttp2_option_new(&session_option);
+  nghttp2_option_set_recv_client_preface(session_option, 1);
+}
+
+Config::~Config()
+{
+  nghttp2_option_del(session_option);
+}
 
 Stream::Stream(Http2Handler *handler, int32_t stream_id)
   : handler(handler),
     rtimer(nullptr),
     wtimer(nullptr),
     stream_id(stream_id),
-    file(-1),
-    enable_compression(false)
+    file(-1)
 {}
 
 Stream::~Stream()
@@ -212,19 +223,29 @@ void remove_stream_write_timeout(Stream *stream)
 }
 } // namespace
 
+namespace {
+void fill_callback(nghttp2_session_callbacks *callbacks, const Config *config);
+} // namespace
+
 class Sessions {
 public:
   Sessions(event_base *evbase, const Config *config, SSL_CTX *ssl_ctx)
     : evbase_(evbase),
       config_(config),
       ssl_ctx_(ssl_ctx),
+      callbacks_(nullptr),
       next_session_id_(1)
-  {}
+  {
+    nghttp2_session_callbacks_new(&callbacks_);
+
+    fill_callback(callbacks_, config_);
+  }
   ~Sessions()
   {
     for(auto handler : handlers_) {
       delete handler;
     }
+    nghttp2_session_callbacks_del(callbacks_);
   }
   void add_handler(Http2Handler *handler)
   {
@@ -268,6 +289,10 @@ public:
     }
     return session_id;
   }
+  const nghttp2_session_callbacks* get_callbacks() const
+  {
+    return callbacks_;
+  }
   void accept_connection(int fd)
   {
     int val = 1;
@@ -296,6 +321,7 @@ private:
   event_base *evbase_;
   const Config *config_;
   SSL_CTX *ssl_ctx_;
+  nghttp2_session_callbacks *callbacks_;
   int64_t next_session_id_;
 };
 
@@ -310,36 +336,20 @@ void on_session_closed(Http2Handler *hd, int64_t session_id)
 }
 } // namespace
 
-namespace {
-void fill_callback(nghttp2_session_callbacks& callbacks, const Config *config);
-} // namespace
-
 Http2Handler::Http2Handler(Sessions *sessions,
                            int fd, SSL *ssl, int64_t session_id)
   : session_id_(session_id),
     session_(nullptr),
     sessions_(sessions),
     ssl_(ssl),
-    rev_(nullptr),
-    wev_(nullptr),
+    bev_(nullptr),
     settings_timerev_(nullptr),
-    pending_data_(nullptr),
-    pending_datalen_(0),
-    left_connhd_len_(NGHTTP2_CLIENT_CONNECTION_PREFACE_LEN),
     fd_(fd)
-{
-  nghttp2_buf_wrap_init(&sendbuf_, sendbufarray_, sizeof(sendbufarray_));
-}
+{}
 
 Http2Handler::~Http2Handler()
 {
   on_session_closed(this, session_id_);
-  if(rev_) {
-    event_free(rev_);
-  }
-  if(wev_) {
-    event_free(wev_);
-  }
   if(settings_timerev_) {
     event_free(settings_timerev_);
   }
@@ -348,6 +358,10 @@ Http2Handler::~Http2Handler()
     SSL_set_shutdown(ssl_, SSL_RECEIVED_SHUTDOWN);
     SSL_shutdown(ssl_);
   }
+  if(bev_) {
+    bufferevent_disable(bev_, EV_READ | EV_WRITE);
+    bufferevent_free(bev_);
+  }
   if(ssl_) {
     SSL_free(ssl_);
   }
@@ -361,325 +375,93 @@ void Http2Handler::remove_self()
 }
 
 namespace {
-void rev_cb(evutil_socket_t fd, short what, void *arg)
+void readcb(bufferevent *bev, void *arg)
 {
   int rv;
   auto handler = static_cast<Http2Handler*>(arg);
 
-  if(what & EV_READ) {
-    rv = handler->on_read();
-    if(rv == -1) {
-      delete_handler(handler);
-    }
+  rv = handler->on_read();
+  if(rv == -1) {
+    delete_handler(handler);
   }
 }
 } // namespace
 
 namespace {
-void wev_cb(evutil_socket_t fd, short what, void *arg)
+void writecb(bufferevent *bev, void *arg)
 {
   int rv;
   auto handler = static_cast<Http2Handler*>(arg);
 
-  if(what & EV_WRITE) {
-    rv = handler->on_write();
-    if(rv == -1) {
-      delete_handler(handler);
-    }
+  rv = handler->on_write();
+  if(rv == -1) {
+    delete_handler(handler);
   }
 }
 } // namespace
 
-int Http2Handler::handle_ssl_temporal_error(int err)
-{
-  auto sslerr = SSL_get_error(ssl_, err);
-
-  switch(sslerr) {
-  case SSL_ERROR_WANT_READ:
-    event_add(rev_, nullptr);
-    return 1;
-  case SSL_ERROR_WANT_WRITE:
-    event_add(wev_, nullptr);
-    return 1;
-  }
-
-  return -1;
-}
-
-int Http2Handler::tls_write(const uint8_t *data, size_t datalen)
-{
-  int rv;
-  size_t max_avail;
-
-  // OpenSSL sends at most 16K bytes
-  max_avail = ssl_ ?
-    std::min((ssize_t)16384, nghttp2_buf_avail(&sendbuf_)) :
-    nghttp2_buf_avail(&sendbuf_);
-
-  if(max_avail < datalen) {
-    if(nghttp2_buf_len(&sendbuf_) > 0) {
-      rv = tls_write_pending();
-
-      if(rv == -1) {
-        return -1;
-      }
-
-      if(rv == 1) {
-        pending_data_ = data;
-        pending_datalen_ = datalen;
-
-        return 1;
-      }
-    }
-
-    assert(nghttp2_buf_avail(&sendbuf_) >= (ssize_t)datalen);
-  }
-  //std::cerr << "DBG: copy " << datalen << " bytes" << std::endl;
-  sendbuf_.last = nghttp2_cpymem(sendbuf_.last, data, datalen);
-
-  return 0;
-}
-
-int Http2Handler::tls_write_pending()
-{
-  int rv;
-
-  if(nghttp2_buf_len(&sendbuf_) == 0) {
-    return 0;
-  }
-
-  for(;;) {
-    if(ssl_) {
-      ERR_clear_error();
-
-      rv = SSL_write(ssl_, sendbuf_.pos, nghttp2_buf_len(&sendbuf_));
-
-      if(rv == 0) {
-        return -1;
-      }
-      if(rv < 0) {
-        return handle_ssl_temporal_error(rv);
-      }
-    } else {
-      while((rv = write(fd_, sendbuf_.pos, nghttp2_buf_len(&sendbuf_))) &&
-            rv == -1 && errno == EINTR);
-
-      if(rv == 0) {
-        continue;
-      }
-      if(rv < 0) {
-        if(errno == EAGAIN || errno == EWOULDBLOCK) {
-          event_add(wev_, nullptr);
-          return 1;
-        }
-        return -1;
-      }
-    }
-
-    sendbuf_.pos += rv;
-
-    if(nghttp2_buf_len(&sendbuf_) == 0) {
-      nghttp2_buf_reset(&sendbuf_);
-
-      if(pending_data_) {
-        assert(nghttp2_buf_avail(&sendbuf_) >= (ssize_t)pending_datalen_);
-        sendbuf_.last = nghttp2_cpymem(sendbuf_.last,
-                                      pending_data_, pending_datalen_);
-        pending_data_ = nullptr;
-        pending_datalen_ = 0;
-
-        continue;
-      }
-
-      return 0;
-    }
-  }
-}
-
 namespace {
-void tls_handshake_cb(evutil_socket_t fd, short what, void *arg)
+void eventcb(bufferevent *bev, short events, void *arg)
 {
-  int rv;
   auto handler = static_cast<Http2Handler*>(arg);
 
-  if(what & (EV_READ | EV_WRITE)) {
-    rv = handler->tls_handshake();
-    if(rv == -1) {
-      delete_handler(handler);
-      return;
+  if(events & (BEV_EVENT_EOF | BEV_EVENT_ERROR | BEV_EVENT_TIMEOUT)) {
+    delete_handler(handler);
+
+    return;
+  }
+
+  if(events & BEV_EVENT_CONNECTED) {
+    if(handler->get_sessions()->get_config()->verbose) {
+      std::cerr << "SSL/TLS handshake completed" << std::endl;
     }
-    if(rv == 1) {
+
+    if(handler->verify_npn_result() != 0) {
+      delete_handler(handler);
+
       return;
     }
 
-    rv = handler->on_connect();
-    if(rv != 0) {
+    if(handler->on_connect() != 0) {
       delete_handler(handler);
+
       return;
     }
   }
 }
 } // namespace
 
-int Http2Handler::tls_handshake()
-{
-  int rv;
-
-  ERR_clear_error();
-
-  rv = SSL_accept(ssl_);
-  if(rv == 0) {
-    return -1;
-  }
-  if(rv < 0) {
-    auto sslerr = SSL_get_error(ssl_, rv);
-
-    switch(sslerr) {
-    case SSL_ERROR_NONE:
-    case SSL_ERROR_WANT_X509_LOOKUP:
-    case SSL_ERROR_ZERO_RETURN:
-      break;
-    case SSL_ERROR_WANT_READ:
-      event_add(rev_, nullptr);
-      return 1;
-    case SSL_ERROR_WANT_WRITE:
-      event_add(wev_, nullptr);
-      return 1;
-    }
-  }
-
-  if(sessions_->get_config()->verbose) {
-    std::cerr << "SSL/TLS handshake completed" << std::endl;
-  }
-
-  if(verify_npn_result() != 0) {
-    return -1;
-  }
-
-  event_del(rev_);
-  event_del(wev_);
-
-  event_assign(rev_, sessions_->get_evbase(), fd_, EV_READ, rev_cb, this);
-  event_assign(wev_, sessions_->get_evbase(), fd_, EV_WRITE, wev_cb, this);
-
-  return 0;
-}
-
 int Http2Handler::setup_bev()
 {
+  auto evbase = sessions_->get_evbase();
+
   if(ssl_) {
-    rev_ = event_new(sessions_->get_evbase(), fd_, EV_READ, tls_handshake_cb,
-                     this);
-    wev_ = event_new(sessions_->get_evbase(), fd_, EV_WRITE, tls_handshake_cb,
-                     this);
+    bev_ = bufferevent_openssl_socket_new(evbase, fd_, ssl_,
+                                          BUFFEREVENT_SSL_ACCEPTING,
+                                          BEV_OPT_DEFER_CALLBACKS);
   } else {
-    rev_ = event_new(sessions_->get_evbase(), fd_, EV_READ, rev_cb, this);
-    wev_ = event_new(sessions_->get_evbase(), fd_, EV_WRITE, wev_cb, this);
+    bev_ = bufferevent_socket_new(evbase, fd_, BEV_OPT_DEFER_CALLBACKS);
   }
 
-  event_add(rev_, nullptr);
-  // TODO set up timeout here
+  bufferevent_enable(bev_, EV_READ);
+  bufferevent_setcb(bev_, readcb, writecb, eventcb, this);
 
   return 0;
 }
 
-int Http2Handler::wait_events()
-{
-  int active = 0;
-
-  if(nghttp2_session_want_read(session_)) {
-    event_add(rev_, nullptr);
-    active = 1;
-  }
-
-  if(nghttp2_session_want_write(session_)) {
-    event_add(wev_, nullptr);
-    active = 1;
-  }
-
-  return active ? 0 : -1;
-}
-
-int Http2Handler::on_read()
+int Http2Handler::send()
 {
   int rv;
   uint8_t buf[16384];
-  uint8_t *bufp;
-  size_t nread;
-
-  if(ssl_) {
-    ERR_clear_error();
-    rv = SSL_read(ssl_, buf, sizeof(buf));
-
-    if(rv == 0) {
-      return -1;
-    }
-    if(rv < 0) {
-      return handle_ssl_temporal_error(rv);
-    }
-  } else {
-    while((rv = read(fd_, buf, sizeof(buf))) && rv == -1 && errno == EINTR);
-
-    if(rv == 0) {
-      return -1;
-    }
-    if(rv < 0) {
-      if(errno == EAGAIN || errno == EWOULDBLOCK) {
-        event_add(rev_, nullptr);
-        return 1;
-      }
-      return -1;
-    }
-  }
-
-  nread = rv;
-  bufp = buf;
-
-  if(left_connhd_len_ > 0) {
-    auto len = std::min(left_connhd_len_, nread);
-    const char *conhead = NGHTTP2_CLIENT_CONNECTION_PREFACE;
-
-    if(memcmp(conhead + NGHTTP2_CLIENT_CONNECTION_PREFACE_LEN -
-              left_connhd_len_, bufp, len) != 0) {
-      return -1;
-    }
-
-    left_connhd_len_ -= len;
-    nread -= len;
-
-    if(nread == 0) {
-      wait_events();
-
-      return 0;
-    }
-
-    bufp += len;
-  }
-
-  rv = nghttp2_session_mem_recv(session_, bufp, nread);
-  if(rv < 0) {
-    std::cerr << "nghttp2_session_mem_recv() returned error: "
-              << nghttp2_strerror(rv) << std::endl;
-    return -1;
-  }
-
-  return wait_events();
-}
-
-int Http2Handler::on_write()
-{
-  int rv;
-
-  //std::cerr << "DBG: on_write" << std::endl;
-
-  rv = tls_write_pending();
-  if(rv != 0) {
-    return rv;
-  }
-
+  auto output = bufferevent_get_output(bev_);
+  util::EvbufferBuffer evbbuf(output, buf, sizeof(buf));
   for(;;) {
-    const uint8_t *data;
+    // Check buffer length and break if it is large enough.
+    if(evbuffer_get_length(output) + evbbuf.get_buflen() >= 65536) {
+      break;
+    }
 
+    const uint8_t *data;
     auto datalen = nghttp2_session_mem_send(session_, &data);
 
     if(datalen < 0) {
@@ -687,24 +469,65 @@ int Http2Handler::on_write()
                 << nghttp2_strerror(datalen) << std::endl;
       return -1;
     }
-
     if(datalen == 0) {
       break;
     }
-
-    rv = tls_write(data, datalen);
-
+    rv = evbbuf.add(data, datalen);
     if(rv != 0) {
-      return rv;
+      std::cerr << "evbuffer_add() failed" << std::endl;
+      return -1;
     }
   }
 
-  rv = tls_write_pending();
+  rv = evbbuf.flush();
   if(rv != 0) {
-    return rv;
+    std::cerr << "evbuffer_add() failed" << std::endl;
+    return -1;
   }
 
-  return wait_events();
+  if(nghttp2_session_want_read(session_) == 0 &&
+     nghttp2_session_want_write(session_) == 0 &&
+     evbuffer_get_length(output) == 0) {
+
+    return -1;
+  }
+
+  return 0;
+}
+
+int Http2Handler::on_read()
+{
+  int rv;
+
+  auto input = bufferevent_get_input(bev_);
+
+  for(;;) {
+    auto len = evbuffer_get_contiguous_space(input);
+
+    if(len == 0) {
+      break;
+    }
+
+    auto data = evbuffer_pullup(input, len);
+
+    rv = nghttp2_session_mem_recv(session_, data, len);
+    if(rv < 0) {
+      std::cerr << "nghttp2_session_mem_recv() returned error: "
+                << nghttp2_strerror(rv) << std::endl;
+      return -1;
+    }
+
+    if(evbuffer_drain(input, len) == -1) {
+      std::cerr << "evbuffer_drain() failed" << std::endl;
+    }
+  }
+
+  return send();
+}
+
+int Http2Handler::on_write()
+{
+  return send();
 }
 
 namespace {
@@ -719,19 +542,17 @@ void settings_timeout_cb(evutil_socket_t fd, short what, void *arg)
 int Http2Handler::on_connect()
 {
   int r;
-  nghttp2_session_callbacks callbacks;
-  fill_callback(callbacks, sessions_->get_config());
-  r = nghttp2_session_server_new(&session_, &callbacks, this);
+
+  r = nghttp2_session_server_new2(&session_, sessions_->get_callbacks(), this,
+                                  sessions_->get_config()->session_option);
   if(r != 0) {
     return r;
   }
   nghttp2_settings_entry entry[4];
-  size_t niv = 2;
+  size_t niv = 1;
 
   entry[0].settings_id = NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS;
   entry[0].value = 100;
-  entry[1].settings_id = NGHTTP2_SETTINGS_COMPRESS_DATA;
-  entry[1].value = 1;
 
   if(sessions_->get_config()->header_table_size >= 0) {
     entry[niv].settings_id = NGHTTP2_SETTINGS_HEADER_TABLE_SIZE;
@@ -840,6 +661,16 @@ int Http2Handler::submit_response(const std::string& status,
                                  data_prd);
 }
 
+int Http2Handler::submit_non_final_response(const std::string& status,
+                                            int32_t stream_id)
+{
+  auto nva = std::vector<nghttp2_nv>{
+    http2::make_nv_ls(":status", status)
+  };
+  return nghttp2_submit_headers(session_, NGHTTP2_FLAG_NONE, stream_id,
+                                nullptr, nva.data(), nva.size(), nullptr);
+}
+
 int Http2Handler::submit_push_promise(Stream *stream,
                                       const std::string& push_path)
 {
@@ -874,14 +705,13 @@ int Http2Handler::submit_push_promise(Stream *stream,
 
   auto promised_stream = util::make_unique<Stream>(this, promised_stream_id);
 
-  append_nv(promised_stream.get(), http2::sort_nva(nva.data(), nva.size()));
+  append_nv(promised_stream.get(), nva);
   add_stream(promised_stream_id, std::move(promised_stream));
 
   return 0;
 }
 
-int Http2Handler::submit_rst_stream(Stream *stream,
-                                    nghttp2_error_code error_code)
+int Http2Handler::submit_rst_stream(Stream *stream, uint32_t error_code)
 {
   remove_stream_read_timeout(stream);
   remove_stream_write_timeout(stream);
@@ -925,16 +755,6 @@ const Config* Http2Handler::get_config() const
   return sessions_->get_config();
 }
 
-size_t Http2Handler::get_left_connhd_len() const
-{
-  return left_connhd_len_;
-}
-
-void Http2Handler::set_left_connhd_len(size_t left)
-{
-  left_connhd_len_ = left;
-}
-
 void Http2Handler::remove_settings_timer()
 {
   if(settings_timerev_) {
@@ -944,24 +764,11 @@ void Http2Handler::remove_settings_timer()
   }
 }
 
-void Http2Handler::terminate_session(nghttp2_error_code error_code)
+void Http2Handler::terminate_session(uint32_t error_code)
 {
   nghttp2_session_terminate_session(session_, error_code);
 }
 
-void Http2Handler::decide_compression(const std::string& path, Stream *stream)
-{
-  if(nghttp2_session_get_remote_settings
-     (session_, NGHTTP2_SETTINGS_COMPRESS_DATA) == 1 &&
-     (util::endsWith(path, ".html") ||
-      util::endsWith(path, ".js") ||
-      util::endsWith(path, ".css") ||
-      util::endsWith(path, ".txt"))) {
-
-    stream->enable_compression = true;
-  }
-}
-
 ssize_t file_read_callback
 (nghttp2_session *session, int32_t stream_id,
  uint8_t *buf, size_t length, uint32_t *data_flags,
@@ -972,65 +779,30 @@ ssize_t file_read_callback
 
   int fd = source->fd;
   ssize_t nread;
-  ssize_t rv;
 
-  // Compressing too small data is not efficient?
-  if(length >= 1024 && stream && stream->enable_compression) {
-    uint8_t srcbuf[4096];
-    auto maxread = std::min(length, sizeof(srcbuf));
+  while((nread = read(fd, buf, length)) == -1 && errno == EINTR);
 
-    while((nread = read(fd, srcbuf, maxread)) == -1 && errno == EINTR);
-    if(nread == -1) {
-      if(stream) {
-        remove_stream_read_timeout(stream);
-        remove_stream_write_timeout(stream);
-      }
+  if(nread == -1) {
+    remove_stream_read_timeout(stream);
+    remove_stream_write_timeout(stream);
 
-      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
-    }
-
-    if(nread > 0) {
-      rv = deflate_data(buf, length, srcbuf, nread);
-
-      if(rv < 0) {
-        memcpy(buf, srcbuf, nread);
-      } else {
-        nread = rv;
-
-        *data_flags |= NGHTTP2_DATA_FLAG_COMPRESSED;
-      }
-    }
-  } else {
-    while((nread = read(fd, buf, length)) == -1 && errno == EINTR);
-    if(nread == -1) {
-      if(stream) {
-        remove_stream_read_timeout(stream);
-        remove_stream_write_timeout(stream);
-      }
-
-      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
-    }
+    return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
   }
 
   if(nread == 0) {
     *data_flags |= NGHTTP2_DATA_FLAG_EOF;
+
+    if(nghttp2_session_get_stream_remote_close(session, stream_id) == 0) {
+      remove_stream_read_timeout(stream);
+      remove_stream_write_timeout(stream);
+
+      hd->submit_rst_stream(stream, NGHTTP2_NO_ERROR);
+    }
   }
 
   return nread;
 }
 
-namespace {
-bool check_url(const std::string& url)
-{
-  // We don't like '\' in url.
-  return !url.empty() && url[0] == '/' &&
-    url.find('\\') == std::string::npos &&
-    url.find("/../") == std::string::npos &&
-    url.find("/./") == std::string::npos &&
-    !util::endsWith(url, "/..") && !util::endsWith(url, "/.");
-}
-} // namespace
-
 namespace {
 void prepare_status_response(Stream *stream, Http2Handler *hd,
                              const std::string& status)
@@ -1060,7 +832,11 @@ void prepare_status_response(Stream *stream, Http2Handler *hd,
     gzclose(write_fd);
     headers.emplace_back("content-encoding", "gzip");
   } else {
-    auto rv = write(pipefd[1], body.c_str(), body.size());
+    ssize_t rv;
+
+    while((rv = write(pipefd[1], body.c_str(), body.size())) == -1 &&
+          errno == EINTR);
+
     if(rv != static_cast<ssize_t>(body.size())) {
       std::cerr << "Could not write all response body: " << rv << std::endl;
     }
@@ -1104,7 +880,7 @@ void prepare_response(Stream *stream, Http2Handler *hd, bool allow_push = true)
     url = url.substr(0, query_pos);
   }
   url = util::percentDecode(url.begin(), url.end());
-  if(!check_url(url)) {
+  if(!util::check_path(url)) {
     prepare_status_response(stream, hd, STATUS_404);
     return;
   }
@@ -1125,26 +901,34 @@ void prepare_response(Stream *stream, Http2Handler *hd, bool allow_push = true)
   int file = open(path.c_str(), O_RDONLY | O_BINARY);
   if(file == -1) {
     prepare_status_response(stream, hd, STATUS_404);
-  } else {
-    struct stat buf;
-    if(fstat(file, &buf) == -1) {
-      close(file);
-      prepare_status_response(stream, hd, STATUS_404);
-    } else {
-      stream->file = file;
-      nghttp2_data_provider data_prd;
-      data_prd.source.fd = file;
-      data_prd.read_callback = file_read_callback;
-      if(last_mod_found && buf.st_mtime <= last_mod) {
-        prepare_status_response(stream, hd, STATUS_304);
-      } else {
-        hd->decide_compression(path, stream);
 
-        hd->submit_file_response(STATUS_200, stream, buf.st_mtime,
-                                 buf.st_size, &data_prd);
-      }
-    }
+    return;
   }
+
+  struct stat buf;
+
+  if(fstat(file, &buf) == -1) {
+    close(file);
+    prepare_status_response(stream, hd, STATUS_404);
+
+    return;
+  }
+
+  stream->file = file;
+
+  nghttp2_data_provider data_prd;
+
+  data_prd.source.fd = file;
+  data_prd.read_callback = file_read_callback;
+
+  if(last_mod_found && buf.st_mtime <= last_mod) {
+    prepare_status_response(stream, hd, STATUS_304);
+
+    return;
+  }
+
+  hd->submit_file_response(STATUS_200, stream, buf.st_mtime,
+                           buf.st_size, &data_prd);
 }
 } // namespace
 
@@ -1179,8 +963,20 @@ int on_header_callback(nghttp2_session *session,
   if(!http2::check_nv(name, namelen, value, valuelen)) {
     return 0;
   }
-  http2::split_add_header(stream->headers, name, namelen, value, valuelen,
-                          flags & NGHTTP2_NV_FLAG_NO_INDEX);
+
+  if(namelen > 0 && name[0] == ':') {
+    if((!stream->headers.empty() &&
+        stream->headers.back().name.c_str()[0] != ':') ||
+     !http2::check_http2_request_pseudo_header(name, namelen)) {
+
+      nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE, frame->hd.stream_id,
+                                NGHTTP2_PROTOCOL_ERROR);
+      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
+    }
+  }
+
+  http2::add_header(stream->headers, name, namelen, value, valuelen,
+                    flags & NGHTTP2_NV_FLAG_NO_INDEX);
   return 0;
 }
 } // namespace
@@ -1250,8 +1046,9 @@ int hd_on_frame_recv_callback
 
     if(frame->hd.flags & NGHTTP2_FLAG_END_STREAM) {
       remove_stream_read_timeout(stream);
-
-      prepare_response(stream, hd);
+      if(!hd->get_config()->early_response) {
+        prepare_response(stream, hd);
+      }
     } else {
       add_stream_read_timeout(stream);
     }
@@ -1267,7 +1064,7 @@ int hd_on_frame_recv_callback
     if(frame->headers.cat == NGHTTP2_HCAT_REQUEST) {
 
       http2::normalize_headers(stream->headers);
-      if(!http2::check_http2_headers(stream->headers)) {
+      if(!http2::check_http2_request_headers(stream->headers)) {
         hd->submit_rst_stream(stream, NGHTTP2_PROTOCOL_ERROR);
         return 0;
       }
@@ -1285,12 +1082,23 @@ int hd_on_frame_recv_callback
         hd->submit_rst_stream(stream, NGHTTP2_PROTOCOL_ERROR);
         return 0;
       }
+
+      auto expect100 = http2::get_header(stream->headers, "expect");
+
+      if(expect100 && util::strieq("100-continue", expect100->value.c_str())) {
+        hd->submit_non_final_response("100", frame->hd.stream_id);
+      }
+
+      if(hd->get_config()->early_response) {
+        prepare_response(stream, hd);
+      }
     }
 
     if(frame->hd.flags & NGHTTP2_FLAG_END_STREAM) {
       remove_stream_read_timeout(stream);
-
-      prepare_response(stream, hd);
+      if(!hd->get_config()->early_response) {
+        prepare_response(stream, hd);
+      }
     } else {
       add_stream_read_timeout(stream);
     }
@@ -1337,8 +1145,9 @@ int hd_on_frame_send_callback
 
     if(frame->hd.flags & NGHTTP2_FLAG_END_STREAM) {
       remove_stream_write_timeout(stream);
-    } else if(nghttp2_session_get_stream_remote_window_size
-              (session, frame->hd.stream_id) == 0) {
+    } else if(std::min(nghttp2_session_get_stream_remote_window_size
+                       (session, frame->hd.stream_id),
+                       nghttp2_session_get_remote_window_size(session)) <= 0) {
       // If stream is blocked by flow control, enable write timeout.
       add_stream_read_timeout_if_pending(stream);
       add_stream_write_timeout(stream);
@@ -1407,7 +1216,7 @@ int on_data_chunk_recv_callback
 
 namespace {
 int on_stream_close_callback
-(nghttp2_session *session, int32_t stream_id, nghttp2_error_code error_code,
+(nghttp2_session *session, int32_t stream_id, uint32_t error_code,
  void *user_data)
 {
   auto hd = static_cast<Http2Handler*>(user_data);
@@ -1423,23 +1232,34 @@ int on_stream_close_callback
 } // namespace
 
 namespace {
-void fill_callback(nghttp2_session_callbacks& callbacks, const Config *config)
+void fill_callback(nghttp2_session_callbacks *callbacks, const Config *config)
 {
-  memset(&callbacks, 0, sizeof(nghttp2_session_callbacks));
-  callbacks.on_stream_close_callback = on_stream_close_callback;
-  callbacks.on_frame_recv_callback = hd_on_frame_recv_callback;
-  callbacks.on_frame_send_callback = hd_on_frame_send_callback;
+  nghttp2_session_callbacks_set_on_stream_close_callback
+    (callbacks, on_stream_close_callback);
+
+  nghttp2_session_callbacks_set_on_frame_recv_callback
+    (callbacks, hd_on_frame_recv_callback);
+
+  nghttp2_session_callbacks_set_on_frame_send_callback
+    (callbacks, hd_on_frame_send_callback);
+
   if(config->verbose) {
-    callbacks.on_invalid_frame_recv_callback =
-      verbose_on_invalid_frame_recv_callback;
-    callbacks.on_unknown_frame_recv_callback =
-      verbose_on_unknown_frame_recv_callback;
+    nghttp2_session_callbacks_set_on_invalid_frame_recv_callback
+      (callbacks, verbose_on_invalid_frame_recv_callback);
   }
-  callbacks.on_data_chunk_recv_callback = on_data_chunk_recv_callback;
-  callbacks.on_header_callback = on_header_callback;
-  callbacks.on_begin_headers_callback = on_begin_headers_callback;
+
+  nghttp2_session_callbacks_set_on_data_chunk_recv_callback
+    (callbacks, on_data_chunk_recv_callback);
+
+  nghttp2_session_callbacks_set_on_header_callback
+    (callbacks, on_header_callback);
+
+  nghttp2_session_callbacks_set_on_begin_headers_callback
+    (callbacks, on_begin_headers_callback);
+
   if(config->padding) {
-    callbacks.select_padding_callback = select_padding_callback;
+    nghttp2_session_callbacks_set_select_padding_callback
+      (callbacks, select_padding_callback);
   }
 }
 } // namespace
@@ -1455,7 +1275,9 @@ void worker_readcb(bufferevent *bev, void *arg)
   auto input = bufferevent_get_input(bev);
   while(evbuffer_get_length(input) >= sizeof(ClientInfo)) {
     ClientInfo client;
-    evbuffer_remove(input, &client, sizeof(client));
+    if(evbuffer_remove(input, &client, sizeof(client)) == -1) {
+      std::cerr << "evbuffer_remove() failed" << std::endl;
+    }
     sessions->accept_connection(client.fd);
   }
 }
@@ -1700,7 +1522,7 @@ int HttpServer::run()
     SSL_CTX_set_mode(ssl_ctx, SSL_MODE_RELEASE_BUFFERS);
     SSL_CTX_set_mode(ssl_ctx, SSL_MODE_ENABLE_PARTIAL_WRITE);
 
-    SSL_CTX_set_cipher_list(ssl_ctx, "HIGH:!aNULL:!MD5");
+    SSL_CTX_set_cipher_list(ssl_ctx, ssl::DEFAULT_CIPHER_LIST);
 
 
     const unsigned char sid_ctx[] = "nghttpd";
@@ -1709,9 +1531,12 @@ int HttpServer::run()
 
 #ifndef OPENSSL_NO_EC
 
-#if OPENSSL_VERSION_NUMBER >= 0x10002000L
-    SSL_CTX_set_ecdh_auto(ssl_ctx, 1);
-#else // OPENSSL_VERSION_NUBMER < 0x10002000L
+  // Disabled SSL_CTX_set_ecdh_auto, because computational cost of
+  // chosen curve is much higher than P-256.
+
+// #if OPENSSL_VERSION_NUMBER >= 0x10002000L
+//     SSL_CTX_set_ecdh_auto(ssl_ctx, 1);
+// #else // OPENSSL_VERSION_NUBMER < 0x10002000L
     // Use P-256, which is sufficiently secure at the time of this
     // writing.
     auto ecdh = EC_KEY_new_by_curve_name(NID_X9_62_prime256v1);
@@ -1722,10 +1547,32 @@ int HttpServer::run()
     }
     SSL_CTX_set_tmp_ecdh(ssl_ctx, ecdh);
     EC_KEY_free(ecdh);
-#endif // OPENSSL_VERSION_NUBMER < 0x10002000L
+// #endif // OPENSSL_VERSION_NUBMER < 0x10002000L
 
 #endif // OPENSSL_NO_EC
 
+    if(!config_->dh_param_file.empty()) {
+      // Read DH parameters from file
+      auto bio = BIO_new_file(config_->dh_param_file.c_str(), "r");
+      if(bio == nullptr) {
+        std::cerr << "BIO_new_file() failed: "
+                  << ERR_error_string(ERR_get_error(), nullptr) << std::endl;
+        return -1;
+      }
+
+      auto dh = PEM_read_bio_DHparams(bio, nullptr, nullptr, nullptr);
+
+      if(dh == nullptr) {
+        std::cerr << "PEM_read_bio_DHparams() failed: "
+                  << ERR_error_string(ERR_get_error(), nullptr) << std::endl;
+        return -1;
+      }
+
+      SSL_CTX_set_tmp_dh(ssl_ctx, dh);
+      DH_free(dh);
+      BIO_free(bio);
+    }
+
     if(SSL_CTX_use_PrivateKey_file(ssl_ctx,
                                    config_->private_key_file.c_str(),
                                    SSL_FILETYPE_PEM) != 1) {
