@@ -52,7 +52,7 @@
 namespace nghttp2 {
 
 namespace {
-const char* strstatus(nghttp2_error_code error_code)
+const char* strstatus(uint32_t error_code)
 {
   switch(error_code) {
   case NGHTTP2_NO_ERROR:
@@ -99,8 +99,10 @@ const char* strsettingsid(int32_t id)
     return "SETTINGS_MAX_CONCURRENT_STREAMS";
   case NGHTTP2_SETTINGS_INITIAL_WINDOW_SIZE:
     return "SETTINGS_INITIAL_WINDOW_SIZE";
-  case NGHTTP2_SETTINGS_COMPRESS_DATA:
-    return "SETTINGS_COMPRESS_DATA";
+  case NGHTTP2_SETTINGS_MAX_FRAME_SIZE:
+    return "SETTINGS_MAX_FRAME_SIZE";
+  case NGHTTP2_SETTINGS_MAX_HEADER_LIST_SIZE:
+    return "SETTINGS_MAX_HEADER_LIST_SIZE";
   default:
     return "UNKNOWN";
   }
@@ -129,10 +131,8 @@ const char* strframetype(uint8_t type)
     return "GOAWAY";
   case NGHTTP2_WINDOW_UPDATE:
     return "WINDOW_UPDATE";
-  case NGHTTP2_ALTSVC:
+  case NGHTTP2_EXT_ALTSVC:
     return "ALTSVC";
-  case NGHTTP2_BLOCKED:
-    return "BLOCKED";
   default:
     return "UNKNOWN";
   }
@@ -191,10 +191,11 @@ void print_nv(nghttp2_nv *nv)
 namespace {
 void print_nv(nghttp2_nv *nva, size_t nvlen)
 {
-  for(auto& nv : http2::sort_nva(nva, nvlen)) {
+  auto end = nva + nvlen;
+  for(; nva != end; ++nva) {
     print_frame_attr_indent();
 
-    print_nv(&nv);
+    print_nv(nva);
   }
 }
 } // namelen
@@ -225,58 +226,28 @@ void print_flags(const nghttp2_frame_hd& hd)
     if(hd.flags & NGHTTP2_FLAG_END_STREAM) {
       s += "END_STREAM";
     }
-    if(hd.flags & NGHTTP2_FLAG_END_SEGMENT) {
+    if(hd.flags & NGHTTP2_FLAG_PADDED) {
       if(!s.empty()) {
         s += " | ";
       }
-      s += "END_SEGMENT";
-    }
-    if(hd.flags & NGHTTP2_FLAG_PAD_LOW) {
-      if(!s.empty()) {
-        s += " | ";
-      }
-      s += "PAD_LOW";
-    }
-    if(hd.flags & NGHTTP2_FLAG_PAD_HIGH) {
-      if(!s.empty()) {
-        s += " | ";
-      }
-      s += "PAD_HIGH";
-    }
-    if(hd.flags & NGHTTP2_FLAG_COMPRESSED) {
-      if(!s.empty()) {
-        s += " | ";
-      }
-      s += "COMPRESSED";
+      s += "PADDED";
     }
     break;
   case NGHTTP2_HEADERS:
     if(hd.flags & NGHTTP2_FLAG_END_STREAM) {
       s += "END_STREAM";
     }
-    if(hd.flags & NGHTTP2_FLAG_END_SEGMENT) {
-      if(!s.empty()) {
-        s += " | ";
-      }
-      s += "END_SEGMENT";
-    }
     if(hd.flags & NGHTTP2_FLAG_END_HEADERS) {
       if(!s.empty()) {
         s += " | ";
       }
       s += "END_HEADERS";
     }
-    if(hd.flags & NGHTTP2_FLAG_PAD_LOW) {
+    if(hd.flags & NGHTTP2_FLAG_PADDED) {
       if(!s.empty()) {
         s += " | ";
       }
-      s += "PAD_LOW";
-    }
-    if(hd.flags & NGHTTP2_FLAG_PAD_HIGH) {
-      if(!s.empty()) {
-        s += " | ";
-      }
-      s += "PAD_HIGH";
+      s += "PADDED";
     }
     if(hd.flags & NGHTTP2_FLAG_PRIORITY) {
       if(!s.empty()) {
@@ -297,17 +268,11 @@ void print_flags(const nghttp2_frame_hd& hd)
     if(hd.flags & NGHTTP2_FLAG_END_HEADERS) {
       s += "END_HEADERS";
     }
-    if(hd.flags & NGHTTP2_FLAG_PAD_LOW) {
+    if(hd.flags & NGHTTP2_FLAG_PADDED) {
       if(!s.empty()) {
         s += " | ";
       }
-      s += "PAD_LOW";
-    }
-    if(hd.flags & NGHTTP2_FLAG_PAD_HIGH) {
-      if(!s.empty()) {
-        s += " | ";
-      }
-      s += "PAD_HIGH";
+      s += "PADDED";
     }
     break;
   case NGHTTP2_PING:
@@ -332,25 +297,6 @@ const char* frame_name_ansi_esc(print_type ptype)
 }
 } // namespace
 
-namespace {
-std::string ascii_dump(const uint8_t *data, size_t len)
-{
-  std::string res;
-
-  for(size_t i = 0; i < len; ++i) {
-    auto c = data[i];
-
-    if(c >= 0x20 && c < 0x7f) {
-      res += c;
-    } else {
-      res += ".";
-    }
-  }
-
-  return res;
-}
-} // namespace
-
 namespace {
 void print_frame(print_type ptype, const nghttp2_frame *frame)
 {
@@ -409,7 +355,7 @@ void print_frame(print_type ptype, const nghttp2_frame *frame)
     break;
   case NGHTTP2_RST_STREAM:
     print_frame_attr_indent();
-    fprintf(outfile, "(error_code=%s(%u))\n",
+    fprintf(outfile, "(error_code=%s(0x%02x))\n",
             strstatus(frame->rst_stream.error_code),
             frame->rst_stream.error_code);
     break;
@@ -419,7 +365,7 @@ void print_frame(print_type ptype, const nghttp2_frame *frame)
             static_cast<unsigned long>(frame->settings.niv));
     for(size_t i = 0; i < frame->settings.niv; ++i) {
       print_frame_attr_indent();
-      fprintf(outfile, "[%s(%d):%u]\n",
+      fprintf(outfile, "[%s(0x%02x):%u]\n",
               strsettingsid(frame->settings.iv[i].settings_id),
               frame->settings.iv[i].settings_id,
               frame->settings.iv[i].value);
@@ -440,38 +386,48 @@ void print_frame(print_type ptype, const nghttp2_frame *frame)
   case NGHTTP2_GOAWAY:
     print_frame_attr_indent();
     fprintf(outfile,
-            "(last_stream_id=%d, error_code=%s(%u), opaque_data(%u)=[%s])\n",
+            "(last_stream_id=%d, error_code=%s(0x%02x), "
+            "opaque_data(%u)=[%s])\n",
             frame->goaway.last_stream_id,
             strstatus(frame->goaway.error_code),
             frame->goaway.error_code,
             static_cast<unsigned int>(frame->goaway.opaque_data_len),
-            ascii_dump(frame->goaway.opaque_data,
-                       frame->goaway.opaque_data_len).c_str());
+            util::ascii_dump(frame->goaway.opaque_data,
+                             frame->goaway.opaque_data_len).c_str());
     break;
   case NGHTTP2_WINDOW_UPDATE:
     print_frame_attr_indent();
     fprintf(outfile, "(window_size_increment=%d)\n",
             frame->window_update.window_size_increment);
     break;
-  case NGHTTP2_ALTSVC:
+  case NGHTTP2_EXT_ALTSVC: {
     print_frame_attr_indent();
+
+    auto altsvc = static_cast<const nghttp2_ext_altsvc*>(frame->ext.payload);
+
     fprintf(outfile, "(max-age=%u, port=%u, protocol_id=",
-            frame->altsvc.max_age, frame->altsvc.port);
-    if(frame->altsvc.protocol_id_len) {
-      fwrite(frame->altsvc.protocol_id, frame->altsvc.protocol_id_len, 1,
-             outfile);
+            altsvc->max_age, altsvc->port);
+
+    if(altsvc->protocol_id_len) {
+      fwrite(altsvc->protocol_id, altsvc->protocol_id_len, 1, outfile);
     }
+
     fprintf(outfile, ", host=");
-    if(frame->altsvc.host_len) {
-      fwrite(frame->altsvc.host, frame->altsvc.host_len, 1, outfile);
+
+    if(altsvc->host_len) {
+      fwrite(altsvc->host, altsvc->host_len, 1, outfile);
     }
+
     fprintf(outfile, ", origin=");
-    if(frame->altsvc.origin_len) {
-      fwrite(frame->altsvc.origin, frame->altsvc.origin_len, 1, outfile);
+
+    if(altsvc->origin_len) {
+      fwrite(altsvc->origin, altsvc->origin_len, 1, outfile);
     }
+
     fprintf(outfile, ")\n");
 
     break;
+  }
   default:
     break;
   }
@@ -485,18 +441,17 @@ int verbose_on_header_callback(nghttp2_session *session,
                                uint8_t flags,
                                void *user_data)
 {
-  nghttp2_nv nva = {
+  nghttp2_nv nv = {
     const_cast<uint8_t*>(name), const_cast<uint8_t*>(value),
     namelen, valuelen
   };
 
-  for(auto& nv : http2::sort_nva(&nva, 1)) {
-    print_timer();
-    fprintf(outfile, " (stream_id=%d, noind=%d) ", frame->hd.stream_id,
-            (flags & NGHTTP2_NV_FLAG_NO_INDEX) != 0);
+  print_timer();
+  fprintf(outfile, " recv (stream_id=%d, noind=%d) ", frame->hd.stream_id,
+          (flags & NGHTTP2_NV_FLAG_NO_INDEX) != 0);
 
-    print_nv(&nv);
-  }
+  print_nv(&nv);
+  fflush(outfile);
 
   return 0;
 }
@@ -513,7 +468,7 @@ int verbose_on_frame_recv_callback
 
 int verbose_on_invalid_frame_recv_callback
 (nghttp2_session *session, const nghttp2_frame *frame,
- nghttp2_error_code error_code, void *user_data)
+ uint32_t error_code, void *user_data)
 {
   print_timer();
   fprintf(outfile, " [INVALID; status=%s] recv ", strstatus(error_code));
@@ -522,33 +477,6 @@ int verbose_on_invalid_frame_recv_callback
   return 0;
 }
 
-namespace {
-void dump_header(const uint8_t *head, size_t headlen)
-{
-  size_t i;
-  print_frame_attr_indent();
-  fprintf(outfile, "Header dump: ");
-  for(i = 0; i < headlen; ++i) {
-    fprintf(outfile, "%02X ", head[i]);
-  }
-  fprintf(outfile, "\n");
-}
-} // namespace
-
-int verbose_on_unknown_frame_recv_callback(nghttp2_session *session,
-                                           const uint8_t *head,
-                                           size_t headlen,
-                                           const uint8_t *payload,
-                                           size_t payloadlen,
-                                           void *user_data)
-{
-  print_timer();
-  fprintf(outfile, " recv unknown frame\n");
-  dump_header(head, headlen);
-  fflush(outfile);
-  return 0;
-}
-
 int verbose_on_frame_send_callback
 (nghttp2_session *session, const nghttp2_frame *frame, void *user_data)
 {
@@ -559,6 +487,23 @@ int verbose_on_frame_send_callback
   return 0;
 }
 
+int verbose_on_data_chunk_recv_callback
+(nghttp2_session *session, uint8_t flags, int32_t stream_id,
+ const uint8_t *data, size_t len, void *user_data)
+{
+  print_timer();
+  auto srecv = nghttp2_session_get_stream_effective_recv_data_length
+    (session, stream_id);
+  auto crecv = nghttp2_session_get_effective_recv_data_length(session);
+
+  fprintf(outfile,
+          " recv (stream_id=%d, length=%zu, srecv=%d, crecv=%d) DATA\n",
+          stream_id, len, srecv, crecv);
+  fflush(outfile);
+
+  return 0;
+}
+
 namespace {
 std::chrono::steady_clock::time_point base_tv;
 } // namespace
