@@ -44,6 +44,7 @@
 #include <event2/bufferevent_ssl.h>
 
 #include <openssl/err.h>
+#include <openssl/conf.h>
 
 #include "http-parser/http_parser.h"
 
@@ -108,13 +109,16 @@ Stream::Stream()
   : status_success(-1)
 {}
 
-Client::Client(Worker *worker)
+Client::Client(Worker *worker, size_t req_todo)
   : worker(worker),
     ssl(nullptr),
     bev(nullptr),
     next_addr(config.addrs),
     reqidx(0),
-    state(CLIENT_IDLE)
+    state(CLIENT_IDLE),
+    req_todo(req_todo),
+    req_started(0),
+    req_done(0)
 {}
 
 Client::~Client()
@@ -158,12 +162,19 @@ int Client::connect()
   return 0;
 }
 
-void Client::disconnect()
+void Client::fail()
 {
   process_abandoned_streams();
+
   if(worker->stats.req_done == worker->stats.req_todo) {
     worker->schedule_terminate();
   }
+
+  disconnect();
+}
+
+void Client::disconnect()
+{
   int fd = -1;
   streams.clear();
   session.reset();
@@ -192,13 +203,16 @@ void Client::submit_request()
 {
   session->submit_request();
   ++worker->stats.req_started;
+  ++req_started;
 }
 
 void Client::process_abandoned_streams()
 {
-  worker->stats.req_failed += streams.size();
-  worker->stats.req_error += streams.size();
-  worker->stats.req_done += streams.size();
+  auto req_abandoned = req_todo - req_done;
+
+  worker->stats.req_failed += req_abandoned;
+  worker->stats.req_error += req_abandoned;
+  worker->stats.req_done += req_abandoned;
 }
 
 void Client::report_progress()
@@ -265,6 +279,7 @@ void Client::on_header(int32_t stream_id,
 void Client::on_stream_close(int32_t stream_id, bool success)
 {
   ++worker->stats.req_done;
+  ++req_done;
   if(success && streams[stream_id].status_success == 1) {
     ++worker->stats.req_success;
   } else {
@@ -277,7 +292,7 @@ void Client::on_stream_close(int32_t stream_id, bool success)
     return;
   }
 
-  if(worker->stats.req_started < worker->stats.req_todo) {
+  if(req_started < req_todo) {
     submit_request();
     return;
   }
@@ -287,13 +302,13 @@ int Client::on_connect()
 {
   session->on_connect();
 
-  auto nreq = std::min(worker->stats.req_todo - worker->stats.req_started,
-                       std::min(worker->stats.req_todo / worker->clients.size(),
-                                (size_t)config.max_concurrent_streams));
+  auto nreq = std::min(req_todo - req_started,
+                       (size_t)config.max_concurrent_streams);
+
   for(; nreq > 0; --nreq) {
     submit_request();
   }
-  return 0;
+  return session->on_write();
 }
 
 int Client::on_read()
@@ -315,17 +330,29 @@ int Client::on_write()
 Worker::Worker(uint32_t id, SSL_CTX *ssl_ctx, size_t req_todo, size_t nclients,
                Config *config)
   : stats{0}, evbase(event_base_new()), ssl_ctx(ssl_ctx), config(config),
-    id(id), term_timer_started(false)
+    term_timer(nullptr), id(id)
 {
   stats.req_todo = req_todo;
   progress_interval = std::max((size_t)1, req_todo / 10);
+
+  auto nreqs_per_client = req_todo / nclients;
+  auto nreqs_rem = req_todo % nclients;
+
   for(size_t i = 0; i < nclients; ++i) {
-    clients.push_back(util::make_unique<Client>(this));
+    auto req_todo = nreqs_per_client;
+    if(nreqs_rem > 0) {
+      ++req_todo;
+      --nreqs_rem;
+    }
+    clients.push_back(util::make_unique<Client>(this, req_todo));
   }
 }
 
 Worker::~Worker()
 {
+  if(term_timer) {
+    event_free(term_timer);
+  }
   event_base_free(evbase);
 }
 
@@ -334,7 +361,7 @@ void Worker::run()
   for(auto& client : clients) {
     if(client->connect() != 0) {
       std::cerr << "client could not connect to host" << std::endl;
-      client->disconnect();
+      client->fail();
     }
   }
   event_base_loop(evbase, 0);
@@ -350,11 +377,11 @@ void term_timeout_cb(evutil_socket_t fd, short what, void *arg)
 
 void Worker::schedule_terminate()
 {
-  if(term_timer_started) {
+  if(term_timer) {
     return;
   }
-  term_timer_started = true;
-  auto term_timer = evtimer_new(evbase, term_timeout_cb, this);
+
+  term_timer = evtimer_new(evbase, term_timeout_cb, this);
   timeval timeout = { 0, 0 };
   evtimer_add(term_timer, &timeout);
 }
@@ -400,7 +427,7 @@ void eventcb(bufferevent *bev, short events, void *ptr)
 
       if(!next_proto) {
         debug_nextproto_error();
-        client->disconnect();
+        client->fail();
         return;
       }
 
@@ -416,12 +443,12 @@ void eventcb(bufferevent *bev, short events, void *ptr)
                                                            spdy_version);
         } else {
           debug_nextproto_error();
-          client->disconnect();
+          client->fail();
           return;
         }
 #else // !HAVE_SPDYLAY
         debug_nextproto_error();
-        client->disconnect();
+        client->fail();
         return;
 #endif // !HAVE_SPDYLAY
       }
@@ -458,7 +485,7 @@ void eventcb(bufferevent *bev, short events, void *ptr)
     return;
   }
   if(events & BEV_EVENT_EOF) {
-    client->disconnect();
+    client->fail();
     return;
   }
   if(events & (BEV_EVENT_ERROR | BEV_EVENT_TIMEOUT)) {
@@ -470,7 +497,7 @@ void eventcb(bufferevent *bev, short events, void *ptr)
       }
     }
     debug("error/eof\n");
-    client->disconnect();
+    client->fail();
     return;
   }
 }
@@ -483,7 +510,7 @@ void readcb(bufferevent *bev, void *ptr)
   auto client = static_cast<Client*>(ptr);
   rv = client->on_read();
   if(rv != 0) {
-    client->disconnect();
+    client->fail();
   }
 }
 } // namespace
@@ -498,7 +525,7 @@ void writecb(bufferevent *bev, void *ptr)
   auto client = static_cast<Client*>(ptr);
   rv = client->on_write();
   if(rv != 0) {
-    client->disconnect();
+    client->fail();
   }
 }
 } // namespace
@@ -550,15 +577,6 @@ std::string get_reqline(const char *uri, const http_parser_url& u)
 }
 } // namespace
 
-namespace {
-std::unique_ptr<Worker> run(std::unique_ptr<Worker> worker)
-{
-  worker->run();
-
-  return worker;
-}
-} // namespace
-
 namespace {
 int client_select_next_proto_cb(SSL* ssl,
                                 unsigned char **out, unsigned char *outlen,
@@ -788,11 +806,13 @@ int main(int argc, char **argv)
   memset(&act, 0, sizeof(struct sigaction));
   act.sa_handler = SIG_IGN;
   sigaction(SIGPIPE, &act, nullptr);
+  OPENSSL_config(nullptr);
+  OpenSSL_add_all_algorithms();
   SSL_load_error_strings();
   SSL_library_init();
 
 #ifndef NOTHREADS
-  ssl::LibsslGlobalLock();
+  ssl::LibsslGlobalLock lock;
 #endif // NOTHREADS
 
   auto ssl_ctx = SSL_CTX_new(SSLv23_client_method());
@@ -890,6 +910,10 @@ int main(int argc, char **argv)
 
   resolve_host();
 
+  if(config.nclients == 1) {
+    config.nthreads = 1;
+  }
+
   size_t nreqs_per_thread = config.nreqs / config.nthreads;
   ssize_t nreqs_rem = config.nreqs % config.nthreads;
 
@@ -898,10 +922,10 @@ int main(int argc, char **argv)
 
   std::cout << "starting benchmark..." << std::endl;
 
-  std::vector<std::future<std::unique_ptr<Worker>>> futures;
   auto start = std::chrono::steady_clock::now();
 
-  std::vector<std::unique_ptr<Worker>> workers;
+#ifndef NOTHREADS
+  std::vector<std::future<Stats>> futures;
   for(size_t i = 0; i < config.nthreads - 1; ++i) {
     auto nreqs = nreqs_per_thread + (nreqs_rem-- > 0);
     auto nclients = nclients_per_thread + (nclients_rem-- > 0);
@@ -911,8 +935,17 @@ int main(int argc, char **argv)
               << std::endl;
     auto worker = util::make_unique<Worker>(i, ssl_ctx, nreqs, nclients,
                                             &config);
-    futures.push_back(std::async(std::launch::async, run, std::move(worker)));
+    futures.push_back
+      (std::async(std::launch::async,
+                  [](std::unique_ptr<Worker> worker)
+                  {
+                    worker->run();
+
+                    return worker->stats;
+                  }, std::move(worker)));
   }
+#endif // NOTHREADS
+
   auto nreqs_last = nreqs_per_thread + (nreqs_rem-- > 0);
   auto nclients_last = nclients_per_thread + (nclients_rem-- > 0);
   std::cout << "spawning thread #" << (config.nthreads - 1) << ": "
@@ -923,9 +956,9 @@ int main(int argc, char **argv)
                 &config);
   worker.run();
 
+#ifndef NOTHREADS
   for(auto& fut : futures) {
-    auto subworker = fut.get();
-    auto& stats = subworker->stats;
+    auto stats = fut.get();
 
     worker.stats.req_todo += stats.req_todo;
     worker.stats.req_started += stats.req_started;
@@ -937,11 +970,11 @@ int main(int argc, char **argv)
     worker.stats.bytes_head += stats.bytes_head;
     worker.stats.bytes_body += stats.bytes_body;
 
-    for(size_t i = 0; i < sizeof(stats.status) / sizeof(stats.status[0]);
-        ++i) {
+    for(size_t i = 0; i < util::array_size(stats.status); ++i) {
       worker.stats.status[i] += stats.status[i];
     }
   }
+#endif // NOTHREADS
 
   auto end = std::chrono::steady_clock::now();
   auto duration =
@@ -995,6 +1028,9 @@ int main(int argc, char **argv)
             << worker.stats.bytes_head << " bytes headers, "
             << worker.stats.bytes_body << " bytes data"
             << std::endl;
+
+  SSL_CTX_free(ssl_ctx);
+
   return 0;
 }
 
