@@ -54,6 +54,7 @@
 
 #include <openssl/ssl.h>
 #include <openssl/err.h>
+#include <openssl/conf.h>
 
 #include <event.h>
 #include <event2/event.h>
@@ -66,6 +67,7 @@
 #include "app_helper.h"
 #include "HtmlParser.h"
 #include "util.h"
+#include "libevent_util.h"
 #include "base64.h"
 #include "http2.h"
 #include "nghttp2_gzip.h"
@@ -100,7 +102,8 @@ struct Config {
   bool stat;
   bool upgrade;
   bool continuation;
-  bool compress_data;
+  bool no_content_length;
+
   Config()
     : output_upper_thres(1024*1024),
       padding(0),
@@ -118,7 +121,7 @@ struct Config {
       stat(false),
       upgrade(false),
       continuation(false),
-      compress_data(false)
+      no_content_length(false)
   {
     nghttp2_option_new(&http2_option);
     nghttp2_option_set_peer_max_concurrent_streams
@@ -176,7 +179,6 @@ struct Request {
   Headers push_req_nva;
   // URI without fragment
   std::string uri;
-  std::string status;
   http_parser_url u;
   std::shared_ptr<Dependency> dep;
   nghttp2_priority_spec pri_spec;
@@ -187,10 +189,12 @@ struct Request {
   HtmlParser *html_parser;
   const nghttp2_data_provider *data_prd;
   int32_t stream_id;
+  int status;
   // Recursion level: 0: first entity, 1: entity linked from first entity
   int level;
   // RequestPriority value defined in HtmlParser.h
   int pri;
+  bool expect_final_response;
 
   // For pushed request, |uri| is empty and |u| is zero-cleared.
   Request(const std::string& uri, const http_parser_url &u,
@@ -207,8 +211,10 @@ struct Request {
       html_parser(nullptr),
       data_prd(data_prd),
       stream_id(-1),
+      status(0),
       level(level),
-      pri(pri)
+      pri(pri),
+      expect_final_response(false)
   {}
 
   ~Request()
@@ -314,6 +320,16 @@ struct Request {
     }
   }
 
+  bool response_pseudo_header_allowed() const
+  {
+    return res_nva.empty() || res_nva.back().name.c_str()[0] == ':';
+  }
+
+  bool push_request_pseudo_header_allowed() const
+  {
+    return res_nva.empty() || push_req_nva.back().name.c_str()[0] == ':';
+  }
+
   void record_request_time()
   {
     stat.stage = STAT_ON_REQUEST;
@@ -347,7 +363,7 @@ Config config;
 namespace {
 size_t populate_settings(nghttp2_settings_entry *iv)
 {
-  size_t niv = 3;
+  size_t niv = 2;
 
   iv[0].settings_id = NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS;
   iv[0].value = 100;
@@ -359,9 +375,6 @@ size_t populate_settings(nghttp2_settings_entry *iv)
     iv[1].value = NGHTTP2_INITIAL_WINDOW_SIZE;
   }
 
-  iv[2].settings_id = NGHTTP2_SETTINGS_COMPRESS_DATA;
-  iv[2].value = 1;
-
   if(config.header_table_size >= 0) {
     iv[niv].settings_id = NGHTTP2_SETTINGS_HEADER_TABLE_SIZE;
     iv[niv].value = config.header_table_size;
@@ -432,13 +445,11 @@ struct HttpClient {
   event *settings_timerev;
   addrinfo *addrs;
   addrinfo *next_addr;
-  nghttp2_gzip *inflater;
   // The number of completed requests, including failed ones.
   size_t complete;
   // The length of settings_payload
   size_t settings_payloadlen;
   client_state state;
-  int32_t last_inflate_error_stream_id;
   // The HTTP status code of the response message of HTTP Upgrade.
   unsigned int upgrade_response_status_code;
   // true if the response message of HTTP Upgrade request is fully
@@ -458,11 +469,9 @@ struct HttpClient {
       settings_timerev(nullptr),
       addrs(nullptr),
       next_addr(nullptr),
-      inflater(nullptr),
       complete(0),
       settings_payloadlen(0),
       state(STATE_IDLE),
-      last_inflate_error_stream_id(0),
       upgrade_response_status_code(0),
       upgrade_response_complete(false)
   {}
@@ -494,12 +503,12 @@ struct HttpClient {
     hints.ai_flags = AI_ADDRCONFIG;
     rv = getaddrinfo(host.c_str(), util::utos(port).c_str(), &hints, &addrs);
     if(rv != 0) {
-      std::cerr << "getaddrinfo() failed: "
+      std::cerr << "[ERROR] getaddrinfo() failed: "
                 << gai_strerror(rv) << std::endl;
       return -1;
     }
     if(addrs == nullptr) {
-      std::cerr << "No address returned" << std::endl;
+      std::cerr << "[ERROR] No address returned" << std::endl;
       return -1;
     }
     next_addr = addrs;
@@ -513,7 +522,7 @@ struct HttpClient {
       // We are establishing TLS connection.
       ssl = SSL_new(ssl_ctx);
       if(!ssl) {
-        std::cerr << "SSL_new() failed: "
+        std::cerr << "[ERROR] SSL_new() failed: "
                   << ERR_error_string(ERR_get_error(), nullptr) << std::endl;
         return -1;
       }
@@ -678,7 +687,7 @@ struct HttpClient {
       auto htperr = HTTP_PARSER_ERRNO(htp.get());
 
       if(htperr != HPE_OK) {
-        std::cerr << "Failed to parse HTTP Upgrade response header: "
+        std::cerr << "[ERROR] Failed to parse HTTP Upgrade response header: "
                   << "(" << http_errno_name(htperr) << ") "
                   << http_errno_description(htperr) << std::endl;
         return -1;
@@ -715,7 +724,7 @@ struct HttpClient {
           return 0;
         }
 
-        std::cerr << "HTTP Upgrade failed" << std::endl;
+        std::cerr << "[ERROR] HTTP Upgrade failed" << std::endl;
 
         return -1;
       }
@@ -745,7 +754,7 @@ struct HttpClient {
       rv = nghttp2_session_upgrade(session, settings_payload,
                                    settings_payloadlen, stream_user_data);
       if(rv != 0) {
-        std::cerr << "nghttp2_session_upgrade() returned error: "
+        std::cerr << "[ERROR] nghttp2_session_upgrade() returned error: "
                   << nghttp2_strerror(rv) << std::endl;
         return -1;
       }
@@ -813,7 +822,7 @@ struct HttpClient {
       rv = nghttp2_session_mem_recv(session, mem, inputlen);
 
       if(rv < 0) {
-        std::cerr << "nghttp2_session_mem_recv() returned error: "
+        std::cerr << "[ERROR] nghttp2_session_mem_recv() returned error: "
                   << nghttp2_strerror(rv) << std::endl;
         return -1;
       }
@@ -840,7 +849,7 @@ struct HttpClient {
       auto datalen = nghttp2_session_mem_send(session, &data);
 
       if(datalen < 0) {
-        std::cerr << "nghttp2_session_mem_send() returned error: "
+        std::cerr << "[ERROR] nghttp2_session_mem_send() returned error: "
                   << nghttp2_strerror(datalen) << std::endl;
         return -1;
       }
@@ -849,13 +858,13 @@ struct HttpClient {
       }
       rv = evbbuf.add(data, datalen);
       if(rv != 0) {
-        std::cerr << "evbuffer_add() failed" << std::endl;
+        std::cerr << "[ERROR] evbuffer_add() failed" << std::endl;
         return -1;
       }
     }
     rv = evbbuf.flush();
     if(rv != 0) {
-      std::cerr << "evbuffer_add() failed" << std::endl;
+      std::cerr << "[ERROR] evbuffer_add() failed" << std::endl;
       return -1;
     }
     if(nghttp2_session_want_read(session) == 0 &&
@@ -906,55 +915,23 @@ struct HttpClient {
     }
     if(path_cache.count(uri)) {
       return false;
-    } else {
-      if(config.multiply == 1) {
-        path_cache.insert(uri);
-      }
-
-      reqvec.push_back(util::make_unique<Request>(uri, u, data_prd,
-                                                  data_length,
-                                                  pri_spec, std::move(dep),
-                                                  pri, level));
-      return true;
     }
+
+    if(config.multiply == 1) {
+      path_cache.insert(uri);
+    }
+
+    reqvec.push_back(util::make_unique<Request>(uri, u, data_prd,
+                                                data_length,
+                                                pri_spec, std::move(dep),
+                                                pri, level));
+    return true;
   }
   void record_handshake_time()
   {
     stat.on_handshake_time = get_time();
   }
 
-  bool check_inflater(int32_t stream_id)
-  {
-    if(inflater == nullptr || last_inflate_error_stream_id == stream_id) {
-      return false;
-    }
-
-    last_inflate_error_stream_id = 0;
-
-    return true;
-  }
-
-  bool reset_inflater()
-  {
-    int rv;
-    nghttp2_gzip *gzip;
-
-    if(inflater) {
-      nghttp2_gzip_inflate_del(inflater);
-      inflater = nullptr;
-    }
-
-    rv = nghttp2_gzip_inflate_new(&gzip);
-
-    if(rv != 0) {
-      return false;
-    }
-
-    inflater = gzip;
-
-    return true;
-  }
-
   void on_request(Request *req)
   {
     req->record_request_time();
@@ -1053,13 +1030,13 @@ int submit_request
      {"accept-encoding", "gzip, deflate"},
      {"user-agent", "nghttp2/" NGHTTP2_VERSION}};
   if(config.continuation) {
-    for(size_t i = 0; i < 8; ++i) {
+    for(size_t i = 0; i < 6; ++i) {
       build_headers.emplace_back("continuation-test-" + util::utos(i+1),
                                  std::string(4096, '-'));
     }
   }
   auto num_initial_headers = build_headers.size();
-  if(req->data_prd) {
+  if(!config.no_content_length && req->data_prd) {
     build_headers.emplace_back("content-length", util::utos(req->data_length));
   }
   for(auto& kv : headers) {
@@ -1080,13 +1057,7 @@ int submit_request
     build_headers.emplace_back(kv.first, kv.second, no_index);
   }
   std::stable_sort(std::begin(build_headers), std::end(build_headers),
-                   [](const Headers::value_type& lhs,
-                      const Headers::value_type& rhs)
-                   {
-                     return lhs.name < rhs.name;
-                   });
-
-  build_headers = http2::concat_norm_headers(std::move(build_headers));
+                   http2::name_less);
 
   auto nva = std::vector<nghttp2_nv>();
   nva.reserve(build_headers.size());
@@ -1099,7 +1070,7 @@ int submit_request
                                           nva.data(), nva.size(),
                                           req->data_prd, req);
   if(stream_id < 0) {
-    std::cerr << "nghttp2_submit_request() returned error: "
+    std::cerr << "[ERROR] nghttp2_submit_request() returned error: "
               << nghttp2_strerror(stream_id) << std::endl;
     return -1;
   }
@@ -1165,6 +1136,17 @@ int on_data_chunk_recv_callback
     return 0;
   }
 
+  if(config.verbose) {
+    verbose_on_data_chunk_recv_callback(session, flags, stream_id, data, len,
+                                        user_data);
+  }
+
+  if(req->status == 0) {
+    nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE,
+                              stream_id, NGHTTP2_PROTOCOL_ERROR);
+    return 0;
+  }
+
   if(req->inflater) {
     while(len > 0) {
       const size_t MAX_OUTLEN = 4096;
@@ -1184,65 +1166,15 @@ int on_data_chunk_recv_callback
       data += tlen;
       len -= tlen;
     }
-  } else if(flags & NGHTTP2_FLAG_COMPRESSED) {
-    if(len == 0 || !client->check_inflater(stream_id)) {
-      return 0;
-    }
 
-    const size_t MAX_OUTLEN = 4096;
-    uint8_t out[MAX_OUTLEN];
-    size_t outlen;
-
-    do {
-      outlen = MAX_OUTLEN;
-      auto tlen = len;
-
-      int rv = nghttp2_gzip_inflate(client->inflater, out, &outlen,
-                                    data, &tlen);
-      if(rv != 0) {
-        goto per_frame_decomp_error;
-      }
-
-      if(!config.null_out) {
-        std::cout.write(reinterpret_cast<const char*>(out), outlen);
-      }
-
-      update_html_parser(client, req, out, outlen, 0);
-
-      data += tlen;
-      len -= tlen;
-
-      if(nghttp2_gzip_inflate_finished(client->inflater)) {
-        // When Z_STREAM_END was reached, remaining input length
-        // must be 0.
-        if(len > 0) {
-          goto per_frame_decomp_error;
-        }
-
-        break;
-      }
-    } while(len > 0 || outlen > 0);
-
-  } else {
-    if(!config.null_out) {
-      std::cout.write(reinterpret_cast<const char*>(data), len);
-    }
-    update_html_parser(client, req, data, len, 0);
+    return 0;
   }
 
-  return 0;
-
- per_frame_decomp_error:
-  // If per-frame decompression failed, we remember the stream ID so
-  // that subsequent chunk of DATA is ignored.
-  client->last_inflate_error_stream_id = stream_id;
-
-  if(!client->reset_inflater()) {
-    return NGHTTP2_ERR_CALLBACK_FAILURE;
+  if(!config.null_out) {
+    std::cout.write(reinterpret_cast<const char*>(data), len);
   }
 
-  nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE, stream_id,
-                            NGHTTP2_INTERNAL_ERROR);
+  update_html_parser(client, req, data, len, 0);
 
   return 0;
 }
@@ -1274,6 +1206,9 @@ namespace {
 void check_response_header(nghttp2_session *session, Request* req)
 {
   bool gzip = false;
+
+  req->expect_final_response = false;
+
   for(auto& nv : req->res_nva) {
     if("content-encoding" == nv.name) {
       gzip = util::strieq("gzip", nv.value) ||
@@ -1281,9 +1216,32 @@ void check_response_header(nghttp2_session *session, Request* req)
       continue;
     }
     if(":status" == nv.name) {
-      req->status.assign(nv.value);
+      int status;
+      if(req->status != 0 ||
+         (status = http2::parse_http_status_code(nv.value)) == -1) {
+
+        nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE,
+                                  req->stream_id, NGHTTP2_PROTOCOL_ERROR);
+        return;
+      }
+
+      req->status = status;
     }
   }
+
+  if(req->status == 0) {
+    nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE, req->stream_id,
+                              NGHTTP2_PROTOCOL_ERROR);
+    return;
+  }
+
+  if(req->status / 100 == 1) {
+    req->expect_final_response = true;
+    req->status = 0;
+    req->res_nva.clear();
+    return;
+  }
+
   if(gzip) {
     if(!req->inflater) {
       req->init_inflater();
@@ -1341,29 +1299,63 @@ int on_header_callback(nghttp2_session *session,
     verbose_on_header_callback(session, frame, name, namelen, value, valuelen,
                                flags, user_data);
   }
+
+  if(!http2::check_nv(name, namelen, value, valuelen)) {
+    nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE, frame->hd.stream_id,
+                              NGHTTP2_PROTOCOL_ERROR);
+    return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
+  }
+
   switch(frame->hd.type) {
   case NGHTTP2_HEADERS: {
-    if(frame->headers.cat != NGHTTP2_HCAT_RESPONSE &&
-       frame->headers.cat != NGHTTP2_HCAT_PUSH_RESPONSE) {
-      break;
-    }
     auto req = (Request*)nghttp2_session_get_stream_user_data
       (session, frame->hd.stream_id);
+
     if(!req) {
       break;
     }
-    http2::split_add_header(req->res_nva, name, namelen, value, valuelen,
-                            flags & NGHTTP2_NV_FLAG_NO_INDEX);
+
+    if(frame->headers.cat != NGHTTP2_HCAT_RESPONSE &&
+       frame->headers.cat != NGHTTP2_HCAT_PUSH_RESPONSE &&
+       (frame->headers.cat != NGHTTP2_HCAT_HEADERS ||
+        !req->expect_final_response)) {
+      break;
+    }
+
+    if(namelen > 0 && name[0] == ':') {
+      if(!req->response_pseudo_header_allowed() ||
+         !http2::check_http2_response_pseudo_header(name, namelen)) {
+        nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE,
+                                  frame->hd.stream_id,
+                                  NGHTTP2_PROTOCOL_ERROR);
+        return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
+      }
+    }
+
+    http2::add_header(req->res_nva, name, namelen, value, valuelen,
+                      flags & NGHTTP2_NV_FLAG_NO_INDEX);
     break;
   }
   case NGHTTP2_PUSH_PROMISE: {
     auto req = (Request*)nghttp2_session_get_stream_user_data
       (session, frame->push_promise.promised_stream_id);
+
     if(!req) {
       break;
     }
-    http2::split_add_header(req->push_req_nva, name, namelen, value, valuelen,
-                            flags & NGHTTP2_NV_FLAG_NO_INDEX);
+
+    if(namelen > 0 && name[0] == ':') {
+      if(!req->push_request_pseudo_header_allowed() ||
+         !http2::check_http2_request_pseudo_header(name, namelen)) {
+        nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE,
+                                  frame->push_promise.promised_stream_id,
+                                  NGHTTP2_PROTOCOL_ERROR);
+        break;
+      }
+    }
+
+    http2::add_header(req->push_req_nva, name, namelen, value, valuelen,
+                      flags & NGHTTP2_NV_FLAG_NO_INDEX);
     break;
   }
   }
@@ -1377,43 +1369,45 @@ int on_frame_recv_callback2
 {
   int rv = 0;
 
+  if(config.verbose) {
+    verbose_on_frame_recv_callback(session, frame, user_data);
+  }
+
   auto client = get_session(user_data);
   switch(frame->hd.type) {
-  case NGHTTP2_DATA:
-    if(frame->hd.flags & NGHTTP2_FLAG_COMPRESSED) {
-
-      auto inflate_finished = nghttp2_gzip_inflate_finished(client->inflater);
-
-      if(!client->reset_inflater()) {
-        rv = nghttp2_session_terminate_session(session,
-                                               NGHTTP2_INTERNAL_ERROR);
-
-        if(nghttp2_is_fatal(rv)) {
-          rv = NGHTTP2_ERR_CALLBACK_FAILURE;
-        } else {
-          rv = 0;
-        }
-      }
-      // Error if compressed block does not end in frame.
-      if(!inflate_finished) {
-        nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE,
-                                  frame->hd.stream_id, NGHTTP2_PROTOCOL_ERROR);
-      }
-    }
-    break;
   case NGHTTP2_HEADERS: {
-    if(frame->headers.cat != NGHTTP2_HCAT_RESPONSE &&
-       frame->headers.cat != NGHTTP2_HCAT_PUSH_RESPONSE) {
-      break;
-    }
     auto req = (Request*)nghttp2_session_get_stream_user_data
       (session, frame->hd.stream_id);
     // If this is the HTTP Upgrade with OPTIONS method to avoid POST,
     // req is nullptr.
-    if(req) {
+    if(!req) {
+      break;
+    }
+
+    if(frame->headers.cat == NGHTTP2_HCAT_RESPONSE ||
+       frame->headers.cat == NGHTTP2_HCAT_PUSH_RESPONSE) {
       req->record_response_time();
       check_response_header(session, req);
+
+      break;
     }
+
+    if(frame->headers.cat == NGHTTP2_HCAT_HEADERS) {
+      if(req->expect_final_response) {
+        check_response_header(session, req);
+      } else {
+        nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE,
+                                  frame->hd.stream_id, NGHTTP2_PROTOCOL_ERROR);
+        break;
+      }
+    }
+
+    if(req->status == 0 && (frame->hd.flags & NGHTTP2_FLAG_END_STREAM)) {
+      nghttp2_submit_rst_stream(session, NGHTTP2_FLAG_NONE,
+                                frame->hd.stream_id, NGHTTP2_PROTOCOL_ERROR);
+      break;
+    }
+
     break;
   }
   case NGHTTP2_SETTINGS:
@@ -1475,16 +1469,13 @@ int on_frame_recv_callback2
     break;
   }
   }
-  if(config.verbose) {
-    verbose_on_frame_recv_callback(session, frame, user_data);
-  }
   return rv;
 }
 } // namespace
 
 namespace {
 int on_stream_close_callback
-(nghttp2_session *session, int32_t stream_id, nghttp2_error_code error_code,
+(nghttp2_session *session, int32_t stream_id, uint32_t error_code,
  void *user_data)
 {
   auto client = get_session(user_data);
@@ -1545,7 +1536,7 @@ void print_stats(const HttpClient& client)
 namespace {
 void print_protocol_nego_error()
 {
-  std::cerr << "HTTP/2 protocol was not selected."
+  std::cerr << "[ERROR] HTTP/2 protocol was not selected."
             << " (nghttp2 expects " << NGHTTP2_PROTO_VERSION_ID << ")"
             << std::endl;
 }
@@ -1626,7 +1617,7 @@ void eventcb(bufferevent *bev, short events, void *ptr)
     int val = 1;
     if(setsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
                   reinterpret_cast<char *>(&val), sizeof(val)) == -1) {
-      std::cerr << "Setting option TCP_NODELAY failed: errno="
+      std::cerr << "[ERROR] Setting option TCP_NODELAY failed: errno="
                 << errno << std::endl;
     }
     if(client->need_upgrade()) {
@@ -1687,12 +1678,12 @@ void eventcb(bufferevent *bev, short events, void *ptr)
   if(events & (BEV_EVENT_ERROR | BEV_EVENT_TIMEOUT)) {
     if(events & BEV_EVENT_ERROR) {
       if(client->state == STATE_IDLE) {
-        std::cerr << "Could not connect to the host" << std::endl;
+        std::cerr << "[ERROR] Could not connect to the host" << std::endl;
       } else {
-        std::cerr << "Network error" << std::endl;
+        std::cerr << "[ERROR] Network error" << std::endl;
       }
     } else {
-      std::cerr << "Timeout" << std::endl;
+      std::cerr << "[ERROR] Timeout" << std::endl;
     }
     auto state = client->state;
     client->disconnect();
@@ -1720,7 +1711,7 @@ int communicate(const std::string& scheme, const std::string& host,
   if(scheme == "https") {
     ssl_ctx = SSL_CTX_new(SSLv23_client_method());
     if(!ssl_ctx) {
-      std::cerr << "Failed to create SSL_CTX: "
+      std::cerr << "[ERROR] Failed to create SSL_CTX: "
                 << ERR_error_string(ERR_get_error(), nullptr) << std::endl;
       result = -1;
       goto fin;
@@ -1734,7 +1725,8 @@ int communicate(const std::string& scheme, const std::string& host,
     if(!config.keyfile.empty()) {
       if(SSL_CTX_use_PrivateKey_file(ssl_ctx, config.keyfile.c_str(),
                                      SSL_FILETYPE_PEM) != 1) {
-        std::cerr << ERR_error_string(ERR_get_error(), nullptr) << std::endl;
+        std::cerr << "[ERROR] "
+                  << ERR_error_string(ERR_get_error(), nullptr) << std::endl;
         result = -1;
         goto fin;
       }
@@ -1742,7 +1734,8 @@ int communicate(const std::string& scheme, const std::string& host,
     if(!config.certfile.empty()) {
       if(SSL_CTX_use_certificate_chain_file(ssl_ctx,
                                             config.certfile.c_str()) != 1) {
-        std::cerr << ERR_error_string(ERR_get_error(), nullptr) << std::endl;
+        std::cerr << "[ERROR] "
+                  << ERR_error_string(ERR_get_error(), nullptr) << std::endl;
         result = -1;
         goto fin;
       }
@@ -1761,10 +1754,6 @@ int communicate(const std::string& scheme, const std::string& host,
   {
     HttpClient client{callbacks, evbase, ssl_ctx};
 
-    if(!client.reset_inflater()) {
-      goto fin;
-    }
-
     nghttp2_priority_spec pri_spec;
 
     if(config.weight != NGHTTP2_DEFAULT_WEIGHT) {
@@ -1820,37 +1809,12 @@ ssize_t file_read_callback
   assert(req);
   int fd = source->fd;
   ssize_t nread;
-  ssize_t rv;
 
-  // Compressing too small data is not efficient?
-  if(length >= 1024 && config.compress_data &&
-     nghttp2_session_get_remote_settings
-     (session, NGHTTP2_SETTINGS_COMPRESS_DATA) == 1) {
-    uint8_t srcbuf[4096];
-    auto maxread = std::min(length, sizeof(srcbuf));
+  while((nread = pread(fd, buf, length, req->data_offset)) == -1 &&
+        errno == EINTR);
 
-    while((nread = read(fd, srcbuf, maxread)) == -1 && errno == EINTR);
-    if(nread == -1) {
-      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
-    }
-
-    if(nread > 0) {
-      rv = deflate_data(buf, length, srcbuf, nread);
-
-      if(rv < 0) {
-        memcpy(buf, srcbuf, nread);
-      } else {
-        nread = rv;
-
-        *data_flags |= NGHTTP2_DATA_FLAG_COMPRESSED;
-      }
-    }
-  } else {
-    while((nread = pread(fd, buf, length, req->data_offset)) == -1 &&
-          errno == EINTR);
-    if(nread == -1) {
-      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
-    }
+  if(nread == -1) {
+    return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
   }
 
   if(nread == 0) {
@@ -1866,22 +1830,37 @@ ssize_t file_read_callback
 namespace {
 int run(char **uris, int n)
 {
-  nghttp2_session_callbacks callbacks;
-  memset(&callbacks, 0, sizeof(nghttp2_session_callbacks));
-  callbacks.on_stream_close_callback = on_stream_close_callback;
-  callbacks.on_frame_recv_callback = on_frame_recv_callback2;
+  nghttp2_session_callbacks *callbacks;
+
+  nghttp2_session_callbacks_new(&callbacks);
+  auto cbsdel = util::defer(callbacks, nghttp2_session_callbacks_del);
+
+  nghttp2_session_callbacks_set_on_stream_close_callback
+    (callbacks, on_stream_close_callback);
+
+  nghttp2_session_callbacks_set_on_frame_recv_callback
+    (callbacks, on_frame_recv_callback2);
+
   if(config.verbose) {
-    callbacks.on_frame_send_callback = verbose_on_frame_send_callback;
-    callbacks.on_invalid_frame_recv_callback =
-      verbose_on_invalid_frame_recv_callback;
-    callbacks.on_unknown_frame_recv_callback =
-      verbose_on_unknown_frame_recv_callback;
+    nghttp2_session_callbacks_set_on_frame_send_callback
+      (callbacks, verbose_on_frame_send_callback);
+
+    nghttp2_session_callbacks_set_on_invalid_frame_recv_callback
+      (callbacks, verbose_on_invalid_frame_recv_callback);
   }
-  callbacks.on_data_chunk_recv_callback = on_data_chunk_recv_callback;
-  callbacks.on_begin_headers_callback = on_begin_headers_callback;
-  callbacks.on_header_callback = on_header_callback;
+
+  nghttp2_session_callbacks_set_on_data_chunk_recv_callback
+    (callbacks, on_data_chunk_recv_callback);
+
+  nghttp2_session_callbacks_set_on_begin_headers_callback
+    (callbacks, on_begin_headers_callback);
+
+  nghttp2_session_callbacks_set_on_header_callback
+    (callbacks, on_header_callback);
+
   if(config.padding) {
-    callbacks.select_padding_callback = select_padding_callback;
+    nghttp2_session_callbacks_set_select_padding_callback
+      (callbacks, select_padding_callback);
   }
 
   std::string prev_scheme;
@@ -1895,12 +1874,14 @@ int run(char **uris, int n)
   if(!config.datafile.empty()) {
     data_fd = open(config.datafile.c_str(), O_RDONLY | O_BINARY);
     if(data_fd == -1) {
-      std::cerr << "Could not open file " << config.datafile << std::endl;
+      std::cerr << "[ERROR] Could not open file "
+                << config.datafile << std::endl;
       return 1;
     }
     if(fstat(data_fd, &data_stat) == -1) {
       close(data_fd);
-      std::cerr << "Could not stat file " << config.datafile << std::endl;
+      std::cerr << "[ERROR] Could not stat file "
+                << config.datafile << std::endl;
       return 1;
     }
     data_prd.source.fd = data_fd;
@@ -1921,7 +1902,7 @@ int run(char **uris, int n)
          port != prev_port) {
         if(!requests.empty()) {
           if (communicate(prev_scheme, prev_host, prev_port,
-                          std::move(requests), &callbacks) != 0) {
+                          std::move(requests), callbacks) != 0) {
             ++failures;
           }
           requests.clear();
@@ -1936,7 +1917,7 @@ int run(char **uris, int n)
   }
   if(!requests.empty()) {
     if (communicate(prev_scheme, prev_host, prev_port, std::move(requests),
-                    &callbacks) != 0) {
+                    callbacks) != 0) {
       ++failures;
     }
   }
@@ -1992,9 +1973,6 @@ Options:
                      be in PEM format.
   -d, --data=<FILE>  Post FILE to  server. If '-' is  given, data will
                      be read from stdin.
-  -g, --compress-data
-                     When used  with -d option, compress  request body
-                     on the fly using per-frame compression.
   -m, --multiply=<N> Request each URI <N> times.  By default, same URI
                      is not requested twice.   This option disables it
                      too.
@@ -2019,6 +1997,8 @@ Options:
                      padding.  Specify 0 to disable padding.
   --color            Force colored log output.
   --continuation     Send large header to test CONTINUATION.
+  --no-content-length
+                     Don't send content-length header field.
   --version          Display version information and exit.
   -h, --help         Display this help and exit.)"
       << std::endl;
@@ -2042,7 +2022,6 @@ int main(int argc, char **argv)
       {"help", no_argument, nullptr, 'h'},
       {"header", required_argument, nullptr, 'H'},
       {"data", required_argument, nullptr, 'd'},
-      {"compress-data", no_argument, nullptr, 'g'},
       {"multiply", required_argument, nullptr, 'm'},
       {"upgrade", no_argument, nullptr, 'u'},
       {"weight", required_argument, nullptr, 'p'},
@@ -2054,6 +2033,7 @@ int main(int argc, char **argv)
       {"color", no_argument, &flag, 3},
       {"continuation", no_argument, &flag, 4},
       {"version", no_argument, &flag, 5},
+      {"no-content-length", no_argument, &flag, 6},
       {nullptr, 0, nullptr, 0 }
     };
     int option_index = 0;
@@ -2151,7 +2131,7 @@ int main(int argc, char **argv)
 #ifdef HAVE_LIBXML2
       config.get_assets = true;
 #else // !HAVE_LIBXML2
-      std::cerr << "Warning: -a, --get-assets option cannot be used because\n"
+      std::cerr << "[WARNING]: -a, --get-assets option cannot be used because\n"
                 << "the binary was not compiled with libxml2."
                 << std::endl;
 #endif // !HAVE_LIBXML2
@@ -2162,9 +2142,6 @@ int main(int argc, char **argv)
     case 'd':
       config.datafile = strcmp("-", optarg) == 0 ? "/dev/stdin" : optarg;
       break;
-    case 'g':
-      config.compress_data = true;
-      break;
     case 'm':
       config.multiply = strtoul(optarg, nullptr, 10);
       break;
@@ -2201,6 +2178,10 @@ int main(int argc, char **argv)
         // version option
         print_version(std::cout);
         exit(EXIT_SUCCESS);
+      case 6:
+        // no-content-length option
+        config.no_content_length = true;
+        break;
       }
       break;
     default:
@@ -2217,6 +2198,8 @@ int main(int argc, char **argv)
   memset(&act, 0, sizeof(struct sigaction));
   act.sa_handler = SIG_IGN;
   sigaction(SIGPIPE, &act, nullptr);
+  OPENSSL_config(nullptr);
+  OpenSSL_add_all_algorithms();
   SSL_load_error_strings();
   SSL_library_init();
   reset_timer();
