@@ -37,6 +37,7 @@
 
 #include <openssl/ssl.h>
 #include <openssl/err.h>
+#include <openssl/conf.h>
 #include <nghttp2/nghttp2.h>
 
 #include "app_helper.h"
@@ -133,6 +134,13 @@ Options:
                      Set the number of worker threads.
                      Default: 1
   -e, --error-gzip   Make error response gzipped.
+  --dh-param-file=<PATH>
+                     Path to  file that contains DH  parameters in PEM
+                     format.  Without  this option, DHE  cipher suites
+                     are not available.
+  --early-response   Start  sending response  when request  HEADERS is
+                     received,   rather  than   complete  request   is
+                     received.
   --version          Display version information and exit.
   -h, --help         Display this help and exit.)"
       << std::endl;
@@ -159,6 +167,8 @@ int main(int argc, char **argv)
       {"no-tls", no_argument, &flag, 1},
       {"color", no_argument, &flag, 2},
       {"version", no_argument, &flag, 3},
+      {"dh-param-file", required_argument, &flag, 4},
+      {"early-response", no_argument, &flag, 5},
       {nullptr, 0, nullptr, 0}
     };
     int option_index = 0;
@@ -233,6 +243,14 @@ int main(int argc, char **argv)
         // version
         print_version(std::cout);
         exit(EXIT_SUCCESS);
+      case 4:
+        // dh-param-file
+        config.dh_param_file = optarg;
+        break;
+      case 5:
+        // early-response
+        config.early_response = true;
+        break;
       }
       break;
     default:
@@ -273,11 +291,12 @@ int main(int argc, char **argv)
   memset(&act, 0, sizeof(struct sigaction));
   act.sa_handler = SIG_IGN;
   sigaction(SIGPIPE, &act, nullptr);
+  OPENSSL_config(nullptr);
   OpenSSL_add_all_algorithms();
   SSL_load_error_strings();
   SSL_library_init();
 #ifndef NOTHREADS
-  ssl::LibsslGlobalLock();
+  ssl::LibsslGlobalLock lock;
 #endif // NOTHREADS
 
   reset_timer();
