@@ -29,10 +29,25 @@
 #  include <config.h>
 #endif // HAVE_CONFIG_H
 
+#include <sys/types.h>
+#include <sys/socket.h>
+
 #include <cassert>
 
 #include "shrpx_log.h"
 
+#ifndef SOCK_NONBLOCK
+#define SOCK_NONBLOCK 0
+#endif // !SOCK_NONBLOCK
+
+#ifndef SOCK_CLOEXEC
+#define SOCK_CLOEXEC 0
+#endif // !SOCK_CLOEXEC
+
+#ifndef HAVE__EXIT
+#define _Exit(status) _exit(status)
+#endif // !HAVE__EXIT
+
 #define DIE()                                   \
   assert(0);
 
