@@ -33,12 +33,15 @@
 #include "shrpx_config.h"
 #include "shrpx_http_downstream_connection.h"
 #include "shrpx_http2_downstream_connection.h"
-#include "shrpx_accesslog.h"
 #include "shrpx_ssl.h"
+#include "shrpx_worker.h"
+#include "shrpx_worker_config.h"
+#include "shrpx_downstream_connection_pool.h"
 #ifdef HAVE_SPDYLAY
 #include "shrpx_spdy_upstream.h"
 #endif // HAVE_SPDYLAY
 #include "util.h"
+#include "libevent_util.h"
 
 using namespace nghttp2;
 
@@ -48,6 +51,10 @@ namespace {
 void upstream_readcb(bufferevent *bev, void *arg)
 {
   auto handler = static_cast<ClientHandler*>(arg);
+  auto upstream = handler->get_upstream();
+  if(upstream) {
+    upstream->reset_timeouts();
+  }
   int rv = handler->on_read();
   if(rv != 0) {
     delete handler;
@@ -59,6 +66,11 @@ namespace {
 void upstream_writecb(bufferevent *bev, void *arg)
 {
   auto handler = static_cast<ClientHandler*>(arg);
+  auto upstream = handler->get_upstream();
+  if(upstream) {
+    upstream->reset_timeouts();
+  }
+
   // We actually depend on write low-water mark == 0.
   if(handler->get_outbuf_length() > 0) {
     // Possibly because of deferred callback, we may get this callback
@@ -69,7 +81,7 @@ void upstream_writecb(bufferevent *bev, void *arg)
     delete handler;
     return;
   }
-  auto upstream = handler->get_upstream();
+
   if(!upstream) {
     return;
   }
@@ -122,13 +134,6 @@ void upstream_eventcb(bufferevent *bev, short events, void *arg)
           CLOG(INFO, handler) << "SSL/TLS session reused";
         }
       }
-      // At this point, input buffer is already filled with some
-      // bytes.  The read callback is not called until new data
-      // come. So consume input buffer here.
-      if(handler->get_upstream()->on_read() != 0) {
-        delete handler;
-        return;
-      }
     }
   }
 }
@@ -222,38 +227,19 @@ void upstream_http1_connhd_readcb(bufferevent *bev, void *arg)
 }
 } // namespace
 
-namespace {
-void tls_raw_readcb(evbuffer *buffer, const evbuffer_cb_info *info, void *arg)
-{
-  auto handler = static_cast<ClientHandler*>(arg);
-  if(handler->get_tls_renegotiation()) {
-    if(LOG_ENABLED(INFO)) {
-      CLOG(INFO, handler) << "Close connection due to TLS renegotiation";
-    }
-    delete handler;
-  }
-}
-} // namespace
-
-namespace {
-void tls_raw_writecb(evbuffer *buffer, const evbuffer_cb_info *info, void *arg)
-{
-  auto handler = static_cast<ClientHandler*>(arg);
-  // upstream_writecb() is called when external bufferevent
-  // handler->bev's output buffer gets empty. But the underlying
-  // bufferevent may have pending output buffer.
-  upstream_writecb(handler->get_bev(), handler);
-}
-} // namespace
-
 ClientHandler::ClientHandler(bufferevent *bev,
                              bufferevent_rate_limit_group *rate_limit_group,
                              int fd, SSL *ssl,
-                             const char *ipaddr)
+                             const char *ipaddr,
+                             WorkerStat *worker_stat,
+                             DownstreamConnectionPool *dconn_pool)
   : ipaddr_(ipaddr),
+    dconn_pool_(dconn_pool),
     bev_(bev),
     http2session_(nullptr),
     ssl_(ssl),
+    reneg_shutdown_timerev_(nullptr),
+    worker_stat_(worker_stat),
     left_connhd_len_(NGHTTP2_CLIENT_CONNECTION_PREFACE_LEN),
     fd_(fd),
     should_close_after_write_(false),
@@ -262,33 +248,25 @@ ClientHandler::ClientHandler(bufferevent *bev,
 {
   int rv;
 
-  auto rate_limit_bev = bufferevent_get_underlying(bev_);
-  if(!rate_limit_bev) {
-    rate_limit_bev = bev_;
-  }
+  ++worker_stat->num_connections;
 
-  rv = bufferevent_set_rate_limit(rate_limit_bev,
-                                  get_config()->rate_limit_cfg);
+  rv = bufferevent_set_rate_limit(bev_, get_config()->rate_limit_cfg);
   if(rv == -1) {
     CLOG(FATAL, this) << "bufferevent_set_rate_limit() failed";
   }
 
-  rv = bufferevent_add_to_rate_limit_group(rate_limit_bev, rate_limit_group);
+  rv = bufferevent_add_to_rate_limit_group(bev_, rate_limit_group);
   if(rv == -1) {
     CLOG(FATAL, this) << "bufferevent_add_to_rate_limit_group() failed";
   }
 
-  bufferevent_enable(bev_, EV_READ | EV_WRITE);
+  util::bev_enable_unless(bev_, EV_READ | EV_WRITE);
   bufferevent_setwatermark(bev_, EV_READ, 0, SHRPX_READ_WATERMARK);
   set_upstream_timeouts(&get_config()->upstream_read_timeout,
                         &get_config()->upstream_write_timeout);
   if(ssl_) {
     SSL_set_app_data(ssl_, reinterpret_cast<char*>(this));
     set_bev_cb(nullptr, upstream_writecb, upstream_eventcb);
-    auto input = bufferevent_get_input(bufferevent_get_underlying(bev_));
-    evbuffer_add_cb(input, tls_raw_readcb, this);
-    auto output = bufferevent_get_output(bufferevent_get_underlying(bev_));
-    evbuffer_add_cb(output, tls_raw_writecb, this);
   } else {
     // For non-TLS version, first create HttpsUpstream. It may be
     // upgraded to HTTP/2 through HTTP Upgrade or direct HTTP/2
@@ -303,34 +281,36 @@ ClientHandler::~ClientHandler()
   if(LOG_ENABLED(INFO)) {
     CLOG(INFO, this) << "Deleting";
   }
+
+  --worker_stat_->num_connections;
+
+  // TODO If backend is http/2, and it is in CONNECTED state, signal
+  // it and make it loopbreak when output is zero.
+  if(worker_config->graceful_shutdown && worker_stat_->num_connections == 0) {
+    event_base_loopbreak(get_evbase());
+  }
+
+  if(reneg_shutdown_timerev_) {
+    event_free(reneg_shutdown_timerev_);
+  }
+
   if(ssl_) {
     SSL_set_app_data(ssl_, nullptr);
     SSL_set_shutdown(ssl_, SSL_RECEIVED_SHUTDOWN);
     SSL_shutdown(ssl_);
   }
 
-  auto underlying = bufferevent_get_underlying(bev_);
+  bufferevent_remove_from_rate_limit_group(bev_);
 
-  if(underlying) {
-    bufferevent_remove_from_rate_limit_group(underlying);
-  } else {
-    bufferevent_remove_from_rate_limit_group(bev_);
-  }
-
-  bufferevent_disable(bev_, EV_READ | EV_WRITE);
+  util::bev_disable_unless(bev_, EV_READ | EV_WRITE);
   bufferevent_free(bev_);
+
   if(ssl_) {
     SSL_free(ssl_);
   }
-  if(underlying) {
-    bufferevent_disable(underlying, EV_READ | EV_WRITE);
-    bufferevent_free(underlying);
-  }
+
   shutdown(fd_, SHUT_WR);
   close(fd_);
-  for(auto dconn : dconn_pool_) {
-    delete dconn;
-  }
   if(LOG_ENABLED(INFO)) {
     CLOG(INFO, this) << "Deleted";
   }
@@ -361,19 +341,15 @@ void ClientHandler::set_bev_cb
 void ClientHandler::set_upstream_timeouts(const timeval *read_timeout,
                                           const timeval *write_timeout)
 {
-  auto bev = bufferevent_get_underlying(bev_);
-
-  if(!bev) {
-    bev = bev_;
-  }
-
-  bufferevent_set_timeouts(bev, read_timeout, write_timeout);
+  bufferevent_set_timeouts(bev_, read_timeout, write_timeout);
 }
 
 int ClientHandler::validate_next_proto()
 {
   const unsigned char *next_proto = nullptr;
   unsigned int next_proto_len;
+  int rv;
+
   // First set callback for catch all cases
   set_bev_cb(upstream_readcb, upstream_writecb, upstream_eventcb);
   SSL_get0_next_proto_negotiated(ssl_, &next_proto, &next_proto_len);
@@ -384,7 +360,6 @@ int ClientHandler::validate_next_proto()
         CLOG(INFO, this) << "The negotiated next protocol: " << proto;
       }
       if(!ssl::in_proto_list(get_config()->npn_list,
-                             get_config()->npn_list_len,
                              next_proto, next_proto_len)) {
         break;
       }
@@ -392,25 +367,53 @@ int ClientHandler::validate_next_proto()
          memcmp(NGHTTP2_PROTO_VERSION_ID, next_proto,
                 NGHTTP2_PROTO_VERSION_ID_LEN) == 0) {
 
-        // For NPN, we must check security requirement here.
-        if(!ssl::check_http2_requirement(ssl_)) {
-          return -1;
-        }
-
         set_bev_cb(upstream_http2_connhd_readcb, upstream_writecb,
                    upstream_eventcb);
-        upstream_ = util::make_unique<Http2Upstream>(this);
+
+        auto http2_upstream = util::make_unique<Http2Upstream>(this);
+
+        if(!ssl::check_http2_requirement(ssl_)) {
+          rv = http2_upstream->terminate_session(NGHTTP2_INADEQUATE_SECURITY);
+
+          if(rv != 0) {
+            return -1;
+          }
+        }
+
+        upstream_ = std::move(http2_upstream);
+
+        // At this point, input buffer is already filled with some
+        // bytes.  The read callback is not called until new data
+        // come. So consume input buffer here.
+        upstream_http2_connhd_readcb(bev_, this);
+
         return 0;
       } else {
 #ifdef HAVE_SPDYLAY
         uint16_t version = spdylay_npn_get_version(next_proto, next_proto_len);
         if(version) {
           upstream_ = util::make_unique<SpdyUpstream>(version, this);
+
+          // At this point, input buffer is already filled with some
+          // bytes.  The read callback is not called until new data
+          // come. So consume input buffer here.
+          if(upstream_->on_read() != 0) {
+            return -1;
+          }
+
           return 0;
         }
 #endif // HAVE_SPDYLAY
         if(next_proto_len == 8 && memcmp("http/1.1", next_proto, 8) == 0) {
           upstream_ = util::make_unique<HttpsUpstream>(this);
+
+          // At this point, input buffer is already filled with some
+          // bytes.  The read callback is not called until new data
+          // come. So consume input buffer here.
+          if(upstream_->on_read() != 0) {
+            return -1;
+          }
+
           return 0;
         }
       }
@@ -427,6 +430,14 @@ int ClientHandler::validate_next_proto()
       CLOG(INFO, this) << "No protocol negotiated. Fallback to HTTP/1.1";
     }
     upstream_ = util::make_unique<HttpsUpstream>(this);
+
+    // At this point, input buffer is already filled with some bytes.
+    // The read callback is not called until new data come. So consume
+    // input buffer here.
+    if(upstream_->on_read() != 0) {
+      return -1;
+    }
+
     return 0;
   }
   if(LOG_ENABLED(INFO)) {
@@ -460,12 +471,14 @@ void ClientHandler::set_should_close_after_write(bool f)
   should_close_after_write_ = f;
 }
 
-void ClientHandler::pool_downstream_connection(DownstreamConnection *dconn)
+void ClientHandler::pool_downstream_connection
+(std::unique_ptr<DownstreamConnection> dconn)
 {
   if(LOG_ENABLED(INFO)) {
-    CLOG(INFO, this) << "Pooling downstream connection DCONN:" << dconn;
+    CLOG(INFO, this) << "Pooling downstream connection DCONN:" << dconn.get();
   }
-  dconn_pool_.insert(dconn);
+  dconn->set_client_handler(nullptr);
+  dconn_pool_->add_downstream_connection(std::move(dconn));
 }
 
 void ClientHandler::remove_downstream_connection(DownstreamConnection *dconn)
@@ -474,40 +487,43 @@ void ClientHandler::remove_downstream_connection(DownstreamConnection *dconn)
     CLOG(INFO, this) << "Removing downstream connection DCONN:" << dconn
                      << " from pool";
   }
-  dconn_pool_.erase(dconn);
+  dconn_pool_->remove_downstream_connection(dconn);
 }
 
-DownstreamConnection* ClientHandler::get_downstream_connection()
+std::unique_ptr<DownstreamConnection>
+ClientHandler::get_downstream_connection()
 {
-  if(dconn_pool_.empty()) {
+  auto dconn = dconn_pool_->pop_downstream_connection();
+
+  if(!dconn) {
     if(LOG_ENABLED(INFO)) {
       CLOG(INFO, this) << "Downstream connection pool is empty."
                        << " Create new one";
     }
+
     if(http2session_) {
-      return new Http2DownstreamConnection(this);
+      dconn = util::make_unique<Http2DownstreamConnection>
+        (dconn_pool_, http2session_);
     } else {
-      return new HttpDownstreamConnection(this);
-    }
-  } else {
-    auto dconn = *std::begin(dconn_pool_);
-    dconn_pool_.erase(dconn);
-    if(LOG_ENABLED(INFO)) {
-      CLOG(INFO, this) << "Reuse downstream connection DCONN:" << dconn
-                       << " from pool";
+      dconn = util::make_unique<HttpDownstreamConnection>(dconn_pool_);
     }
+    dconn->set_client_handler(this);
     return dconn;
   }
+
+  dconn->set_client_handler(this);
+
+  if(LOG_ENABLED(INFO)) {
+    CLOG(INFO, this) << "Reuse downstream connection DCONN:" << dconn.get()
+                     << " from pool";
+  }
+
+  return dconn;
 }
 
 size_t ClientHandler::get_outbuf_length()
 {
-  auto underlying = bufferevent_get_underlying(bev_);
-  auto len = evbuffer_get_length(bufferevent_get_output(bev_));
-  if(underlying) {
-    len += evbuffer_get_length(bufferevent_get_output(underlying));
-  }
-  return len;
+  return evbuffer_get_length(bufferevent_get_output(bev_));
 }
 
 SSL* ClientHandler::get_ssl() const
@@ -525,6 +541,17 @@ Http2Session* ClientHandler::get_http2_session() const
   return http2session_;
 }
 
+void ClientHandler::set_http1_connect_blocker
+(ConnectBlocker *http1_connect_blocker)
+{
+  http1_connect_blocker_ = http1_connect_blocker;
+}
+
+ConnectBlocker* ClientHandler::get_http1_connect_blocker() const
+{
+  return http1_connect_blocker_;
+}
+
 size_t ClientHandler::get_left_connhd_len() const
 {
   return left_connhd_len_;
@@ -588,8 +615,35 @@ bool ClientHandler::get_tls_handshake() const
   return tls_handshake_;
 }
 
+namespace {
+void shutdown_cb(evutil_socket_t fd, short what, void *arg)
+{
+  auto handler = static_cast<ClientHandler*>(arg);
+
+  if(LOG_ENABLED(INFO)) {
+    CLOG(INFO, handler) << "Close connection due to TLS renegotiation";
+  }
+
+  delete handler;
+}
+} // namespace
+
 void ClientHandler::set_tls_renegotiation(bool f)
 {
+  if(tls_renegotiation_ == false) {
+    if(LOG_ENABLED(INFO)) {
+      CLOG(INFO, this) << "TLS renegotiation detected. "
+                       << "Start shutdown timer now.";
+    }
+
+    reneg_shutdown_timerev_ = evtimer_new(get_evbase(), shutdown_cb, this);
+    event_priority_set(reneg_shutdown_timerev_, 0);
+
+    timeval timeout = {0, 0};
+
+    // TODO What to do if this failed?
+    evtimer_add(reneg_shutdown_timerev_, &timeout);
+  }
   tls_renegotiation_ = f;
 }
 
