@@ -27,7 +27,6 @@
 
 #include "shrpx.h"
 
-#include <set>
 #include <memory>
 
 #include <event.h>
@@ -41,12 +40,17 @@ class Upstream;
 class DownstreamConnection;
 class Http2Session;
 class HttpsUpstream;
+class ConnectBlocker;
+class DownstreamConnectionPool;
+struct WorkerStat;
 
 class ClientHandler {
 public:
   ClientHandler(bufferevent *bev,
                 bufferevent_rate_limit_group *rate_limit_group,
-                int fd, SSL *ssl, const char *ipaddr);
+                int fd, SSL *ssl, const char *ipaddr,
+                WorkerStat *worker_stat,
+                DownstreamConnectionPool *dconn_pool);
   ~ClientHandler();
   int on_read();
   int on_event();
@@ -62,13 +66,15 @@ public:
   void set_should_close_after_write(bool f);
   Upstream* get_upstream();
 
-  void pool_downstream_connection(DownstreamConnection *dconn);
+  void pool_downstream_connection(std::unique_ptr<DownstreamConnection> dconn);
   void remove_downstream_connection(DownstreamConnection *dconn);
-  DownstreamConnection* get_downstream_connection();
+  std::unique_ptr<DownstreamConnection> get_downstream_connection();
   size_t get_outbuf_length();
   SSL* get_ssl() const;
   void set_http2_session(Http2Session *http2session);
   Http2Session* get_http2_session() const;
+  void set_http1_connect_blocker(ConnectBlocker *http1_connect_blocker);
+  ConnectBlocker* get_http1_connect_blocker() const;
   size_t get_left_connhd_len() const;
   void set_left_connhd_len(size_t left);
   // Call this function when HTTP/2 connection header is received at
@@ -86,14 +92,17 @@ public:
   void set_tls_renegotiation(bool f);
   bool get_tls_renegotiation() const;
 private:
-  std::set<DownstreamConnection*> dconn_pool_;
   std::unique_ptr<Upstream> upstream_;
   std::string ipaddr_;
+  DownstreamConnectionPool *dconn_pool_;
   bufferevent *bev_;
   // Shared HTTP2 session for each thread. NULL if backend is not
   // HTTP2. Not deleted by this object.
   Http2Session *http2session_;
+  ConnectBlocker *http1_connect_blocker_;
   SSL *ssl_;
+  event *reneg_shutdown_timerev_;
+  WorkerStat *worker_stat_;
   // The number of bytes of HTTP/2 client connection header to read
   size_t left_connhd_len_;
   int fd_;
