@@ -74,7 +74,12 @@ const char SHRPX_OPT_FRONTEND_READ_TIMEOUT[] = "frontend-read-timeout";
 const char SHRPX_OPT_FRONTEND_WRITE_TIMEOUT[] = "frontend-write-timeout";
 const char SHRPX_OPT_BACKEND_READ_TIMEOUT[] = "backend-read-timeout";
 const char SHRPX_OPT_BACKEND_WRITE_TIMEOUT[] = "backend-write-timeout";
-const char SHRPX_OPT_ACCESSLOG[] = "accesslog";
+const char SHRPX_OPT_STREAM_READ_TIMEOUT[] = "stream-read-timeout";
+const char SHRPX_OPT_STREAM_WRITE_TIMEOUT[] = "stream-write-timeout";
+const char SHRPX_OPT_ACCESSLOG_FILE[] = "accesslog-file";
+const char SHRPX_OPT_ACCESSLOG_SYSLOG[] = "accesslog-syslog";
+const char SHRPX_OPT_ERRORLOG_FILE[] = "errorlog-file";
+const char SHRPX_OPT_ERRORLOG_SYSLOG[] = "errorlog-syslog";
 const char
 SHRPX_OPT_BACKEND_KEEP_ALIVE_TIMEOUT[] = "backend-keep-alive-timeout";
 const char
@@ -89,11 +94,9 @@ const char SHRPX_OPT_BACKEND_NO_TLS[] = "backend-no-tls";
 const char SHRPX_OPT_BACKEND_TLS_SNI_FIELD[] = "backend-tls-sni-field";
 const char SHRPX_OPT_PID_FILE[] = "pid-file";
 const char SHRPX_OPT_USER[] = "user";
-const char SHRPX_OPT_SYSLOG[] = "syslog";
 const char SHRPX_OPT_SYSLOG_FACILITY[] = "syslog-facility";
 const char SHRPX_OPT_BACKLOG[] = "backlog";
 const char SHRPX_OPT_CIPHERS[] = "ciphers";
-const char SHRPX_OPT_HONOR_CIPHER_ORDER[] = "honor-cipher-order";
 const char SHRPX_OPT_CLIENT[] = "client";
 const char SHRPX_OPT_INSECURE[] = "insecure";
 const char SHRPX_OPT_CACERT[] = "cacert";
@@ -123,6 +126,12 @@ const char SHRPX_OPT_FRONTEND_FRAME_DEBUG[] = "frontend-frame-debug";
 const char SHRPX_OPT_PADDING[] = "padding";
 const char SHRPX_OPT_ALTSVC[] = "altsvc";
 const char SHRPX_OPT_ADD_RESPONSE_HEADER[] = "add-response-header";
+const char SHRPX_OPT_WORKER_FRONTEND_CONNECTIONS[] =
+  "worker-frontend-connections";
+const char SHRPX_OPT_NO_LOCATION_REWRITE[] = "no-location-rewrite";
+const char SHRPX_OPT_BACKEND_CONNECTIONS_PER_FRONTEND[] =
+  "backend-connections-per-frontend";
+const char SHRPX_OPT_LISTENER_DISABLE_TIMEOUT[] = "listener-disable-timeout";
 
 namespace {
 Config *config = nullptr;
@@ -190,17 +199,19 @@ bool is_secure(const char *filename)
 }
 } // namespace
 
-namespace {
 FILE* open_file_for_write(const char *filename)
 {
   auto f = fopen(filename, "wb");
-  if(f == NULL) {
+  if(f == nullptr) {
     LOG(ERROR) << "Failed to open " << filename << " for writing. Cause: "
                << strerror(errno);
+    return nullptr;
   }
+
+  evutil_make_socket_closeonexec(fileno(f));
+
   return f;
 }
-} // namespace
 
 std::string read_passwd_from_file(const char *filename)
 {
@@ -222,20 +233,28 @@ std::string read_passwd_from_file(const char *filename)
   return line;
 }
 
-void set_config_str(char **destp, const char *val)
+std::unique_ptr<char[]> strcopy(const char *val)
 {
-  if(*destp) {
-    free(*destp);
-  }
-  *destp = strdup(val);
+  auto len = strlen(val);
+  auto res = util::make_unique<char[]>(len + 1);
+  memcpy(res.get(), val, len + 1);
+  return res;
 }
 
-std::unique_ptr<char*[]> parse_config_str_list(size_t *outlen, const char *s)
+std::unique_ptr<char[]> strcopy(const std::string& val)
+{
+  auto len = val.size();
+  auto res = util::make_unique<char[]>(len + 1);
+  memcpy(res.get(), val.c_str(), len + 1);
+  return res;
+}
+
+std::vector<char*> parse_config_str_list(const char *s)
 {
   size_t len = 1;
   for(const char *first = s, *p = nullptr; (p = strchr(first, ','));
       ++len, first = p + 1);
-  auto list = util::make_unique<char*[]>(len);
+  auto list = std::vector<char*>(len);
   auto first = strdup(s);
   len = 0;
   for(;;) {
@@ -248,10 +267,20 @@ std::unique_ptr<char*[]> parse_config_str_list(size_t *outlen, const char *s)
     first = p + 1;
   }
   list[len++] = first;
-  *outlen = len;
+
   return list;
 }
 
+void clear_config_str_list(std::vector<char*>& list)
+{
+  if(list.empty()) {
+    return;
+  }
+
+  free(list[0]);
+  list.clear();
+}
+
 std::pair<std::string, std::string> parse_header(const char *optarg)
 {
   // We skip possible ":" at the start of optarg.
@@ -268,6 +297,60 @@ std::pair<std::string, std::string> parse_header(const char *optarg)
   return {std::string(optarg, colon), std::string(value, strlen(value))};
 }
 
+template<typename T>
+int parse_uint(T *dest, const char *opt, const char *optarg)
+{
+  char *end = nullptr;
+
+  errno = 0;
+
+  auto val = strtol(optarg, &end, 10);
+
+  if(!optarg[0] || errno != 0 || *end || val < 0) {
+    LOG(ERROR) << opt << ": bad value.  Specify an integer >= 0.";
+    return -1;
+  }
+
+  *dest = val;
+
+  return 0;
+}
+
+template<typename T>
+int parse_int(T *dest, const char *opt, const char *optarg)
+{
+  char *end = nullptr;
+
+  errno = 0;
+
+  auto val = strtol(optarg, &end, 10);
+
+  if(!optarg[0] || errno != 0 || *end) {
+    LOG(ERROR) << opt << ": bad value.  Specify an integer.";
+    return -1;
+  }
+
+  *dest = val;
+
+  return 0;
+}
+
+namespace {
+int parse_timeval(timeval *dest, const char *opt, const char *optarg)
+{
+  time_t sec;
+
+  if(parse_uint(&sec, opt, optarg) != 0) {
+    return -1;
+  }
+
+  dest->tv_sec = sec;
+  dest->tv_usec = 0;
+
+  return 0;
+}
+} // namespace
+
 int parse_config(const char *opt, const char *optarg)
 {
   char host[NI_MAXHOST];
@@ -275,131 +358,264 @@ int parse_config(const char *opt, const char *optarg)
   if(util::strieq(opt, SHRPX_OPT_BACKEND)) {
     if(split_host_port(host, sizeof(host), &port, optarg) == -1) {
       return -1;
-    } else {
-      set_config_str(&mod_config()->downstream_host, host);
-      mod_config()->downstream_port = port;
     }
-  } else if(util::strieq(opt, SHRPX_OPT_FRONTEND)) {
+
+    mod_config()->downstream_host = strcopy(host);
+    mod_config()->downstream_port = port;
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND)) {
     if(split_host_port(host, sizeof(host), &port, optarg) == -1) {
       return -1;
-    } else {
-      set_config_str(&mod_config()->host, host);
-      mod_config()->port = port;
     }
-  } else if(util::strieq(opt, SHRPX_OPT_WORKERS)) {
-    mod_config()->num_worker = strtol(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_HTTP2_MAX_CONCURRENT_STREAMS)) {
-    mod_config()->http2_max_concurrent_streams = strtol(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_LOG_LEVEL)) {
+
+    mod_config()->host = strcopy(host);
+    mod_config()->port = port;
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_WORKERS)) {
+    return parse_uint(&mod_config()->num_worker, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_HTTP2_MAX_CONCURRENT_STREAMS)) {
+    return parse_uint(&mod_config()->http2_max_concurrent_streams,
+                      opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_LOG_LEVEL)) {
     if(Log::set_severity_level_by_name(optarg) == -1) {
-      LOG(ERROR) << "Invalid severity level: " << optarg;
+      LOG(ERROR) << opt << ": Invalid severity level: " << optarg;
       return -1;
     }
-  } else if(util::strieq(opt, SHRPX_OPT_DAEMON)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_DAEMON)) {
     mod_config()->daemon = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_HTTP2_PROXY)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_HTTP2_PROXY)) {
     mod_config()->http2_proxy = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_HTTP2_BRIDGE)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_HTTP2_BRIDGE)) {
     mod_config()->http2_bridge = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_CLIENT_PROXY)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_CLIENT_PROXY)) {
     mod_config()->client_proxy = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_ADD_X_FORWARDED_FOR)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_ADD_X_FORWARDED_FOR)) {
     mod_config()->add_x_forwarded_for = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_NO_VIA)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_NO_VIA)) {
     mod_config()->no_via = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_READ_TIMEOUT)) {
-    timeval tv = {strtol(optarg, nullptr, 10), 0};
-    mod_config()->http2_upstream_read_timeout = tv;
-  } else if(util::strieq(opt, SHRPX_OPT_FRONTEND_READ_TIMEOUT)) {
-    timeval tv = {strtol(optarg, nullptr, 10), 0};
-    mod_config()->upstream_read_timeout = tv;
-  } else if(util::strieq(opt, SHRPX_OPT_FRONTEND_WRITE_TIMEOUT)) {
-    timeval tv = {strtol(optarg, nullptr, 10), 0};
-    mod_config()->upstream_write_timeout = tv;
-  } else if(util::strieq(opt, SHRPX_OPT_BACKEND_READ_TIMEOUT)) {
-    timeval tv = {strtol(optarg, nullptr, 10), 0};
-    mod_config()->downstream_read_timeout = tv;
-  } else if(util::strieq(opt, SHRPX_OPT_BACKEND_WRITE_TIMEOUT)) {
-    timeval tv = {strtol(optarg, nullptr, 10), 0};
-    mod_config()->downstream_write_timeout = tv;
-  } else if(util::strieq(opt, SHRPX_OPT_ACCESSLOG)) {
-    mod_config()->accesslog = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_BACKEND_KEEP_ALIVE_TIMEOUT)) {
-    timeval tv = {strtol(optarg, nullptr, 10), 0};
-    mod_config()->downstream_idle_read_timeout = tv;
-  } else if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_WINDOW_BITS) ||
-            util::strieq(opt, SHRPX_OPT_BACKEND_HTTP2_WINDOW_BITS)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_READ_TIMEOUT)) {
+    return parse_timeval(&mod_config()->http2_upstream_read_timeout,
+                         opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND_READ_TIMEOUT)) {
+    return parse_timeval(&mod_config()->upstream_read_timeout, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND_WRITE_TIMEOUT)) {
+    return parse_timeval(&mod_config()->upstream_write_timeout, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKEND_READ_TIMEOUT)) {
+    return parse_timeval(&mod_config()->downstream_read_timeout, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKEND_WRITE_TIMEOUT)) {
+    return parse_timeval(&mod_config()->downstream_write_timeout, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_STREAM_READ_TIMEOUT)) {
+    return parse_timeval(&mod_config()->stream_read_timeout, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_STREAM_WRITE_TIMEOUT)) {
+    return parse_timeval(&mod_config()->stream_write_timeout, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_ACCESSLOG_FILE)) {
+    mod_config()->accesslog_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_ACCESSLOG_SYSLOG)) {
+    mod_config()->accesslog_syslog = util::strieq(optarg, "yes");
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_ERRORLOG_FILE)) {
+    mod_config()->errorlog_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_ERRORLOG_SYSLOG)) {
+    mod_config()->errorlog_syslog = util::strieq(optarg, "yes");
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKEND_KEEP_ALIVE_TIMEOUT)) {
+    return parse_timeval(&mod_config()->downstream_idle_read_timeout,
+                         opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_WINDOW_BITS) ||
+     util::strieq(opt, SHRPX_OPT_BACKEND_HTTP2_WINDOW_BITS)) {
+
     size_t *resp;
-    const char *optname;
+
     if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_WINDOW_BITS)) {
       resp = &mod_config()->http2_upstream_window_bits;
-      optname = SHRPX_OPT_FRONTEND_HTTP2_WINDOW_BITS;
     } else {
       resp = &mod_config()->http2_downstream_window_bits;
-      optname = SHRPX_OPT_BACKEND_HTTP2_WINDOW_BITS;
     }
+
     errno = 0;
-    unsigned long int n = strtoul(optarg, nullptr, 10);
-    if(errno == 0 && n < 31) {
-      *resp = n;
-    } else {
-      LOG(ERROR) << "--" << optname
-                 << " specify the integer in the range [0, 30], inclusive";
+
+    int n;
+
+    if(parse_uint(&n, opt, optarg) != 0) {
       return -1;
     }
-  } else if(util::strieq(opt,
-                         SHRPX_OPT_FRONTEND_HTTP2_CONNECTION_WINDOW_BITS) ||
-            util::strieq(opt,
-                         SHRPX_OPT_BACKEND_HTTP2_CONNECTION_WINDOW_BITS)) {
+
+    if(n >= 31) {
+      LOG(ERROR) << opt
+                 << ": specify the integer in the range [0, 30], inclusive";
+      return -1;
+    }
+
+    *resp = n;
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_CONNECTION_WINDOW_BITS) ||
+     util::strieq(opt, SHRPX_OPT_BACKEND_HTTP2_CONNECTION_WINDOW_BITS)) {
+
     size_t *resp;
-    const char *optname;
+
     if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_CONNECTION_WINDOW_BITS)) {
       resp = &mod_config()->http2_upstream_connection_window_bits;
-      optname = SHRPX_OPT_FRONTEND_HTTP2_CONNECTION_WINDOW_BITS;
     } else {
       resp = &mod_config()->http2_downstream_connection_window_bits;
-      optname = SHRPX_OPT_BACKEND_HTTP2_CONNECTION_WINDOW_BITS;
     }
+
     errno = 0;
-    unsigned long int n = strtoul(optarg, 0, 10);
-    if(errno == 0 && n >= 16 && n < 31) {
-      *resp = n;
-    } else {
-      LOG(ERROR) << "--" << optname
-                 << " specify the integer in the range [16, 30], inclusive";
+
+    int n;
+
+    if(parse_uint(&n, opt, optarg) != 0) {
       return -1;
     }
-  } else if(util::strieq(opt, SHRPX_OPT_FRONTEND_NO_TLS)) {
+
+    if(n < 16 || n >= 31) {
+      LOG(ERROR) << opt
+                 << ": specify the integer in the range [16, 30], inclusive";
+      return -1;
+    }
+
+    *resp = n;
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND_NO_TLS)) {
     mod_config()->upstream_no_tls = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_BACKEND_NO_TLS)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKEND_NO_TLS)) {
     mod_config()->downstream_no_tls = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_BACKEND_TLS_SNI_FIELD)) {
-    set_config_str(&mod_config()->backend_tls_sni_name, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_PID_FILE)) {
-    set_config_str(&mod_config()->pid_file, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_USER)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKEND_TLS_SNI_FIELD)) {
+    mod_config()->backend_tls_sni_name = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_PID_FILE)) {
+    mod_config()->pid_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_USER)) {
     auto pwd = getpwnam(optarg);
     if(!pwd) {
-      LOG(ERROR) << "--user: failed to get uid from " << optarg
+      LOG(ERROR) << opt << ": failed to get uid from " << optarg
                  << ": " << strerror(errno);
       return -1;
     }
     mod_config()->uid = pwd->pw_uid;
     mod_config()->gid = pwd->pw_gid;
-  } else if(util::strieq(opt, SHRPX_OPT_PRIVATE_KEY_FILE)) {
-    set_config_str(&mod_config()->private_key_file, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_PRIVATE_KEY_PASSWD_FILE)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_PRIVATE_KEY_FILE)) {
+    mod_config()->private_key_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_PRIVATE_KEY_PASSWD_FILE)) {
     auto passwd = read_passwd_from_file(optarg);
     if (passwd.empty()) {
-      LOG(ERROR) << "Couldn't read key file's passwd from " << optarg;
+      LOG(ERROR) << opt << ": Couldn't read key file's passwd from " << optarg;
       return -1;
     }
-    set_config_str(&mod_config()->private_key_passwd, passwd.c_str());
-  } else if(util::strieq(opt, SHRPX_OPT_CERTIFICATE_FILE)) {
-    set_config_str(&mod_config()->cert_file, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_DH_PARAM_FILE)) {
-    set_config_str(&mod_config()->dh_param_file, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_SUBCERT)) {
+    mod_config()->private_key_passwd = strcopy(passwd);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_CERTIFICATE_FILE)) {
+    mod_config()->cert_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_DH_PARAM_FILE)) {
+    mod_config()->dh_param_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_SUBCERT)) {
     // Private Key file and certificate file separated by ':'.
     const char *sp = strchr(optarg, ':');
     if(sp) {
@@ -407,32 +623,75 @@ int parse_config(const char *opt, const char *optarg)
       // TODO Do we need private key for subcert?
       mod_config()->subcerts.emplace_back(keyfile, sp+1);
     }
-  } else if(util::strieq(opt, SHRPX_OPT_SYSLOG)) {
-    mod_config()->syslog = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_SYSLOG_FACILITY)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_SYSLOG_FACILITY)) {
     int facility = int_syslog_facility(optarg);
     if(facility == -1) {
-      LOG(ERROR) << "Unknown syslog facility: " << optarg;
+      LOG(ERROR) << opt << ": Unknown syslog facility: " << optarg;
       return -1;
     }
     mod_config()->syslog_facility = facility;
-  } else if(util::strieq(opt, SHRPX_OPT_BACKLOG)) {
-    mod_config()->backlog = strtol(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_CIPHERS)) {
-    set_config_str(&mod_config()->ciphers, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_HONOR_CIPHER_ORDER)) {
-    mod_config()->honor_cipher_order = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_CLIENT)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKLOG)) {
+    int n;
+    if(parse_int(&n, opt, optarg) != 0) {
+      return -1;
+    }
+
+    if(n < -1) {
+      LOG(ERROR) << opt << ": " << optarg << " is not allowed";
+
+      return -1;
+    }
+
+    mod_config()->backlog = n;
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_CIPHERS)) {
+    mod_config()->ciphers = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_CLIENT)) {
     mod_config()->client = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_INSECURE)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_INSECURE)) {
     mod_config()->insecure = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_CACERT)) {
-    set_config_str(&mod_config()->cacert, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_BACKEND_IPV4)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_CACERT)) {
+    mod_config()->cacert = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKEND_IPV4)) {
     mod_config()->backend_ipv4 = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_BACKEND_IPV6)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKEND_IPV6)) {
     mod_config()->backend_ipv6 = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_BACKEND_HTTP_PROXY_URI)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKEND_HTTP_PROXY_URI)) {
     // parse URI and get hostname, port and optionally userinfo.
     http_parser_url u;
     memset(&u, 0, sizeof(u));
@@ -445,99 +704,168 @@ int parse_config(const char *opt, const char *optarg)
         // userinfo component is empty string.
         if(!val.empty()) {
           val = util::percentDecode(val.begin(), val.end());
-          set_config_str(&mod_config()->downstream_http_proxy_userinfo,
-                         val.c_str());
+          mod_config()->downstream_http_proxy_userinfo = strcopy(val);
         }
       }
       if(u.field_set & UF_HOST) {
         http2::copy_url_component(val, &u, UF_HOST, optarg);
-        set_config_str(&mod_config()->downstream_http_proxy_host, val.c_str());
+        mod_config()->downstream_http_proxy_host = strcopy(val);
       } else {
-        LOG(ERROR) << "backend-http-proxy-uri does not contain hostname";
+        LOG(ERROR) << opt << ": no hostname specified";
         return -1;
       }
       if(u.field_set & UF_PORT) {
         mod_config()->downstream_http_proxy_port = u.port;
       } else {
-        LOG(ERROR) << "backend-http-proxy-uri does not contain port";
+        LOG(ERROR) << opt << ": no port specified";
         return -1;
       }
     } else {
-      LOG(ERROR) << "Could not parse backend-http-proxy-uri";
-        return -1;
+      LOG(ERROR) << opt << ": parse error";
+      return -1;
     }
-  } else if(util::strieq(opt, SHRPX_OPT_READ_RATE)) {
-    mod_config()->read_rate = strtoul(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_READ_BURST)) {
-    mod_config()->read_burst = strtoul(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_WRITE_RATE)) {
-    mod_config()->write_rate = strtoul(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_WRITE_BURST)) {
-    mod_config()->write_burst = strtoul(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_WORKER_READ_RATE)) {
-    mod_config()->worker_read_rate = strtoul(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_WORKER_READ_BURST)) {
-    mod_config()->worker_read_burst = strtoul(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_WORKER_WRITE_RATE)) {
-    mod_config()->worker_write_rate = strtoul(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_WORKER_WRITE_BURST)) {
-    mod_config()->worker_write_burst = strtoul(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_NPN_LIST)) {
-    delete [] mod_config()->npn_list;
-    mod_config()->npn_list = parse_config_str_list(&mod_config()->npn_list_len,
-                                                   optarg).release();
-  } else if(util::strieq(opt, SHRPX_OPT_TLS_PROTO_LIST)) {
-    delete [] mod_config()->tls_proto_list;
-    mod_config()->tls_proto_list = parse_config_str_list
-      (&mod_config()->tls_proto_list_len, optarg).release();
-  } else if(util::strieq(opt, SHRPX_OPT_VERIFY_CLIENT)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_READ_RATE)) {
+    return parse_uint(&mod_config()->read_rate, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_READ_BURST)) {
+    int n;
+
+    if(parse_uint(&n, opt, optarg) != 0) {
+      return -1;
+    }
+
+    if(n == 0) {
+      LOG(ERROR) << opt << ": specify integer strictly larger than 0";
+
+      return -1;
+    }
+
+    mod_config()->read_burst = n;
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_WRITE_RATE)) {
+    return parse_uint(&mod_config()->write_rate, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_WRITE_BURST)) {
+    return parse_uint(&mod_config()->write_burst, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_WORKER_READ_RATE)) {
+    return parse_uint(&mod_config()->worker_read_rate, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_WORKER_READ_BURST)) {
+    return parse_uint(&mod_config()->worker_read_burst, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_WORKER_WRITE_RATE)) {
+    return parse_uint(&mod_config()->worker_write_rate, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_WORKER_WRITE_BURST)) {
+    return parse_uint(&mod_config()->worker_write_burst, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_NPN_LIST)) {
+    clear_config_str_list(mod_config()->npn_list);
+
+    mod_config()->npn_list = parse_config_str_list(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_TLS_PROTO_LIST)) {
+    clear_config_str_list(mod_config()->tls_proto_list);
+
+    mod_config()->tls_proto_list = parse_config_str_list(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_VERIFY_CLIENT)) {
     mod_config()->verify_client = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_VERIFY_CLIENT_CACERT)) {
-    set_config_str(&mod_config()->verify_client_cacert, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_CLIENT_PRIVATE_KEY_FILE)) {
-    set_config_str(&mod_config()->client_private_key_file, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_CLIENT_CERT_FILE)) {
-    set_config_str(&mod_config()->client_cert_file, optarg);
-  } else if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_DUMP_REQUEST_HEADER)) {
-    auto f = open_file_for_write(optarg);
-    if(f == NULL) {
-      return -1;
-    }
-    mod_config()->http2_upstream_dump_request_header = f;
-  } else if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_DUMP_RESPONSE_HEADER)) {
-    auto f = open_file_for_write(optarg);
-    if(f == NULL) {
-      return -1;
-    }
-    mod_config()->http2_upstream_dump_response_header = f;
-  } else if(util::strieq(opt, SHRPX_OPT_HTTP2_NO_COOKIE_CRUMBLING)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_VERIFY_CLIENT_CACERT)) {
+    mod_config()->verify_client_cacert = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_CLIENT_PRIVATE_KEY_FILE)) {
+    mod_config()->client_private_key_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_CLIENT_CERT_FILE)) {
+    mod_config()->client_cert_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_DUMP_REQUEST_HEADER)) {
+    mod_config()->http2_upstream_dump_request_header_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND_HTTP2_DUMP_RESPONSE_HEADER)) {
+    mod_config()->http2_upstream_dump_response_header_file = strcopy(optarg);
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_HTTP2_NO_COOKIE_CRUMBLING)) {
     mod_config()->http2_no_cookie_crumbling = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_FRONTEND_FRAME_DEBUG)) {
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_FRONTEND_FRAME_DEBUG)) {
     mod_config()->upstream_frame_debug = util::strieq(optarg, "yes");
-  } else if(util::strieq(opt, SHRPX_OPT_PADDING)) {
-    mod_config()->padding = strtoul(optarg, nullptr, 10);
-  } else if(util::strieq(opt, SHRPX_OPT_ALTSVC)) {
-    size_t len;
 
-    auto tokens = parse_config_str_list(&len, optarg);
+    return 0;
+  }
 
-    if(len < 2) {
+  if(util::strieq(opt, SHRPX_OPT_PADDING)) {
+    return parse_uint(&mod_config()->padding, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_ALTSVC)) {
+    auto tokens = parse_config_str_list(optarg);
+
+    if(tokens.size() < 2) {
       // Requires at least protocol_id and port
-      LOG(ERROR) << "altsvc: too few parameters: " << optarg;
+      LOG(ERROR) << opt << ": too few parameters: " << optarg;
       return -1;
     }
 
-    if(len > 4) {
+    if(tokens.size() > 4) {
       // We only need protocol_id, port, host and origin
-      LOG(ERROR) << "altsvc: too many parameters: " << optarg;
+      LOG(ERROR) << opt << ": too many parameters: " << optarg;
       return -1;
     }
 
-    errno = 0;
-    auto port = strtoul(tokens[1], nullptr, 10);
+    int port;
 
-    if(errno != 0 || port < 1 || port > std::numeric_limits<uint16_t>::max()) {
-      LOG(ERROR) << "altsvc: port is invalid: " << tokens[1];
+    if(parse_uint(&port, opt, tokens[1]) != 0) {
+      return -1;
+    }
+
+    if(port < 1 ||
+       port > static_cast<int>(std::numeric_limits<uint16_t>::max())) {
+      LOG(ERROR) << opt << ": port is invalid: " << tokens[1];
       return -1;
     }
 
@@ -548,11 +876,11 @@ int parse_config(const char *opt, const char *optarg)
     altsvc.protocol_id = tokens[0];
     altsvc.protocol_id_len = strlen(altsvc.protocol_id);
 
-    if(len > 2) {
+    if(tokens.size() > 2) {
       altsvc.host = tokens[2];
       altsvc.host_len = strlen(altsvc.host);
 
-      if(len > 3) {
+      if(tokens.size() > 3) {
         altsvc.origin = tokens[3];
         altsvc.origin_len = strlen(altsvc.origin);
       }
@@ -560,21 +888,62 @@ int parse_config(const char *opt, const char *optarg)
 
     mod_config()->altsvcs.push_back(std::move(altsvc));
 
-  } else if(util::strieq(opt, SHRPX_OPT_ADD_RESPONSE_HEADER)) {
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_ADD_RESPONSE_HEADER)) {
     auto p = parse_header(optarg);
     if(p.first.empty()) {
-      LOG(ERROR) << "add-response-header: header field name is empty: "
-                 << optarg;
+      LOG(ERROR) << opt << ": header field name is empty: " << optarg;
       return -1;
     }
     mod_config()->add_response_headers.push_back(std::move(p));
-  } else if(util::strieq(opt, "conf")) {
-    LOG(WARNING) << "conf is ignored";
-  } else {
-    LOG(ERROR) << "Unknown option: " << opt;
-    return -1;
+
+    return 0;
   }
-  return 0;
+
+  if(util::strieq(opt, SHRPX_OPT_WORKER_FRONTEND_CONNECTIONS)) {
+    return parse_uint(&mod_config()->worker_frontend_connections, opt, optarg);
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_NO_LOCATION_REWRITE)) {
+    mod_config()->no_location_rewrite = util::strieq(optarg, "yes");
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_BACKEND_CONNECTIONS_PER_FRONTEND)) {
+    int n;
+
+    if(parse_uint(&n, opt, optarg) != 0) {
+      return -1;
+    }
+
+    if(n < 1) {
+      LOG(ERROR) << opt << ": specify the integer more than or equal to 1";
+
+      return -1;
+    }
+
+    mod_config()->max_downstream_connections = n;
+
+    return 0;
+  }
+
+  if(util::strieq(opt, SHRPX_OPT_LISTENER_DISABLE_TIMEOUT)) {
+    return parse_timeval(&mod_config()->listener_disable_timeout,
+                         opt, optarg);
+  }
+
+  if(util::strieq(opt, "conf")) {
+    LOG(WARNING) << "conf: ignored";
+
+    return 0;
+  }
+
+  LOG(ERROR) << "Unknown option: " << opt;
+
+  return -1;
 }
 
 int load_config(const char *filename)
@@ -657,47 +1026,85 @@ int int_syslog_facility(const char *strfacility)
 {
   if(util::strieq(strfacility, "auth")) {
     return LOG_AUTH;
-  } else if(util::strieq(strfacility, "authpriv")) {
-    return LOG_AUTHPRIV;
-  } else if(util::strieq(strfacility, "cron")) {
-    return LOG_CRON;
-  } else if(util::strieq(strfacility, "daemon")) {
-    return LOG_DAEMON;
-  } else if(util::strieq(strfacility, "ftp")) {
-    return LOG_FTP;
-  } else if(util::strieq(strfacility, "kern")) {
-    return LOG_KERN;
-  } else if(util::strieq(strfacility, "local0")) {
-    return LOG_LOCAL0;
-  } else if(util::strieq(strfacility, "local1")) {
-    return LOG_LOCAL1;
-  } else if(util::strieq(strfacility, "local2")) {
-    return LOG_LOCAL2;
-  } else if(util::strieq(strfacility, "local3")) {
-    return LOG_LOCAL3;
-  } else if(util::strieq(strfacility, "local4")) {
-    return LOG_LOCAL4;
-  } else if(util::strieq(strfacility, "local5")) {
-    return LOG_LOCAL5;
-  } else if(util::strieq(strfacility, "local6")) {
-    return LOG_LOCAL6;
-  } else if(util::strieq(strfacility, "local7")) {
-    return LOG_LOCAL7;
-  } else if(util::strieq(strfacility, "lpr")) {
-    return LOG_LPR;
-  } else if(util::strieq(strfacility, "mail")) {
-    return LOG_MAIL;
-  } else if(util::strieq(strfacility, "news")) {
-    return LOG_NEWS;
-  } else if(util::strieq(strfacility, "syslog")) {
-    return LOG_SYSLOG;
-  } else if(util::strieq(strfacility, "user")) {
-    return LOG_USER;
-  } else if(util::strieq(strfacility, "uucp")) {
-    return LOG_UUCP;
-  } else {
-    return -1;
   }
+
+  if(util::strieq(strfacility, "authpriv")) {
+    return LOG_AUTHPRIV;
+  }
+
+  if(util::strieq(strfacility, "cron")) {
+    return LOG_CRON;
+  }
+
+  if(util::strieq(strfacility, "daemon")) {
+    return LOG_DAEMON;
+  }
+
+  if(util::strieq(strfacility, "ftp")) {
+    return LOG_FTP;
+  }
+
+  if(util::strieq(strfacility, "kern")) {
+    return LOG_KERN;
+  }
+
+  if(util::strieq(strfacility, "local0")) {
+    return LOG_LOCAL0;
+  }
+
+  if(util::strieq(strfacility, "local1")) {
+    return LOG_LOCAL1;
+  }
+
+  if(util::strieq(strfacility, "local2")) {
+    return LOG_LOCAL2;
+  }
+
+  if(util::strieq(strfacility, "local3")) {
+    return LOG_LOCAL3;
+  }
+
+  if(util::strieq(strfacility, "local4")) {
+    return LOG_LOCAL4;
+  }
+
+  if(util::strieq(strfacility, "local5")) {
+    return LOG_LOCAL5;
+  }
+
+  if(util::strieq(strfacility, "local6")) {
+    return LOG_LOCAL6;
+  }
+
+  if(util::strieq(strfacility, "local7")) {
+    return LOG_LOCAL7;
+  }
+
+  if(util::strieq(strfacility, "lpr")) {
+    return LOG_LPR;
+  }
+
+  if(util::strieq(strfacility, "mail")) {
+    return LOG_MAIL;
+  }
+
+  if(util::strieq(strfacility, "news")) {
+    return LOG_NEWS;
+  }
+
+  if(util::strieq(strfacility, "syslog")) {
+    return LOG_SYSLOG;
+  }
+
+  if(util::strieq(strfacility, "user")) {
+    return LOG_USER;
+  }
+
+  if(util::strieq(strfacility, "uucp")) {
+    return LOG_UUCP;
+  }
+
+  return -1;
 }
 
 } // namespace shrpx
