@@ -39,6 +39,7 @@
 #include "shrpx_error.h"
 #include "shrpx_http.h"
 #include "shrpx_http2_session.h"
+#include "shrpx_worker_config.h"
 #include "http2.h"
 #include "util.h"
 
@@ -47,9 +48,9 @@ using namespace nghttp2;
 namespace shrpx {
 
 Http2DownstreamConnection::Http2DownstreamConnection
-(ClientHandler *client_handler)
-  : DownstreamConnection(client_handler),
-    http2session_(client_handler->get_http2_session()),
+(DownstreamConnectionPool *dconn_pool, Http2Session *http2session)
+  : DownstreamConnection(dconn_pool),
+    http2session_(http2session),
     request_body_buf_(nullptr),
     sd_(nullptr)
 {}
@@ -63,15 +64,27 @@ Http2DownstreamConnection::~Http2DownstreamConnection()
     evbuffer_free(request_body_buf_);
   }
   if(downstream_) {
+    downstream_->disable_downstream_rtimer();
+    downstream_->disable_downstream_wtimer();
+
     if(submit_rst_stream(downstream_) == 0) {
       http2session_->notify();
     }
+
+    if(downstream_->get_downstream_stream_id() != -1) {
+      http2session_->consume(downstream_->get_downstream_stream_id(),
+                             downstream_->get_response_datalen());
+
+      downstream_->reset_response_datalen();
+
+      http2session_->notify();
+    }
   }
   http2session_->remove_downstream_connection(this);
   // Downstream and DownstreamConnection may be deleted
   // asynchronously.
   if(downstream_) {
-    downstream_->set_downstream_connection(nullptr);
+    downstream_->release_downstream_connection();
   }
   if(LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Deleted";
@@ -108,8 +121,11 @@ int Http2DownstreamConnection::attach_downstream(Downstream *downstream)
   if(http2session_->get_state() == Http2Session::DISCONNECTED) {
     http2session_->notify();
   }
-  downstream->set_downstream_connection(this);
+
   downstream_ = downstream;
+
+  downstream_->init_downstream_timer();
+
   return 0;
 }
 
@@ -121,13 +137,23 @@ void Http2DownstreamConnection::detach_downstream(Downstream *downstream)
   if(submit_rst_stream(downstream) == 0) {
     http2session_->notify();
   }
-  downstream->set_downstream_connection(nullptr);
-  downstream_ = nullptr;
 
-  client_handler_->pool_downstream_connection(this);
+  if(downstream_->get_downstream_stream_id() != -1) {
+    http2session_->consume(downstream_->get_downstream_stream_id(),
+                           downstream_->get_response_datalen());
+
+    downstream_->reset_response_datalen();
+
+    http2session_->notify();
+  }
+
+  downstream->disable_downstream_rtimer();
+  downstream->disable_downstream_wtimer();
+  downstream_ = nullptr;
 }
 
-int Http2DownstreamConnection::submit_rst_stream(Downstream *downstream)
+int Http2DownstreamConnection::submit_rst_stream(Downstream *downstream,
+                                                 uint32_t error_code)
 {
   int rv = -1;
   if(http2session_->get_state() == Http2Session::CONNECTED &&
@@ -143,8 +169,7 @@ int Http2DownstreamConnection::submit_rst_stream(Downstream *downstream)
                           << downstream->get_downstream_stream_id();
       }
       rv = http2session_->submit_rst_stream
-        (downstream->get_downstream_stream_id(),
-         NGHTTP2_INTERNAL_ERROR);
+        (downstream->get_downstream_stream_id(), error_code);
     }
   }
   return rv;
@@ -178,50 +203,57 @@ ssize_t http2_data_read_callback(nghttp2_session *session,
       DCLOG(FATAL, dconn) << "evbuffer_remove() failed";
       return NGHTTP2_ERR_CALLBACK_FAILURE;
     }
-    if(nread == 0) {
-      if(downstream->get_request_state() == Downstream::MSG_COMPLETE) {
-        if(!downstream->get_upgrade_request() ||
-           (downstream->get_response_state() == Downstream::HEADER_COMPLETE &&
-            !downstream->get_upgraded())) {
-          *data_flags |= NGHTTP2_DATA_FLAG_EOF;
-        } else {
-          return NGHTTP2_ERR_DEFERRED;
-        }
-        break;
-      } else {
-        // This is important because it will handle flow control
-        // stuff.
-        if(downstream->get_upstream()->resume_read(SHRPX_NO_BUFFER,
-                                                   downstream) != 0) {
-          // In this case, downstream may be deleted.
-          return NGHTTP2_ERR_CALLBACK_FAILURE;
-        }
-        // Check dconn is still alive because Upstream::resume_read()
-        // may delete downstream which will delete dconn.
-        if(sd->dconn == nullptr) {
-          return NGHTTP2_ERR_DEFERRED;
-        }
-        if(evbuffer_get_length(body) == 0) {
-          // Check get_request_state() == MSG_COMPLETE just in case
-          if(downstream->get_request_state() == Downstream::MSG_COMPLETE) {
-            *data_flags |= NGHTTP2_DATA_FLAG_EOF;
-            break;
-          }
-          return NGHTTP2_ERR_DEFERRED;
-        }
-      }
-    } else {
-      // Send WINDOW_UPDATE before buffer is empty to avoid delay
-      // because of RTT.
-      if(!downstream->get_output_buffer_full() &&
-         downstream->get_upstream()->resume_read(SHRPX_NO_BUFFER,
-                                                 downstream) == -1) {
+
+    if(nread > 0) {
+      // This is important because it will handle flow control
+      // stuff.
+      if(downstream->get_upstream()->resume_read
+         (SHRPX_NO_BUFFER, downstream, nread) != 0) {
         // In this case, downstream may be deleted.
+        return NGHTTP2_ERR_CALLBACK_FAILURE;
+      }
+
+      // Check dconn is still alive because Upstream::resume_read()
+      // may delete downstream which will delete dconn.
+      if(sd->dconn == nullptr) {
+        return NGHTTP2_ERR_DEFERRED;
+      }
+
+      break;
+    }
+
+    if(downstream->get_request_state() == Downstream::MSG_COMPLETE) {
+      if(!downstream->get_upgrade_request() ||
+         (downstream->get_response_state() == Downstream::HEADER_COMPLETE &&
+          !downstream->get_upgraded())) {
+        *data_flags |= NGHTTP2_DATA_FLAG_EOF;
+      } else {
+        downstream->disable_downstream_wtimer();
+
         return NGHTTP2_ERR_DEFERRED;
       }
       break;
+    } else {
+      if(evbuffer_get_length(body) == 0) {
+        // Check get_request_state() == MSG_COMPLETE just in case
+        if(downstream->get_request_state() == Downstream::MSG_COMPLETE) {
+          *data_flags |= NGHTTP2_DATA_FLAG_EOF;
+          break;
+        }
+
+        downstream->disable_downstream_wtimer();
+
+        return NGHTTP2_ERR_DEFERRED;
+      }
     }
   }
+
+  if(evbuffer_get_length(body) > 0 && !downstream->get_output_buffer_full()) {
+    downstream->reset_downstream_wtimer();
+  } else {
+    downstream->disable_downstream_wtimer();
+  }
+
   return nread;
 }
 } // namespace
@@ -243,7 +275,7 @@ int Http2DownstreamConnection::push_request_headers()
     downstream_->crumble_request_cookie();
   }
   downstream_->normalize_request_headers();
-  downstream_->concat_norm_request_headers();
+
   auto end_headers = std::end(downstream_->get_request_headers());
 
   // 7 means:
@@ -301,7 +333,12 @@ int Http2DownstreamConnection::push_request_headers()
       http2::copy_url_component(path, &u, UF_PATH, url);
       http2::copy_url_component(query, &u, UF_QUERY, url);
       if(path.empty()) {
-        path = "/";
+        if(!authority.empty() &&
+           downstream_->get_request_method() == "OPTIONS") {
+          path = "*";
+        } else {
+          path = "/";
+        }
       }
       if(!query.empty()) {
         path += "?";
@@ -347,11 +384,6 @@ int Http2DownstreamConnection::push_request_headers()
 
   http2::copy_norm_headers_to_nva(nva, downstream_->get_request_headers());
 
-  bool content_length = false;
-  if(downstream_->get_norm_request_header("content-length") != end_headers) {
-    content_length = true;
-  }
-
   bool chunked_encoding = false;
   auto transfer_encoding =
     downstream_->get_norm_request_header("transfer-encoding");
@@ -373,7 +405,8 @@ int Http2DownstreamConnection::push_request_headers()
     nva.push_back(http2::make_nv_ls("x-forwarded-for", (*xff).value));
   }
 
-  if(downstream_->get_request_method() != "CONNECT") {
+  if(!get_config()->http2_proxy && !get_config()->client_proxy &&
+     downstream_->get_request_method() != "CONNECT") {
     // We use same protocol with :scheme header field
     if(scheme.empty()) {
       if(client_handler_->get_ssl()) {
@@ -413,8 +446,12 @@ int Http2DownstreamConnection::push_request_headers()
     DCLOG(INFO, this) << "HTTP request headers\n" << ss.str();
   }
 
+  auto content_length =
+    downstream_->get_norm_request_header("content-length") != end_headers;
+
   if(downstream_->get_request_method() == "CONNECT" ||
-     chunked_encoding || content_length) {
+     chunked_encoding || content_length ||
+     downstream_->get_request_http2_expect_body()) {
     // Request-body is expected.
     nghttp2_data_provider data_prd;
     data_prd.source.ptr = this;
@@ -429,6 +466,9 @@ int Http2DownstreamConnection::push_request_headers()
     DCLOG(FATAL, this) << "nghttp2_submit_request() failed";
     return -1;
   }
+
+  downstream_->reset_downstream_wtimer();
+
   http2session_->notify();
   return 0;
 }
@@ -446,6 +486,9 @@ int Http2DownstreamConnection::push_upload_data_chunk(const uint8_t *data,
     if(rv != 0) {
       return -1;
     }
+
+    downstream_->ensure_downstream_wtimer();
+
     http2session_->notify();
   }
   return 0;
@@ -459,41 +502,44 @@ int Http2DownstreamConnection::end_upload_data()
     if(rv != 0) {
       return -1;
     }
+
+    downstream_->ensure_downstream_wtimer();
+
     http2session_->notify();
   }
   return 0;
 }
 
-int Http2DownstreamConnection::resume_read(IOCtrlReason reason)
+int Http2DownstreamConnection::resume_read
+(IOCtrlReason reason, size_t consumed)
 {
-  int rv1 = 0, rv2 = 0;
-  if(http2session_->get_state() == Http2Session::CONNECTED &&
-     http2session_->get_flow_control()) {
-    int32_t window_size_increment;
-    window_size_increment = http2::determine_window_update_transmission
-      (http2session_->get_session(), 0);
-    if(window_size_increment != -1) {
-      rv1 = http2session_->submit_window_update(nullptr, window_size_increment);
-      if(rv1 == 0) {
-        http2session_->notify();
-      }
-    }
-    if(downstream_ && downstream_->get_downstream_stream_id() != -1) {
-      window_size_increment = http2::determine_window_update_transmission
-        (http2session_->get_session(), downstream_->get_downstream_stream_id());
-      if(window_size_increment != -1) {
-        rv2 = http2session_->submit_window_update(this, window_size_increment);
-        if(rv2 == 0) {
-          http2session_->notify();
-        }
-      }
-    }
-  }
-  if(rv1 == 0 && rv2 == 0) {
+  int rv;
+
+  if(http2session_->get_state() != Http2Session::CONNECTED ||
+     !http2session_->get_flow_control()) {
     return 0;
   }
-  DLOG(WARNING, this) << "Sending WINDOW_UPDATE failed";
-  return -1;
+
+  if(!downstream_ || downstream_->get_downstream_stream_id() == -1) {
+    return 0;
+  }
+
+  if(consumed > 0) {
+    assert(downstream_->get_response_datalen() >= consumed);
+
+    rv = http2session_->consume(downstream_->get_downstream_stream_id(),
+                                consumed);
+
+    if(rv != 0) {
+      return -1;
+    }
+
+    downstream_->dec_response_datalen(consumed);
+
+    http2session_->notify();
+  }
+
+  return 0;
 }
 
 int Http2DownstreamConnection::on_read()
@@ -538,7 +584,7 @@ StreamData* Http2DownstreamConnection::detach_stream_data()
 bool Http2DownstreamConnection::get_output_buffer_full()
 {
   if(request_body_buf_) {
-    return http2session_->get_outbuf_length() +
+    return
       evbuffer_get_length(request_body_buf_) >= Http2Session::OUTBUF_MAX_THRES;
   } else {
     return false;
@@ -564,4 +610,13 @@ int Http2DownstreamConnection::on_priority_change(int32_t pri)
   return 0;
 }
 
+int Http2DownstreamConnection::on_timeout()
+{
+  if(!downstream_) {
+    return 0;
+  }
+
+  return submit_rst_stream(downstream_, NGHTTP2_NO_ERROR);
+}
+
 } // namespace shrpx
