@@ -26,6 +26,7 @@
 
 #include <netinet/tcp.h>
 #include <unistd.h>
+
 #include <vector>
 
 #include <openssl/err.h>
@@ -40,8 +41,10 @@
 #include "shrpx_client_handler.h"
 #include "shrpx_ssl.h"
 #include "shrpx_http.h"
+#include "shrpx_worker_config.h"
 #include "http2.h"
 #include "util.h"
+#include "libevent_util.h"
 #include "base64.h"
 
 using namespace nghttp2;
@@ -87,7 +90,7 @@ int Http2Session::disconnect()
   }
   if(bev_) {
     int fd = bufferevent_getfd(bev_);
-    bufferevent_disable(bev_, EV_READ | EV_WRITE);
+    util::bev_disable_unless(bev_, EV_READ | EV_WRITE);
     bufferevent_free(bev_);
     bev_ = nullptr;
     if(fd != -1) {
@@ -191,13 +194,14 @@ int Http2Session::init_notification()
 {
   int rv;
   int sockpair[2];
-  rv = socketpair(AF_UNIX, SOCK_STREAM, 0, sockpair);
+  rv = socketpair(AF_UNIX, SOCK_STREAM | SOCK_NONBLOCK | SOCK_CLOEXEC, 0,
+                  sockpair);
   if(rv == -1) {
-    SSLOG(FATAL, this) << "socketpair() failed: errno=" << errno;
+    auto error = errno;
+    SSLOG(FATAL, this) << "socketpair() failed: errno=" << error;
     return -1;
   }
-  evutil_make_socket_nonblocking(sockpair[0]);
-  evutil_make_socket_nonblocking(sockpair[1]);
+
   wrbev_ = bufferevent_socket_new(evbase_, sockpair[0],
                                   BEV_OPT_CLOSE_ON_FREE|
                                   BEV_OPT_DEFER_CALLBACKS);
@@ -216,7 +220,7 @@ int Http2Session::init_notification()
     close(sockpair[1]);
     return -1;
   }
-  bufferevent_enable(rdbev_, EV_READ);
+  util::bev_enable_unless(rdbev_, EV_READ);
   bufferevent_setcb(rdbev_, notify_readcb, nullptr, notify_eventcb, this);
   return 0;
 }
@@ -226,6 +230,7 @@ void readcb(bufferevent *bev, void *ptr)
 {
   int rv;
   auto http2session = static_cast<Http2Session*>(ptr);
+  http2session->reset_timeouts();
   rv = http2session->on_read();
   if(rv != 0) {
     http2session->disconnect();
@@ -236,11 +241,12 @@ void readcb(bufferevent *bev, void *ptr)
 namespace {
 void writecb(bufferevent *bev, void *ptr)
 {
+  auto http2session = static_cast<Http2Session*>(ptr);
+  http2session->reset_timeouts();
   if(evbuffer_get_length(bufferevent_get_output(bev)) > 0) {
     return;
   }
   int rv;
-  auto http2session = static_cast<Http2Session*>(ptr);
   rv = http2session->on_write();
   if(rv != 0) {
     http2session->disconnect();
@@ -257,14 +263,13 @@ void eventcb(bufferevent *bev, short events, void *ptr)
       SSLOG(INFO, http2session) << "Connection established";
     }
     http2session->set_state(Http2Session::CONNECTED);
-    if(!get_config()->downstream_no_tls) {
-      if(!ssl::check_http2_requirement(http2session->get_ssl()) ||
-         (!get_config()->insecure && http2session->check_cert() != 0)) {
+    if(!get_config()->downstream_no_tls &&
+       !get_config()->insecure &&
+       http2session->check_cert() != 0) {
 
-        http2session->disconnect();
+      http2session->disconnect();
 
-        return;
-      }
+      return;
     }
 
     if(http2session->on_connect() != 0) {
@@ -276,8 +281,9 @@ void eventcb(bufferevent *bev, short events, void *ptr)
     int val = 1;
     if(setsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
                   reinterpret_cast<char*>(&val), sizeof(val)) == -1) {
+      auto error = errno;
       SSLOG(WARNING, http2session)
-        << "Setting option TCP_NODELAY failed: errno=" << errno;
+        << "Setting option TCP_NODELAY failed: errno=" << error;
     }
     return;
   }
@@ -338,15 +344,16 @@ void proxy_eventcb(bufferevent *bev, short events, void *ptr)
       SSLOG(INFO, http2session) << "Connected to the proxy";
     }
     std::string req = "CONNECT ";
-    req += get_config()->downstream_hostport;
+    req += get_config()->downstream_hostport.get();
     req += " HTTP/1.1\r\nHost: ";
-    req += get_config()->downstream_host;
+    req += get_config()->downstream_host.get();
     req += "\r\n";
     if(get_config()->downstream_http_proxy_userinfo) {
       req += "Proxy-Authorization: Basic ";
-      size_t len = strlen(get_config()->downstream_http_proxy_userinfo);
-      req += base64::encode(get_config()->downstream_http_proxy_userinfo,
-                            get_config()->downstream_http_proxy_userinfo+len);
+      size_t len = strlen(get_config()->downstream_http_proxy_userinfo.get());
+      req += base64::encode
+        (get_config()->downstream_http_proxy_userinfo.get(),
+         get_config()->downstream_http_proxy_userinfo.get() + len);
       req += "\r\n";
     }
     req += "\r\n";
@@ -393,15 +400,25 @@ int Http2Session::initiate_connection()
   if(get_config()->downstream_http_proxy_host && state_ == DISCONNECTED) {
     if(LOG_ENABLED(INFO)) {
       SSLOG(INFO, this) << "Connecting to the proxy "
-                        << get_config()->downstream_http_proxy_host << ":"
+                        << get_config()->downstream_http_proxy_host.get()
+                        << ":"
                         << get_config()->downstream_http_proxy_port;
     }
-    bev_ = bufferevent_socket_new(evbase_, -1, BEV_OPT_DEFER_CALLBACKS);
-    if(!bev_) {
-      SSLOG(ERROR, this) << "bufferevent_socket_new() failed";
+
+    auto fd = socket(get_config()->downstream_http_proxy_addr.storage.ss_family,
+                     SOCK_STREAM | SOCK_CLOEXEC, 0);
+
+    if(fd == -1) {
       return SHRPX_ERR_NETWORK;
     }
-    bufferevent_enable(bev_, EV_READ);
+
+    bev_ = bufferevent_socket_new(evbase_, fd, BEV_OPT_DEFER_CALLBACKS);
+    if(!bev_) {
+      SSLOG(ERROR, this) << "bufferevent_socket_new() failed";
+      close(fd);
+      return SHRPX_ERR_NETWORK;
+    }
+    util::bev_enable_unless(bev_, EV_READ);
     bufferevent_set_timeouts(bev_, &get_config()->downstream_read_timeout,
                              &get_config()->downstream_write_timeout);
 
@@ -414,7 +431,8 @@ int Http2Session::initiate_connection()
        get_config()->downstream_http_proxy_addrlen);
     if(rv != 0) {
       SSLOG(ERROR, this) << "Failed to connect to the proxy "
-                         << get_config()->downstream_http_proxy_host << ":"
+                         << get_config()->downstream_http_proxy_host.get()
+                         << ":"
                          << get_config()->downstream_http_proxy_port;
       return SHRPX_ERR_NETWORK;
     }
@@ -442,13 +460,13 @@ int Http2Session::initiate_connection()
 
       const char *sni_name = nullptr;
       if ( get_config()->backend_tls_sni_name ) {
-        sni_name = get_config()->backend_tls_sni_name;
+        sni_name = get_config()->backend_tls_sni_name.get();
       }
       else {
-        sni_name = get_config()->downstream_host;
+        sni_name = get_config()->downstream_host.get();
       }
 
-      if(!util::numeric_host(sni_name)) {
+      if(sni_name && !util::numeric_host(sni_name)) {
         // TLS extensions: SNI. There is no documentation about the return
         // code for this function (actually this is macro wrapping SSL_ctrl
         // at the time of this writing).
@@ -456,6 +474,13 @@ int Http2Session::initiate_connection()
       }
       // If state_ == PROXY_CONNECTED, we has connected to the proxy
       // using fd_ and tunnel has been established.
+      if(state_ == DISCONNECTED) {
+        assert(fd_ == -1);
+
+        fd_ = socket(get_config()->downstream_addr.storage.ss_family,
+                     SOCK_STREAM | SOCK_CLOEXEC, 0);
+      }
+
       bev_ = bufferevent_openssl_socket_new(evbase_, fd_, ssl_,
                                             BUFFEREVENT_SSL_CONNECTING,
                                             BEV_OPT_DEFER_CALLBACKS);
@@ -468,43 +493,52 @@ int Http2Session::initiate_connection()
          // TODO maybe not thread-safe?
          const_cast<sockaddr*>(&get_config()->downstream_addr.sa),
          get_config()->downstream_addrlen);
-    } else if(state_ == DISCONNECTED) {
-      // Without TLS and proxy.
-      bev_ = bufferevent_socket_new(evbase_, -1, BEV_OPT_DEFER_CALLBACKS);
-      if(!bev_) {
-        SSLOG(ERROR, this) << "bufferevent_socket_new() failed";
-        return SHRPX_ERR_NETWORK;
-      }
-      rv = bufferevent_socket_connect
-        (bev_,
-         const_cast<sockaddr*>(&get_config()->downstream_addr.sa),
-         get_config()->downstream_addrlen);
     } else {
-      assert(state_ == PROXY_CONNECTED);
-      // Without TLS but with proxy.
+      if(state_ == DISCONNECTED) {
+        // Without TLS and proxy.
+        assert(fd_ == -1);
+
+        fd_ = socket(get_config()->downstream_addr.storage.ss_family,
+                     SOCK_STREAM | SOCK_CLOEXEC, 0);
+
+        if(fd_ == -1) {
+          return SHRPX_ERR_NETWORK;
+        }
+      }
+
       bev_ = bufferevent_socket_new(evbase_, fd_, BEV_OPT_DEFER_CALLBACKS);
       if(!bev_) {
         SSLOG(ERROR, this) << "bufferevent_socket_new() failed";
         return SHRPX_ERR_NETWORK;
       }
-      // Connection already established.
-      eventcb(bev_, BEV_EVENT_CONNECTED, this);
-      // eventcb() has no return value. Check state_ to whether it was
-      // failed or not.
+
       if(state_ == DISCONNECTED) {
-        return -1;
+        rv = bufferevent_socket_connect
+          (bev_,
+           const_cast<sockaddr*>(&get_config()->downstream_addr.sa),
+           get_config()->downstream_addrlen);
+      } else {
+        // Without TLS but with proxy.
+
+        // Connection already established.
+        eventcb(bev_, BEV_EVENT_CONNECTED, this);
+        // eventcb() has no return value. Check state_ to whether it was
+        // failed or not.
+        if(state_ == DISCONNECTED) {
+          return -1;
+        }
       }
     }
+
     if(rv != 0) {
       return SHRPX_ERR_NETWORK;
     }
 
     bufferevent_setwatermark(bev_, EV_READ, 0, SHRPX_READ_WATERMARK);
-    bufferevent_enable(bev_, EV_READ);
+    util::bev_enable_unless(bev_, EV_READ);
     bufferevent_setcb(bev_, readcb, writecb, eventcb, this);
     // Set timeout for HTTP2 session
-    bufferevent_set_timeouts(bev_, &get_config()->downstream_read_timeout,
-                             &get_config()->downstream_write_timeout);
+    reset_timeouts();
 
     // We have been already connected when no TLS and proxy is used.
     if(state_ != CONNECTED) {
@@ -516,6 +550,7 @@ int Http2Session::initiate_connection()
 
   // Unreachable
   DIE();
+  return 0;
 }
 
 void Http2Session::unwrap_free_bev()
@@ -636,8 +671,7 @@ int Http2Session::submit_request(Http2DownstreamConnection *dconn,
   return 0;
 }
 
-int Http2Session::submit_rst_stream(int32_t stream_id,
-                                   nghttp2_error_code error_code)
+int Http2Session::submit_rst_stream(int32_t stream_id, uint32_t error_code)
 {
   assert(state_ == CONNECTED);
   if(LOG_ENABLED(INFO)) {
@@ -656,27 +690,6 @@ int Http2Session::submit_rst_stream(int32_t stream_id,
   return 0;
 }
 
-int Http2Session::submit_window_update(Http2DownstreamConnection *dconn,
-                                      int32_t amount)
-{
-  assert(state_ == CONNECTED);
-  int rv;
-  int32_t stream_id;
-  if(dconn) {
-    stream_id = dconn->get_downstream()->get_downstream_stream_id();
-  } else {
-    stream_id = 0;
-  }
-  rv = nghttp2_submit_window_update(session_, NGHTTP2_FLAG_NONE,
-                                    stream_id, amount);
-  if(rv < NGHTTP2_ERR_FATAL) {
-    SSLOG(FATAL, this) << "nghttp2_submit_window_update() failed: "
-                       << nghttp2_strerror(rv);
-    return -1;
-  }
-  return 0;
-}
-
 int Http2Session::submit_priority(Http2DownstreamConnection *dconn,
                                   int32_t pri)
 {
@@ -743,10 +756,9 @@ void call_downstream_readcb(Http2Session *http2session, Downstream *downstream)
 
 namespace {
 int on_stream_close_callback
-(nghttp2_session *session, int32_t stream_id, nghttp2_error_code error_code,
+(nghttp2_session *session, int32_t stream_id, uint32_t error_code,
  void *user_data)
 {
-  int rv;
   auto http2session = static_cast<Http2Session*>(user_data);
   if(LOG_ENABLED(INFO)) {
     SSLOG(INFO, http2session) << "Stream stream_id=" << stream_id
@@ -763,11 +775,21 @@ int on_stream_close_callback
   if(dconn) {
     auto downstream = dconn->get_downstream();
     if(downstream && downstream->get_downstream_stream_id() == stream_id) {
-      auto upstream = downstream->get_upstream();
-      if(error_code == NGHTTP2_NO_ERROR) {
+
+      if(downstream->get_upgraded() &&
+         downstream->get_response_state() == Downstream::HEADER_COMPLETE) {
+        // For tunneled connection, we have to submit RST_STREAM to
+        // upstream *after* whole response body is sent. We just set
+        // MSG_COMPLETE here. Upstream will take care of that.
+        if(LOG_ENABLED(INFO)) {
+          SSLOG(INFO, http2session) << "RST_STREAM against tunneled stream "
+                                    << "stream_id="
+                                    << stream_id;
+        }
+        downstream->get_upstream()->on_downstream_body_complete(downstream);
         downstream->set_response_state(Downstream::MSG_COMPLETE);
-        rv = upstream->on_downstream_body_complete(downstream);
-        if(rv != 0) {
+      } else if(error_code == NGHTTP2_NO_ERROR) {
+        if(downstream->get_response_state() != Downstream::MSG_COMPLETE) {
           downstream->set_response_state(Downstream::MSG_RESET);
         }
       } else {
@@ -835,10 +857,7 @@ int on_header_callback(nghttp2_session *session,
                        uint8_t flags,
                        void *user_data)
 {
-  if(frame->hd.type != NGHTTP2_HEADERS ||
-     frame->headers.cat != NGHTTP2_HCAT_RESPONSE) {
-    return 0;
-  }
+  auto http2session = static_cast<Http2Session*>(user_data);
   auto sd = static_cast<StreamData*>
     (nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
   if(!sd || !sd->dconn) {
@@ -848,6 +867,13 @@ int on_header_callback(nghttp2_session *session,
   if(!downstream) {
     return 0;
   }
+
+  if(frame->hd.type != NGHTTP2_HEADERS ||
+     (frame->headers.cat != NGHTTP2_HCAT_RESPONSE &&
+      !downstream->get_expect_final_response())) {
+    return 0;
+  }
+
   if(downstream->get_response_headers_sum() > Downstream::MAX_HEADERS_SUM) {
     if(LOG_ENABLED(INFO)) {
       DLOG(INFO, downstream) << "Too large header block size="
@@ -858,6 +884,17 @@ int on_header_callback(nghttp2_session *session,
   if(!http2::check_nv(name, namelen, value, valuelen)) {
     return 0;
   }
+
+  if(namelen > 0 && name[0] == ':') {
+    if(!downstream->response_pseudo_header_allowed() ||
+       !http2::check_http2_response_pseudo_header(name, namelen)) {
+
+      http2session->submit_rst_stream(frame->hd.stream_id,
+                                      NGHTTP2_PROTOCOL_ERROR);
+      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
+    }
+  }
+
   downstream->split_add_response_header(name, namelen, value, valuelen,
                                         flags & NGHTTP2_NV_FLAG_NO_INDEX);
   return 0;
@@ -899,23 +936,20 @@ int on_begin_headers_callback(nghttp2_session *session,
 
 namespace {
 int on_response_headers(Http2Session *http2session,
+                        Downstream *downstream,
                         nghttp2_session *session,
                         const nghttp2_frame *frame)
 {
   int rv;
-  auto sd = static_cast<StreamData*>
-    (nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
-  if(!sd || !sd->dconn) {
-    return 0;
-  }
-  auto downstream = sd->dconn->get_downstream();
-  if(!downstream) {
-    return 0;
-  }
+
+  auto upstream = downstream->get_upstream();
+
   downstream->normalize_response_headers();
   auto& nva = downstream->get_response_headers();
 
-  if(!http2::check_http2_headers(nva)) {
+  downstream->set_expect_final_response(false);
+
+  if(!http2::check_http2_response_headers(nva)) {
     http2session->submit_rst_stream(frame->hd.stream_id,
                                     NGHTTP2_PROTOCOL_ERROR);
     downstream->set_response_state(Downstream::MSG_RESET);
@@ -924,18 +958,53 @@ int on_response_headers(Http2Session *http2session,
   }
 
   auto status = http2::get_unique_header(nva, ":status");
-  if(!status || http2::value_lws(status)) {
+  int status_code;
+
+  if(!status || http2::value_lws(status) ||
+     (status_code = http2::parse_http_status_code(status->value)) == -1) {
+
     http2session->submit_rst_stream(frame->hd.stream_id,
                                     NGHTTP2_PROTOCOL_ERROR);
     downstream->set_response_state(Downstream::MSG_RESET);
     call_downstream_readcb(http2session, downstream);
+
     return 0;
   }
-  downstream->set_response_http_status(strtoul(status->value.c_str(),
-                                               nullptr, 10));
+
+  downstream->set_response_http_status(status_code);
   downstream->set_response_major(2);
   downstream->set_response_minor(0);
 
+  if(LOG_ENABLED(INFO)) {
+    std::stringstream ss;
+    for(auto& nv : nva) {
+      ss << TTY_HTTP_HD << nv.name << TTY_RST << ": " << nv.value << "\n";
+    }
+    SSLOG(INFO, http2session) << "HTTP response headers. stream_id="
+                              << frame->hd.stream_id
+                              << "\n" << ss.str();
+  }
+
+  if(downstream->get_non_final_response()) {
+
+    if(LOG_ENABLED(INFO)) {
+      SSLOG(INFO, http2session) << "This is non-final response.";
+    }
+
+    downstream->set_expect_final_response(true);
+    rv = upstream->on_downstream_header_complete(downstream);
+
+    // Now Dowstream's response headers are erased.
+
+    if(rv != 0) {
+      http2session->submit_rst_stream(frame->hd.stream_id,
+                                      NGHTTP2_PROTOCOL_ERROR);
+      downstream->set_response_state(Downstream::MSG_RESET);
+    }
+
+    return 0;
+  }
+
   auto content_length = http2::get_header(nva, "content-length");
   if(!content_length && downstream->get_request_method() != "HEAD" &&
      downstream->get_request_method() != "CONNECT") {
@@ -954,30 +1023,20 @@ int on_response_headers(Http2Session *http2session,
         // connection open.  In HTTP2, we are supporsed not to
         // receive transfer-encoding.
         downstream->add_response_header("transfer-encoding", "chunked");
+        downstream->set_chunked_response(true);
       }
     }
   }
 
-  if(LOG_ENABLED(INFO)) {
-    std::stringstream ss;
-    for(auto& nv : nva) {
-      ss << TTY_HTTP_HD << nv.name << TTY_RST << ": " << nv.value << "\n";
-    }
-    SSLOG(INFO, http2session) << "HTTP response headers. stream_id="
-                              << frame->hd.stream_id
-                              << "\n" << ss.str();
-  }
-
-  auto upstream = downstream->get_upstream();
   downstream->set_response_state(Downstream::HEADER_COMPLETE);
   downstream->check_upgrade_fulfilled();
   if(downstream->get_upgraded()) {
     downstream->set_response_connection_close(true);
     // On upgrade sucess, both ends can send data
-    if(upstream->resume_read(SHRPX_MSG_BLOCK, downstream) != 0) {
+    if(upstream->resume_read(SHRPX_MSG_BLOCK, downstream, 0) != 0) {
       // If resume_read fails, just drop connection. Not ideal.
       delete upstream->get_client_handler();
-      return 0;
+      return -1;
     }
     downstream->set_request_state(Downstream::HEADER_COMPLETE);
     if(LOG_ENABLED(INFO)) {
@@ -995,7 +1054,7 @@ int on_response_headers(Http2Session *http2session,
                                     NGHTTP2_PROTOCOL_ERROR);
     downstream->set_response_state(Downstream::MSG_RESET);
   }
-  call_downstream_readcb(http2session, downstream);
+
   return 0;
 }
 } // namespace
@@ -1026,12 +1085,84 @@ int on_frame_recv_callback
       http2session->submit_rst_stream(frame->hd.stream_id,
                                       NGHTTP2_INTERNAL_ERROR);
       downstream->set_response_state(Downstream::MSG_RESET);
+
+    } else if(frame->hd.flags & NGHTTP2_FLAG_END_STREAM) {
+
+      downstream->disable_downstream_rtimer();
+
+      if(downstream->get_response_state() == Downstream::HEADER_COMPLETE) {
+
+        downstream->set_response_state(Downstream::MSG_COMPLETE);
+
+        rv = upstream->on_downstream_body_complete(downstream);
+
+        if(rv != 0) {
+          downstream->set_response_state(Downstream::MSG_RESET);
+        }
+      }
     }
+
     call_downstream_readcb(http2session, downstream);
     break;
   }
-  case NGHTTP2_HEADERS:
-    return on_response_headers(http2session, session, frame);
+  case NGHTTP2_HEADERS: {
+    auto sd = static_cast<StreamData*>
+      (nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
+    if(!sd || !sd->dconn) {
+      break;
+    }
+    auto downstream = sd->dconn->get_downstream();
+
+    if(!downstream) {
+      return 0;
+    }
+
+    if(frame->headers.cat == NGHTTP2_HCAT_RESPONSE) {
+      rv = on_response_headers(http2session, downstream, session, frame);
+
+      if(rv != 0) {
+        return 0;
+      }
+    }
+
+    if(frame->headers.cat == NGHTTP2_HCAT_HEADERS) {
+      if(downstream->get_expect_final_response()) {
+        rv = on_response_headers(http2session, downstream, session, frame);
+
+        if(rv != 0) {
+          return 0;
+        }
+      } else if((frame->hd.flags & NGHTTP2_FLAG_END_STREAM) == 0) {
+        http2session->submit_rst_stream(frame->hd.stream_id,
+                                        NGHTTP2_PROTOCOL_ERROR);
+        return 0;
+      }
+    }
+
+    if(frame->hd.flags & NGHTTP2_FLAG_END_STREAM) {
+
+      downstream->disable_downstream_rtimer();
+
+      if(downstream->get_response_state() == Downstream::HEADER_COMPLETE) {
+        downstream->set_response_state(Downstream::MSG_COMPLETE);
+
+        auto upstream = downstream->get_upstream();
+
+        rv = upstream->on_downstream_body_complete(downstream);
+
+        if(rv != 0) {
+          downstream->set_response_state(Downstream::MSG_RESET);
+        }
+      }
+    } else {
+      downstream->reset_downstream_rtimer();
+    }
+
+    // This may delete downstream
+    call_downstream_readcb(http2session, downstream);
+
+    break;
+  }
   case NGHTTP2_RST_STREAM: {
     auto sd = static_cast<StreamData*>
       (nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
@@ -1039,23 +1170,7 @@ int on_frame_recv_callback
       auto downstream = sd->dconn->get_downstream();
       if(downstream &&
          downstream->get_downstream_stream_id() == frame->hd.stream_id) {
-        if(downstream->get_upgraded() &&
-           downstream->get_response_state() == Downstream::HEADER_COMPLETE) {
-          // For tunneled connection, we has to submit RST_STREAM to
-          // upstream *after* whole response body is sent. We just set
-          // MSG_COMPLETE here. Upstream will take care of that.
-          if(LOG_ENABLED(INFO)) {
-            SSLOG(INFO, http2session) << "RST_STREAM against tunneled stream "
-                                      << "stream_id="
-                                      << frame->hd.stream_id;
-          }
-          downstream->get_upstream()->on_downstream_body_complete(downstream);
-          downstream->set_response_state(Downstream::MSG_COMPLETE);
-        } else {
-          // If we got RST_STREAM, just flag MSG_RESET to indicate
-          // upstream connection must be terminated.
-          downstream->set_response_state(Downstream::MSG_RESET);
-        }
+
         downstream->set_response_rst_stream_error_code
           (frame->rst_stream.error_code);
         call_downstream_readcb(http2session, downstream);
@@ -1098,20 +1213,56 @@ int on_data_chunk_recv_callback(nghttp2_session *session,
     (nghttp2_session_get_stream_user_data(session, stream_id));
   if(!sd || !sd->dconn) {
     http2session->submit_rst_stream(stream_id, NGHTTP2_INTERNAL_ERROR);
+
+    if(http2session->consume(stream_id, len) != 0) {
+      return NGHTTP2_ERR_CALLBACK_FAILURE;
+    }
+
     return 0;
   }
   auto downstream = sd->dconn->get_downstream();
-  if(!downstream || downstream->get_downstream_stream_id() != stream_id) {
+  if(!downstream || downstream->get_downstream_stream_id() != stream_id ||
+     !downstream->expect_response_body()) {
+
     http2session->submit_rst_stream(stream_id, NGHTTP2_INTERNAL_ERROR);
+
+    if(http2session->consume(stream_id, len) != 0) {
+      return NGHTTP2_ERR_CALLBACK_FAILURE;
+    }
+
     return 0;
   }
 
+  // We don't want DATA after non-final response, which is illegal in
+  // HTTP.
+  if(downstream->get_non_final_response()) {
+    http2session->submit_rst_stream(stream_id, NGHTTP2_PROTOCOL_ERROR);
+
+    if(http2session->consume(stream_id, len) != 0) {
+      return NGHTTP2_ERR_CALLBACK_FAILURE;
+    }
+
+    return 0;
+  }
+
+  downstream->reset_downstream_rtimer();
+
+  downstream->add_response_bodylen(len);
+
   auto upstream = downstream->get_upstream();
   rv = upstream->on_downstream_body(downstream, data, len, false);
   if(rv != 0) {
     http2session->submit_rst_stream(stream_id, NGHTTP2_INTERNAL_ERROR);
+
+    if(http2session->consume(stream_id, len) != 0) {
+      return NGHTTP2_ERR_CALLBACK_FAILURE;
+    }
+
     downstream->set_response_state(Downstream::MSG_RESET);
   }
+
+  downstream->add_response_datalen(len);
+
   call_downstream_readcb(http2session, downstream);
   return 0;
 }
@@ -1122,6 +1273,31 @@ int on_frame_send_callback(nghttp2_session* session,
                            const nghttp2_frame *frame, void *user_data)
 {
   auto http2session = static_cast<Http2Session*>(user_data);
+
+  if(frame->hd.type == NGHTTP2_DATA || frame->hd.type == NGHTTP2_HEADERS) {
+    if((frame->hd.flags & NGHTTP2_FLAG_END_STREAM) == 0) {
+      return 0;
+    }
+
+    auto sd = static_cast<StreamData*>
+      (nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
+
+    if(!sd || !sd->dconn) {
+      return 0;
+    }
+
+    auto downstream = sd->dconn->get_downstream();
+
+    if(!downstream ||
+       downstream->get_downstream_stream_id() != frame->hd.stream_id) {
+      return 0;
+    }
+
+    downstream->reset_downstream_rtimer();
+
+    return 0;
+  }
+
   if(frame->hd.type == NGHTTP2_SETTINGS &&
      (frame->hd.flags & NGHTTP2_FLAG_ACK) == 0) {
     if(http2session->start_settings_timer() != 0) {
@@ -1166,20 +1342,6 @@ int on_frame_not_send_callback(nghttp2_session *session,
 }
 } // namespace
 
-namespace {
-int on_unknown_frame_recv_callback(nghttp2_session *session,
-                                   const uint8_t *head, size_t headlen,
-                                   const uint8_t *payload, size_t payloadlen,
-                                   void *user_data)
-{
-  auto http2session = static_cast<Http2Session*>(user_data);
-  if(LOG_ENABLED(INFO)) {
-    SSLOG(INFO, http2session) << "Received unknown control frame";
-  }
-  return 0;
-}
-} // namespace
-
 int Http2Session::on_connect()
 {
   int rv;
@@ -1210,21 +1372,43 @@ int Http2Session::on_connect()
       return -1;
     }
   }
-  nghttp2_session_callbacks callbacks;
-  memset(&callbacks, 0, sizeof(callbacks));
-  callbacks.on_stream_close_callback = on_stream_close_callback;
-  callbacks.on_frame_recv_callback = on_frame_recv_callback;
-  callbacks.on_data_chunk_recv_callback = on_data_chunk_recv_callback;
-  callbacks.on_frame_send_callback = on_frame_send_callback;
-  callbacks.on_frame_not_send_callback = on_frame_not_send_callback;
-  callbacks.on_unknown_frame_recv_callback = on_unknown_frame_recv_callback;
-  callbacks.on_header_callback = on_header_callback;
-  callbacks.on_begin_headers_callback = on_begin_headers_callback;
-  if(get_config()->padding) {
-    callbacks.select_padding_callback = http::select_padding_callback;
+  nghttp2_session_callbacks *callbacks;
+  rv = nghttp2_session_callbacks_new(&callbacks);
+
+  if(rv != 0) {
+    return -1;
   }
 
-  rv = nghttp2_session_client_new2(&session_, &callbacks, this,
+  auto callbacks_deleter =
+    util::defer(callbacks, nghttp2_session_callbacks_del);
+
+  nghttp2_session_callbacks_set_on_stream_close_callback
+    (callbacks, on_stream_close_callback);
+
+  nghttp2_session_callbacks_set_on_frame_recv_callback
+    (callbacks, on_frame_recv_callback);
+
+  nghttp2_session_callbacks_set_on_data_chunk_recv_callback
+    (callbacks, on_data_chunk_recv_callback);
+
+  nghttp2_session_callbacks_set_on_frame_send_callback
+    (callbacks, on_frame_send_callback);
+
+  nghttp2_session_callbacks_set_on_frame_not_send_callback
+    (callbacks, on_frame_not_send_callback);
+
+  nghttp2_session_callbacks_set_on_header_callback
+    (callbacks, on_header_callback);
+
+  nghttp2_session_callbacks_set_on_begin_headers_callback
+    (callbacks, on_begin_headers_callback);
+
+  if(get_config()->padding) {
+    nghttp2_session_callbacks_set_select_padding_callback
+      (callbacks, http::select_padding_callback);
+  }
+
+  rv = nghttp2_session_client_new2(&session_, callbacks, this,
                                    get_config()->http2_option);
 
   if(rv != 0) {
@@ -1243,7 +1427,7 @@ int Http2Session::on_connect()
   entry[2].value = (1 << get_config()->http2_downstream_window_bits) - 1;
 
   rv = nghttp2_submit_settings(session_, NGHTTP2_FLAG_NONE, entry,
-                               sizeof(entry)/sizeof(nghttp2_settings_entry));
+                               util::array_size(entry));
   if(rv != 0) {
     return -1;
   }
@@ -1265,16 +1449,49 @@ int Http2Session::on_connect()
     return -1;
   }
 
+  if(!get_config()->downstream_no_tls &&
+     !ssl::check_http2_requirement(ssl_)) {
+
+    rv = terminate_session(NGHTTP2_INADEQUATE_SECURITY);
+
+    if(rv != 0) {
+      return -1;
+    }
+  }
+
   rv = send();
   if(rv != 0) {
     return -1;
   }
 
+  if(!get_config()->downstream_no_tls &&
+     !ssl::check_http2_requirement(ssl_)) {
+
+    return 0;
+  }
+
   // submit pending request
   for(auto dconn : dconns_) {
-    if(dconn->push_request_headers() != 0) {
-      return -1;
+    if(dconn->push_request_headers() == 0) {
+      auto downstream = dconn->get_downstream();
+      auto upstream = downstream->get_upstream();
+      upstream->resume_read(SHRPX_NO_BUFFER, downstream, 0);
+      continue;
     }
+
+    if(LOG_ENABLED(INFO)) {
+      SSLOG(INFO, this) << "backend request failed";
+    }
+
+    auto downstream = dconn->get_downstream();
+
+    if(!downstream) {
+      continue;
+    }
+
+    auto upstream = downstream->get_upstream();
+
+    upstream->on_downstream_abort_request(downstream, 400);
   }
   return 0;
 }
@@ -1397,7 +1614,7 @@ void Http2Session::set_state(int state)
   state_ = state;
 }
 
-int Http2Session::terminate_session(nghttp2_error_code error_code)
+int Http2Session::terminate_session(uint32_t error_code)
 {
   int rv;
   rv = nghttp2_session_terminate_session(session_, error_code);
@@ -1421,4 +1638,30 @@ SSL* Http2Session::get_ssl() const
   return ssl_;
 }
 
+int Http2Session::consume(int32_t stream_id, size_t len)
+{
+  int rv;
+
+  if(!session_) {
+    return 0;
+  }
+
+  rv = nghttp2_session_consume(session_, stream_id, len);
+
+  if(rv != 0) {
+    SSLOG(WARNING, this) << "nghttp2_session_consume() returned error: "
+                         << nghttp2_strerror(rv);
+
+    return -1;
+  }
+
+  return 0;
+}
+
+void Http2Session::reset_timeouts()
+{
+  bufferevent_set_timeouts(bev_, &get_config()->downstream_read_timeout,
+                           &get_config()->downstream_write_timeout);
+}
+
 } // namespace shrpx
