@@ -30,8 +30,12 @@
 #include "shrpx_config.h"
 #include "shrpx_error.h"
 #include "shrpx_http.h"
+#include "shrpx_worker_config.h"
+#include "shrpx_connect_blocker.h"
+#include "shrpx_downstream_connection_pool.h"
 #include "http2.h"
 #include "util.h"
+#include "libevent_util.h"
 
 using namespace nghttp2;
 
@@ -41,15 +45,9 @@ namespace {
 const size_t OUTBUF_MAX_THRES = 64*1024;
 } // namespace
 
-// Workaround for the inability for Bufferevent to remove timeout from
-// bufferevent. Specify this long timeout instead of removing.
-namespace {
-timeval max_timeout = { 86400, 0 };
-} // namespace
-
 HttpDownstreamConnection::HttpDownstreamConnection
-(ClientHandler *client_handler)
-  : DownstreamConnection(client_handler),
+(DownstreamConnectionPool *dconn_pool)
+  : DownstreamConnection(dconn_pool),
     bev_(nullptr),
     ioctrl_(nullptr),
     response_htp_{0}
@@ -58,13 +56,13 @@ HttpDownstreamConnection::HttpDownstreamConnection
 HttpDownstreamConnection::~HttpDownstreamConnection()
 {
   if(bev_) {
-    bufferevent_disable(bev_, EV_READ | EV_WRITE);
+    util::bev_disable_unless(bev_, EV_READ | EV_WRITE);
     bufferevent_free(bev_);
   }
   // Downstream and DownstreamConnection may be deleted
   // asynchronously.
   if(downstream_) {
-    downstream_->set_downstream_connection(nullptr);
+    downstream_->release_downstream_connection();
   }
 }
 
@@ -75,12 +73,32 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream)
   }
   auto upstream = downstream->get_upstream();
   if(!bev_) {
+    auto connect_blocker = client_handler_->get_http1_connect_blocker();
+
+    if(connect_blocker->blocked()) {
+      return -1;
+    }
+
     auto evbase = client_handler_->get_evbase();
+
+    auto fd = socket(get_config()->downstream_addr.storage.ss_family,
+                     SOCK_STREAM | SOCK_CLOEXEC, 0);
+
+    if(fd == -1) {
+      connect_blocker->on_failure();
+
+      return SHRPX_ERR_NETWORK;
+    }
+
     bev_ = bufferevent_socket_new
-      (evbase, -1,
+      (evbase, fd,
        BEV_OPT_CLOSE_ON_FREE | BEV_OPT_DEFER_CALLBACKS);
     if(!bev_) {
+      connect_blocker->on_failure();
+
       DCLOG(INFO, this) << "bufferevent_socket_new() failed";
+      close(fd);
+
       return SHRPX_ERR_NETWORK;
     }
     int rv = bufferevent_socket_connect
@@ -89,15 +107,20 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream)
        const_cast<sockaddr*>(&get_config()->downstream_addr.sa),
        get_config()->downstream_addrlen);
     if(rv != 0) {
+      connect_blocker->on_failure();
+
       bufferevent_free(bev_);
       bev_ = nullptr;
       return SHRPX_ERR_NETWORK;
     }
+
+    connect_blocker->on_success();
+
     if(LOG_ENABLED(INFO)) {
       DCLOG(INFO, this) << "Connecting to downstream server";
     }
   }
-  downstream->set_downstream_connection(this);
+
   downstream_ = downstream;
 
   ioctrl_.set_bev(bev_);
@@ -106,16 +129,14 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream)
   response_htp_.data = downstream_;
 
   bufferevent_setwatermark(bev_, EV_READ, 0, SHRPX_READ_WATERMARK);
-  bufferevent_enable(bev_, EV_READ);
+  util::bev_enable_unless(bev_, EV_READ);
   bufferevent_setcb(bev_,
                     upstream->get_downstream_readcb(),
                     upstream->get_downstream_writecb(),
                     upstream->get_downstream_eventcb(), this);
-  // HTTP request/response model, we first issue request to downstream
-  // server, so just enable write timeout here.
-  bufferevent_set_timeouts(bev_,
-                           &max_timeout,
-                           &get_config()->downstream_write_timeout);
+
+  reset_timeouts();
+
   return 0;
 }
 
@@ -134,15 +155,25 @@ int HttpDownstreamConnection::push_request_headers()
       hdrs += downstream_->get_request_path();
     }
   } else if(get_config()->http2_proxy &&
-     !downstream_->get_request_http2_scheme().empty() &&
-     !downstream_->get_request_http2_authority().empty() &&
-     downstream_->get_request_path().c_str()[0] == '/') {
+            !downstream_->get_request_http2_scheme().empty() &&
+            !downstream_->get_request_http2_authority().empty() &&
+            (downstream_->get_request_path().c_str()[0] == '/' ||
+             downstream_->get_request_path() == "*")) {
     // Construct absolute-form request target because we are going to
     // send a request to a HTTP/1 proxy.
     hdrs += downstream_->get_request_http2_scheme();
     hdrs += "://";
     hdrs += downstream_->get_request_http2_authority();
-    hdrs += downstream_->get_request_path();
+
+    // Server-wide OPTIONS takes following form in proxy request:
+    //
+    // OPTIONS http://example.org HTTP/1.1
+    //
+    // Notice that no slash after authority. See
+    // http://tools.ietf.org/html/rfc7230#section-5.3.4
+    if(downstream_->get_request_path() != "*") {
+      hdrs += downstream_->get_request_path();
+    }
   } else {
     // No proxy case. get_request_path() may be absolute-form but we
     // don't care.
@@ -164,6 +195,14 @@ int HttpDownstreamConnection::push_request_headers()
     hdrs += "\r\n";
   }
 
+  if(downstream_->get_request_method() != "CONNECT" &&
+     downstream_->get_request_http2_expect_body() &&
+     downstream_->get_norm_request_header("content-length") == end_headers) {
+
+    downstream_->set_chunked_request(true);
+    hdrs += "Transfer-Encoding: chunked\r\n";
+  }
+
   if(downstream_->get_request_connection_close()) {
     hdrs += "Connection: close\r\n";
   }
@@ -183,7 +222,8 @@ int HttpDownstreamConnection::push_request_headers()
     http2::sanitize_header_value(hdrs, hdrs.size() - (*xff).value.size());
     hdrs += "\r\n";
   }
-  if(downstream_->get_request_method() != "CONNECT") {
+  if(!get_config()->http2_proxy && !get_config()->client_proxy &&
+     downstream_->get_request_method() != "CONNECT") {
     hdrs += "X-Forwarded-Proto: ";
     if(!downstream_->get_request_http2_scheme().empty()) {
       hdrs += downstream_->get_request_http2_scheme();
@@ -226,7 +266,7 @@ int HttpDownstreamConnection::push_request_headers()
   if(LOG_ENABLED(INFO)) {
     const char *hdrp;
     std::string nhdrs;
-    if(get_config()->tty) {
+    if(worker_config->errorlog_tty) {
       nhdrs = http::colorizeHeaders(hdrs.c_str());
       hdrp = nhdrs.c_str();
     } else {
@@ -242,16 +282,6 @@ int HttpDownstreamConnection::push_request_headers()
     return -1;
   }
 
-  // When downstream request is issued, set read timeout. We don't
-  // know when the request is completely received by the downstream
-  // server. This function may be called before that happens. Overall
-  // it does not cause problem for most of the time.  If the
-  // downstream server is too slow to recv/send, the connection will
-  // be dropped by read timeout.
-  bufferevent_set_timeouts(bev_,
-                           &get_config()->downstream_read_timeout,
-                           &get_config()->downstream_write_timeout);
-
   return 0;
 }
 
@@ -303,6 +333,16 @@ int HttpDownstreamConnection::end_upload_data()
   return 0;
 }
 
+namespace {
+void idle_readcb(bufferevent *bev, void *arg)
+{
+  auto dconn = static_cast<HttpDownstreamConnection*>(arg);
+  auto dconn_pool = dconn->get_dconn_pool();
+  dconn_pool->remove_downstream_connection(dconn);
+  // dconn was deleted
+}
+} // namespace
+
 namespace {
 // Gets called when DownstreamConnection is pooled in ClientHandler.
 void idle_eventcb(bufferevent *bev, short events, void *arg)
@@ -310,13 +350,10 @@ void idle_eventcb(bufferevent *bev, short events, void *arg)
   auto dconn = static_cast<HttpDownstreamConnection*>(arg);
   if(events & BEV_EVENT_CONNECTED) {
     // Downstream was detached before connection established?
-    // This may be safe to be left.
     if(LOG_ENABLED(INFO)) {
       DCLOG(INFO, dconn) << "Idle connection connected?";
     }
-    return;
-  }
-  if(events & BEV_EVENT_EOF) {
+  } else if(events & BEV_EVENT_EOF) {
     if(LOG_ENABLED(INFO)) {
       DCLOG(INFO, dconn) << "Idle connection EOF";
     }
@@ -329,9 +366,9 @@ void idle_eventcb(bufferevent *bev, short events, void *arg)
       DCLOG(INFO, dconn) << "Idle connection network error";
     }
   }
-  auto client_handler = dconn->get_client_handler();
-  client_handler->remove_downstream_connection(dconn);
-  delete dconn;
+  auto dconn_pool = dconn->get_dconn_pool();
+  dconn_pool->remove_downstream_connection(dconn);
+  // dconn was deleted
 }
 } // namespace
 
@@ -340,17 +377,15 @@ void HttpDownstreamConnection::detach_downstream(Downstream *downstream)
   if(LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Detaching from DOWNSTREAM:" << downstream;
   }
-  downstream->set_downstream_connection(0);
-  downstream_ = 0;
+  downstream_ = nullptr;
   ioctrl_.force_resume_read();
-  bufferevent_enable(bev_, EV_READ);
-  bufferevent_setcb(bev_, 0, 0, idle_eventcb, this);
+  util::bev_enable_unless(bev_, EV_READ);
+  bufferevent_setcb(bev_, idle_readcb, nullptr, idle_eventcb, this);
   // On idle state, just enable read timeout. Normally idle downstream
   // connection will get EOF from the downstream server and closed.
   bufferevent_set_timeouts(bev_,
                            &get_config()->downstream_idle_read_timeout,
                            &get_config()->downstream_write_timeout);
-  client_handler_->pool_downstream_connection(this);
 }
 
 bufferevent* HttpDownstreamConnection::get_bev()
@@ -363,7 +398,7 @@ void HttpDownstreamConnection::pause_read(IOCtrlReason reason)
   ioctrl_.pause_read(reason);
 }
 
-int HttpDownstreamConnection::resume_read(IOCtrlReason reason)
+int HttpDownstreamConnection::resume_read(IOCtrlReason reason, size_t consumed)
 {
   ioctrl_.resume_read(reason);
   return 0;
@@ -380,28 +415,57 @@ bool HttpDownstreamConnection::get_output_buffer_full()
   return evbuffer_get_length(output) >= OUTBUF_MAX_THRES;
 }
 
+namespace {
+int htp_msg_begincb(http_parser *htp)
+{
+  auto downstream = static_cast<Downstream*>(htp->data);
+
+  if(downstream->get_response_state() != Downstream::INITIAL) {
+    return -1;
+  }
+
+  return 0;
+}
+} // namespace
+
 namespace {
 int htp_hdrs_completecb(http_parser *htp)
 {
   auto downstream = static_cast<Downstream*>(htp->data);
+  auto upstream = downstream->get_upstream();
+  int rv;
+
   downstream->set_response_http_status(htp->status_code);
   downstream->set_response_major(htp->http_major);
   downstream->set_response_minor(htp->http_minor);
+
+  if(downstream->get_non_final_response()) {
+    // For non-final response code, we just call
+    // on_downstream_header_complete() without changing response
+    // state.
+    rv = upstream->on_downstream_header_complete(downstream);
+
+    if(rv != 0) {
+      return -1;
+    }
+
+    return 0;
+  }
+
   downstream->set_response_connection_close(!http_should_keep_alive(htp));
   downstream->set_response_state(Downstream::HEADER_COMPLETE);
+  downstream->inspect_http1_response();
   downstream->check_upgrade_fulfilled();
   if(downstream->get_upgraded()) {
     downstream->set_response_connection_close(true);
   }
-  if(downstream->get_upstream()->on_downstream_header_complete(downstream)
-     != 0) {
+  if(upstream->on_downstream_header_complete(downstream) != 0) {
     return -1;
   }
 
   if(downstream->get_upgraded()) {
     // Upgrade complete, read until EOF in both ends
-    if(downstream->get_upstream()->resume_read(SHRPX_MSG_BLOCK,
-                                               downstream) != 0) {
+    if(upstream->resume_read(SHRPX_MSG_BLOCK, downstream, 0) != 0) {
       return -1;
     }
     downstream->set_request_state(Downstream::HEADER_COMPLETE);
@@ -418,6 +482,9 @@ int htp_hdrs_completecb(http_parser *htp)
   // 304 status code with nonzero Content-Length, but without response
   // body. See
   // http://tools.ietf.org/html/draft-ietf-httpbis-p1-messaging-20#section-3.3
+
+  // TODO It seems that the cases other than HEAD are handled by
+  // http-parser.  Need test.
   return downstream->get_request_method() == "HEAD" ||
     (100 <= status && status <= 199) || status == 204 ||
     status == 304 ? 1 : 0;
@@ -468,6 +535,9 @@ namespace {
 int htp_bodycb(http_parser *htp, const char *data, size_t len)
 {
   auto downstream = static_cast<Downstream*>(htp->data);
+
+  downstream->add_response_bodylen(len);
+
   return downstream->get_upstream()->on_downstream_body
     (downstream, reinterpret_cast<const uint8_t*>(data), len, true);
 }
@@ -477,6 +547,13 @@ namespace {
 int htp_msg_completecb(http_parser *htp)
 {
   auto downstream = static_cast<Downstream*>(htp->data);
+
+  if(downstream->get_non_final_response()) {
+    downstream->reset_response();
+
+    return 0;
+  }
+
   downstream->set_response_state(Downstream::MSG_COMPLETE);
   // Block reading another response message from (broken?)
   // server. This callback is not called if the connection is
@@ -488,7 +565,7 @@ int htp_msg_completecb(http_parser *htp)
 
 namespace {
 http_parser_settings htp_hooks = {
-  nullptr, // http_cb      on_message_begin;
+  htp_msg_begincb, // http_cb on_message_begin;
   nullptr, // http_data_cb on_url;
   nullptr, // http_data_cb on_status;
   htp_hdr_keycb, // http_data_cb on_header_field;
@@ -501,6 +578,8 @@ http_parser_settings htp_hooks = {
 
 int HttpDownstreamConnection::on_read()
 {
+  reset_timeouts();
+
   auto input = bufferevent_get_input(bev_);
 
   if(downstream_->get_upgraded()) {
@@ -565,6 +644,11 @@ int HttpDownstreamConnection::on_read()
 
 int HttpDownstreamConnection::on_write()
 {
+  reset_timeouts();
+
+  auto upstream = downstream_->get_upstream();
+  upstream->resume_read(SHRPX_NO_BUFFER, downstream_,
+                        downstream_->get_request_datalen());
   return 0;
 }
 
@@ -576,4 +660,11 @@ void HttpDownstreamConnection::on_upstream_change(Upstream *upstream)
                     upstream->get_downstream_eventcb(), this);
 }
 
+void HttpDownstreamConnection::reset_timeouts()
+{
+  bufferevent_set_timeouts(bev_,
+                           &get_config()->downstream_read_timeout,
+                           &get_config()->downstream_write_timeout);
+}
+
 } // namespace shrpx
