@@ -36,7 +36,7 @@
 #include "shrpx_downstream_connection.h"
 #include "shrpx_config.h"
 #include "shrpx_http.h"
-#include "shrpx_accesslog.h"
+#include "shrpx_worker_config.h"
 #include "http2.h"
 #include "util.h"
 
@@ -107,9 +107,14 @@ void on_stream_close_callback
     return;
   }
 
+  upstream->consume(stream_id, downstream->get_request_datalen());
+
+  downstream->reset_request_datalen();
+
   if(downstream->get_request_state() == Downstream::CONNECT_FAIL) {
     upstream->remove_downstream(downstream);
-    delete downstream;
+    // downstrea was deleted
+
     return;
   }
 
@@ -119,13 +124,11 @@ void on_stream_close_callback
     if(!downstream->get_upgraded() &&
        !downstream->get_response_connection_close()) {
       // Keep-alive
-      auto dconn = downstream->get_downstream_connection();
-      if(dconn) {
-        dconn->detach_downstream(downstream);
-      }
+      downstream->detach_downstream_connection();
     }
     upstream->remove_downstream(downstream);
-    delete downstream;
+    // downstrea was deleted
+
     return;
   }
 
@@ -134,7 +137,8 @@ void on_stream_close_callback
   // If shrpx_downstream::push_request_headers() failed, the
   // error is handled here.
   upstream->remove_downstream(downstream);
-  delete downstream;
+  // downstrea was deleted
+
   // How to test this case? Request sufficient large download
   // and make client send RST_STREAM after it gets first DATA
   // frame chunk.
@@ -153,10 +157,12 @@ void on_ctrl_recv_callback
       ULOG(INFO, upstream) << "Received upstream SYN_STREAM stream_id="
                            << frame->syn_stream.stream_id;
     }
-    auto downstream = new Downstream(upstream,
-                                     frame->syn_stream.stream_id,
-                                     frame->syn_stream.pri);
-    upstream->add_downstream(downstream);
+
+    auto downstream = upstream->add_pending_downstream
+      (frame->syn_stream.stream_id, frame->syn_stream.pri);
+
+    downstream->init_upstream_timer();
+    downstream->reset_upstream_rtimer();
     downstream->init_response_body_buf();
 
     auto nv = frame->syn_stream.nv;
@@ -164,7 +170,9 @@ void on_ctrl_recv_callback
     const char *scheme = nullptr;
     const char *host = nullptr;
     const char *method = nullptr;
-    const char *content_length = 0;
+    const char *content_length = nullptr;
+    const char *user_agent = nullptr;
+
     for(size_t i = 0; nv[i]; i += 2) {
       if(strcmp(nv[i], ":path") == 0) {
         path = nv[i+1];
@@ -177,6 +185,8 @@ void on_ctrl_recv_callback
       } else if(nv[i][0] != ':') {
         if(strcmp(nv[i], "content-length") == 0) {
           content_length = nv[i+1];
+        } else if(strcmp(nv[i], "user-agent") == 0) {
+          user_agent = nv[i+1];
         }
         downstream->add_request_header(nv[i], nv[i+1]);
       }
@@ -204,7 +214,15 @@ void on_ctrl_recv_callback
       downstream->set_request_path(path);
     }
 
-    downstream->check_upgrade_request();
+    if(user_agent) {
+      downstream->set_request_user_agent(user_agent);
+    }
+
+    if(!(frame->syn_stream.hd.flags & SPDYLAY_CTRL_FLAG_FIN)) {
+      downstream->set_request_http2_expect_body(true);
+    }
+
+    downstream->inspect_http2_request();
 
     if(LOG_ENABLED(INFO)) {
       std::stringstream ss;
@@ -216,23 +234,14 @@ void on_ctrl_recv_callback
                            << "\n" << ss.str();
     }
 
-    auto dconn = upstream->get_client_handler()->get_downstream_connection();
-    int rv = dconn->attach_downstream(downstream);
-    if(rv != 0) {
-      // If downstream connection fails, issue RST_STREAM.
-      upstream->rst_stream(downstream, SPDYLAY_INTERNAL_ERROR);
-      downstream->set_request_state(Downstream::CONNECT_FAIL);
-      return;
-    }
-    rv = downstream->push_request_headers();
-    if(rv != 0) {
-      upstream->rst_stream(downstream, SPDYLAY_INTERNAL_ERROR);
-      return;
-    }
     downstream->set_request_state(Downstream::HEADER_COMPLETE);
     if(frame->syn_stream.hd.flags & SPDYLAY_CTRL_FLAG_FIN) {
+      downstream->disable_upstream_rtimer();
       downstream->set_request_state(Downstream::MSG_COMPLETE);
     }
+
+    upstream->maintain_downstream_concurrency();
+
     break;
   }
   default:
@@ -241,6 +250,57 @@ void on_ctrl_recv_callback
 }
 } // namespace
 
+void SpdyUpstream::maintain_downstream_concurrency()
+{
+  while(get_config()->max_downstream_connections >
+        downstream_queue_.num_active()) {
+    if(downstream_queue_.pending_empty()) {
+      break;
+    }
+
+    {
+      auto downstream = downstream_queue_.pending_top();
+      if(downstream->get_request_state() != Downstream::HEADER_COMPLETE &&
+         downstream->get_request_state() != Downstream::MSG_COMPLETE) {
+        break;
+      }
+    }
+
+    auto downstream = downstream_queue_.pop_pending();
+
+    if(!downstream) {
+      break;
+    }
+
+    initiate_downstream(std::move(downstream));
+  }
+}
+
+void SpdyUpstream::initiate_downstream(std::unique_ptr<Downstream> downstream)
+{
+  int rv = downstream->attach_downstream_connection
+    (handler_->get_downstream_connection());
+  if(rv != 0) {
+    // If downstream connection fails, issue RST_STREAM.
+    rst_stream(downstream.get(), SPDYLAY_INTERNAL_ERROR);
+    downstream->set_request_state(Downstream::CONNECT_FAIL);
+
+    downstream_queue_.add_failure(std::move(downstream));
+
+    return;
+  }
+  rv = downstream->push_request_headers();
+  if(rv != 0) {
+    rst_stream(downstream.get(), SPDYLAY_INTERNAL_ERROR);
+
+    downstream_queue_.add_failure(std::move(downstream));
+
+    return;
+  }
+
+  downstream_queue_.add_active(std::move(downstream));
+}
+
 namespace {
 void on_data_chunk_recv_callback(spdylay_session *session,
                                  uint8_t flags, int32_t stream_id,
@@ -251,11 +311,18 @@ void on_data_chunk_recv_callback(spdylay_session *session,
   auto downstream = upstream->find_downstream(stream_id);
 
   if(!downstream) {
+    upstream->consume(stream_id, len);
+
     return;
   }
 
+  downstream->reset_upstream_rtimer();
+
   if(downstream->push_upload_data_chunk(data, len) != 0) {
     upstream->rst_stream(downstream, SPDYLAY_INTERNAL_ERROR);
+
+    upstream->consume(stream_id, len);
+
     return;
   }
 
@@ -303,6 +370,7 @@ void on_data_recv_callback(spdylay_session *session, uint8_t flags,
   auto upstream = static_cast<SpdyUpstream*>(user_data);
   auto downstream = upstream->find_downstream(stream_id);
   if(downstream && (flags & SPDYLAY_DATA_FLAG_FIN)) {
+    downstream->disable_upstream_rtimer();
     downstream->end_upload_data();
     downstream->set_request_state(Downstream::MSG_COMPLETE);
   }
@@ -364,8 +432,7 @@ void on_unknown_ctrl_recv_callback(spdylay_session *session,
 namespace {
 // Infer upstream RST_STREAM status code from downstream HTTP/2
 // error code.
-uint32_t infer_upstream_rst_stream_status_code
-(nghttp2_error_code downstream_error_code)
+uint32_t infer_upstream_rst_stream_status_code(uint32_t downstream_error_code)
 {
   // Only propagate *_REFUSED_STREAM so that upstream client can
   // resend request.
@@ -382,8 +449,7 @@ SpdyUpstream::SpdyUpstream(uint16_t version, ClientHandler *handler)
     session_(nullptr)
 {
   //handler->set_bev_cb(spdy_readcb, 0, spdy_eventcb);
-  handler->set_upstream_timeouts(&get_config()->http2_upstream_read_timeout,
-                                 &get_config()->upstream_write_timeout);
+  reset_timeouts();
 
   spdylay_session_callbacks callbacks;
   memset(&callbacks, 0, sizeof(callbacks));
@@ -407,7 +473,7 @@ SpdyUpstream::SpdyUpstream(uint16_t version, ClientHandler *handler)
     flow_control_ = true;
     initial_window_size_ = 1 << get_config()->http2_upstream_window_bits;
     rv = spdylay_session_set_option(session_,
-                                    SPDYLAY_OPT_NO_AUTO_WINDOW_UPDATE, &val,
+                                    SPDYLAY_OPT_NO_AUTO_WINDOW_UPDATE2, &val,
                                     sizeof(val));
     assert(rv == 0);
   } else {
@@ -425,8 +491,7 @@ SpdyUpstream::SpdyUpstream(uint16_t version, ClientHandler *handler)
   entry[1].flags = SPDYLAY_ID_FLAG_SETTINGS_NONE;
 
   rv = spdylay_submit_settings
-    (session_, SPDYLAY_FLAG_SETTINGS_NONE,
-     entry, sizeof(entry)/sizeof(spdylay_settings_entry));
+    (session_, SPDYLAY_FLAG_SETTINGS_NONE, entry, util::array_size(entry));
   assert(rv == 0);
 
   if(version >= SPDYLAY_PROTO_SPDY3_1 &&
@@ -519,7 +584,8 @@ void spdy_downstream_readcb(bufferevent *bev, void *ptr)
     // because there is no consumer now. Downstream connection is also
     // closed in this case.
     upstream->remove_downstream(downstream);
-    delete downstream;
+    // downstrea was deleted
+
     return;
   }
 
@@ -530,8 +596,7 @@ void spdy_downstream_readcb(bufferevent *bev, void *ptr)
     // on_stream_close_callback.
     upstream->rst_stream(downstream, infer_upstream_rst_stream_status_code
                          (downstream->get_response_rst_stream_error_code()));
-    downstream->set_downstream_connection(nullptr);
-    delete dconn;
+    downstream->pop_downstream_connection();
     dconn = nullptr;
   } else {
     auto rv = downstream->on_read();
@@ -551,8 +616,7 @@ void spdy_downstream_readcb(bufferevent *bev, void *ptr)
       downstream->set_response_state(Downstream::MSG_COMPLETE);
       // Clearly, we have to close downstream connection on http parser
       // failure.
-      downstream->set_downstream_connection(nullptr);
-      delete dconn;
+      downstream->pop_downstream_connection();
       dconn = nullptr;
     }
   }
@@ -571,9 +635,7 @@ void spdy_downstream_writecb(bufferevent *bev, void *ptr)
     return;
   }
   auto dconn = static_cast<DownstreamConnection*>(ptr);
-  auto downstream = dconn->get_downstream();
-  auto upstream = static_cast<SpdyUpstream*>(downstream->get_upstream());
-  upstream->resume_read(SHRPX_NO_BUFFER, downstream);
+  dconn->on_write();
 }
 } // namespace
 
@@ -607,14 +669,14 @@ void spdy_downstream_eventcb(bufferevent *bev, short events, void *ptr)
       // If stream was closed already, we don't need to send reply at
       // the first place. We can delete downstream.
       upstream->remove_downstream(downstream);
-      delete downstream;
+      // downstrea was deleted
+
       return;
     }
 
     // Delete downstream connection. If we don't delete it here, it
     // will be pooled in on_stream_close_callback.
-    downstream->set_downstream_connection(nullptr);
-    delete dconn;
+    downstream->pop_downstream_connection();
     dconn = nullptr;
     // downstream wil be deleted in on_stream_close_callback.
     if(downstream->get_response_state() == Downstream::HEADER_COMPLETE) {
@@ -662,14 +724,14 @@ void spdy_downstream_eventcb(bufferevent *bev, short events, void *ptr)
     }
     if(downstream->get_request_state() == Downstream::STREAM_CLOSED) {
       upstream->remove_downstream(downstream);
-      delete downstream;
+      // downstrea was deleted
+
       return;
     }
 
     // Delete downstream connection. If we don't delete it here, it
     // will be pooled in on_stream_close_callback.
-    downstream->set_downstream_connection(nullptr);
-    delete dconn;
+    downstream->pop_downstream_connection();
     dconn = nullptr;
 
     if(downstream->get_response_state() == Downstream::MSG_COMPLETE) {
@@ -723,21 +785,6 @@ int SpdyUpstream::rst_stream(Downstream *downstream, int status_code)
   return 0;
 }
 
-int SpdyUpstream::window_update(Downstream *downstream, int32_t delta)
-{
-  int rv;
-  rv = spdylay_submit_window_update(session_,
-                                    downstream ?
-                                    downstream->get_stream_id() : 0,
-                                    delta);
-  if(rv < SPDYLAY_ERR_FATAL) {
-    ULOG(FATAL, this) << "spdylay_submit_window_update() failed: "
-                      << spdylay_strerror(rv);
-    DIE();
-  }
-  return 0;
-}
-
 namespace {
 ssize_t spdy_data_read_callback(spdylay_session *session,
                                 int32_t stream_id,
@@ -759,6 +806,9 @@ ssize_t spdy_data_read_callback(spdylay_session *session,
      downstream->get_response_state() == Downstream::MSG_COMPLETE) {
     if(!downstream->get_upgraded()) {
       *eof = 1;
+
+      upstream_accesslog(upstream->get_client_handler()->get_ipaddr(),
+                         downstream->get_response_http_status(), downstream);
     } else {
       // For tunneling, issue RST_STREAM to finish the stream.
       if(LOG_ENABLED(INFO)) {
@@ -770,11 +820,17 @@ ssize_t spdy_data_read_callback(spdylay_session *session,
     }
   }
 
-  if(nread == 0 && *eof != 1) {
-    if(downstream->resume_read(SHRPX_NO_BUFFER) != 0) {
-      return SPDYLAY_ERR_CALLBACK_FAILURE;
-    }
+  if(evbuffer_get_length(body) > 0) {
+    downstream->reset_upstream_wtimer();
+  } else {
+    downstream->disable_upstream_wtimer();
+  }
 
+  if(nread > 0 && downstream->resume_read(SHRPX_NO_BUFFER, nread) != 0) {
+    return SPDYLAY_ERR_CALLBACK_FAILURE;
+  }
+
+  if(nread == 0 && *eof != 1) {
     return SPDYLAY_ERR_DEFERRED;
   }
 
@@ -786,6 +842,7 @@ int SpdyUpstream::error_reply(Downstream *downstream, unsigned int status_code)
 {
   int rv;
   auto html = http::create_error_html(status_code);
+  downstream->set_response_http_status(status_code);
   downstream->init_response_body_buf();
   auto body = downstream->get_response_body_buf();
   rv = evbuffer_add(body, html.c_str(), html.size());
@@ -817,10 +874,7 @@ int SpdyUpstream::error_reply(Downstream *downstream, unsigned int status_code)
                       << spdylay_strerror(rv);
     DIE();
   }
-  if(get_config()->accesslog) {
-    upstream_response(get_client_handler()->get_ipaddr(),
-                      status_code, downstream);
-  }
+
   return 0;
 }
 
@@ -839,14 +893,22 @@ bufferevent_event_cb SpdyUpstream::get_downstream_eventcb()
   return spdy_downstream_eventcb;
 }
 
-void SpdyUpstream::add_downstream(Downstream *downstream)
+Downstream* SpdyUpstream::add_pending_downstream
+(int32_t stream_id, int32_t priority)
 {
-  downstream_queue_.add(downstream);
+  auto downstream = util::make_unique<Downstream>(this, stream_id, priority);
+  auto res = downstream.get();
+
+  downstream_queue_.add_pending(std::move(downstream));
+
+  return res;
 }
 
 void SpdyUpstream::remove_downstream(Downstream *downstream)
 {
-  downstream_queue_.remove(downstream);
+  downstream_queue_.remove(downstream->get_stream_id());
+
+  maintain_downstream_concurrency();
 }
 
 Downstream* SpdyUpstream::find_downstream(int32_t stream_id)
@@ -863,18 +925,28 @@ spdylay_session* SpdyUpstream::get_http2_session()
 // spdylay_session_recv. These calls may delete downstream.
 int SpdyUpstream::on_downstream_header_complete(Downstream *downstream)
 {
+  if(downstream->get_non_final_response()) {
+    // SPDY does not support non-final response.  We could send it
+    // with HEADERS and final response in SYN_REPLY, but it is not
+    // official way.
+    downstream->clear_response_headers();
+
+    return 0;
+  }
+
   if(LOG_ENABLED(INFO)) {
     DLOG(INFO, downstream) << "HTTP response header completed";
   }
   downstream->normalize_response_headers();
-  if(!get_config()->http2_proxy && !get_config()->client_proxy) {
+  if(!get_config()->http2_proxy && !get_config()->client_proxy &&
+     !get_config()->no_location_rewrite) {
     downstream->rewrite_norm_location_response_header
       (get_client_handler()->get_upstream_scheme(), get_config()->port);
   }
   size_t nheader = downstream->get_response_headers().size();
-  // 6 means :status, :version and possible via header field.
+  // 8 means server, :status, :version and possible via header field.
   auto nv = util::make_unique<const char*[]>
-    (nheader * 2 + 6 + get_config()->add_response_headers.size() * 2 + 1);
+    (nheader * 2 + 8 + get_config()->add_response_headers.size() * 2 + 1);
 
   size_t hdidx = 0;
   std::string via_value;
@@ -894,11 +966,20 @@ int SpdyUpstream::on_downstream_header_complete(Downstream *downstream)
     } else if(!get_config()->no_via &&
               util::strieq(hd.name.c_str(), "via")) {
       via_value = hd.value;
+    } else if(!get_config()->http2_proxy && !get_config()->client_proxy &&
+              util::strieq(hd.name.c_str(), "server")) {
+      // Rewrite server header field later
     } else {
       nv[hdidx++] = hd.name.c_str();
       nv[hdidx++] = hd.value.c_str();
     }
   }
+
+  if(!get_config()->http2_proxy && !get_config()->client_proxy) {
+    nv[hdidx++] = "server";
+    nv[hdidx++] = get_config()->server_name;
+  }
+
   if(!get_config()->no_via) {
     if(!via_value.empty()) {
       via_value += ", ";
@@ -935,11 +1016,15 @@ int SpdyUpstream::on_downstream_header_complete(Downstream *downstream)
     ULOG(FATAL, this) << "spdylay_submit_response() failed";
     return -1;
   }
-  if(get_config()->accesslog) {
-    upstream_response(get_client_handler()->get_ipaddr(),
-                      downstream->get_response_http_status(),
-                      downstream);
+
+  if(downstream->get_upgraded()) {
+    upstream_accesslog(get_client_handler()->get_ipaddr(),
+                       downstream->get_response_http_status(), downstream);
   }
+
+  downstream->clear_response_headers();
+  downstream->clear_request_headers();
+
   return 0;
 }
 
@@ -958,11 +1043,15 @@ int SpdyUpstream::on_downstream_body(Downstream *downstream,
 
   if(flush) {
     spdylay_session_resume_data(session_, downstream->get_stream_id());
+
+    downstream->ensure_upstream_wtimer();
   }
 
   if(evbuffer_get_length(body) >= INBUF_MAX_THRES) {
     if(!flush) {
       spdylay_session_resume_data(session_, downstream->get_stream_id());
+
+      downstream->ensure_upstream_wtimer();
     }
 
     downstream->pause_read(SHRPX_NO_BUFFER);
@@ -978,7 +1067,10 @@ int SpdyUpstream::on_downstream_body_complete(Downstream *downstream)
   if(LOG_ENABLED(INFO)) {
     DLOG(INFO, downstream) << "HTTP response completed";
   }
+
   spdylay_session_resume_data(session_, downstream->get_stream_id());
+  downstream->ensure_upstream_wtimer();
+
   return 0;
 }
 
@@ -990,41 +1082,68 @@ bool SpdyUpstream::get_flow_control() const
 void SpdyUpstream::pause_read(IOCtrlReason reason)
 {}
 
-namespace {
-int32_t determine_window_update_transmission(spdylay_session *session,
-                                             int32_t stream_id)
-{
-  int32_t recv_length, window_size;
-  if(stream_id == 0) {
-    recv_length = spdylay_session_get_recv_data_length(session);
-    window_size = 1 << get_config()->http2_upstream_connection_window_bits;
-  } else {
-    recv_length = spdylay_session_get_stream_recv_data_length
-      (session, stream_id);
-    window_size = 1 << get_config()->http2_upstream_window_bits;
-  }
-  if(recv_length != -1 && recv_length >= window_size / 2) {
-    return recv_length;
-  }
-  return -1;
-}
-} // namespace
-
-int SpdyUpstream::resume_read(IOCtrlReason reason, Downstream *downstream)
+int SpdyUpstream::resume_read(IOCtrlReason reason, Downstream *downstream,
+                              size_t consumed)
 {
   if(get_flow_control()) {
-    int32_t delta;
-    delta = determine_window_update_transmission(session_, 0);
-    if(delta != -1) {
-      window_update(0, delta);
-    }
-    delta = determine_window_update_transmission
-      (session_, downstream->get_stream_id());
-    if(delta != -1) {
-      window_update(downstream, delta);
+    assert(downstream->get_request_datalen() >= consumed);
+
+    if(consume(downstream->get_stream_id(), consumed) != 0) {
+      return -1;
     }
+
+    downstream->dec_request_datalen(consumed);
   }
+
   return send();
 }
 
+int SpdyUpstream::on_downstream_abort_request(Downstream *downstream,
+                                              unsigned int status_code)
+{
+  int rv;
+
+  rv = error_reply(downstream, status_code);
+
+  if(rv != 0) {
+    return -1;
+  }
+
+  return send();
+}
+
+int SpdyUpstream::consume(int32_t stream_id, size_t len)
+{
+  int rv;
+
+  rv = spdylay_session_consume(session_, stream_id, len);
+
+  if(rv != 0) {
+    ULOG(WARNING, this) << "spdylay_session_consume() returned error: "
+                        << spdylay_strerror(rv);
+    return -1;
+  }
+
+  return 0;
+}
+
+
+int SpdyUpstream::on_timeout(Downstream *downstream)
+{
+  if(LOG_ENABLED(INFO)) {
+    ULOG(INFO, this) << "Stream timeout stream_id="
+                     << downstream->get_stream_id();
+  }
+
+  rst_stream(downstream, SPDYLAY_INTERNAL_ERROR);
+
+  return 0;
+}
+
+void SpdyUpstream::reset_timeouts()
+{
+  handler_->set_upstream_timeouts(&get_config()->http2_upstream_read_timeout,
+                                  &get_config()->upstream_write_timeout);
+}
+
 } // namespace shrpx
