@@ -34,9 +34,15 @@
 
 namespace shrpx {
 
+struct WorkerStat {
+  WorkerStat() : num_connections(0) {}
+
+  size_t num_connections;
+};
+
 class Worker {
 public:
-  Worker(WorkerInfo *info);
+  Worker(const WorkerInfo *info);
   ~Worker();
   void run();
 private:
