@@ -159,8 +159,7 @@ void test_nghttp2_hd_deflate_same_indexed_repr(void)
   CU_ASSERT(0 == nghttp2_hd_deflate_init(&deflater));
   CU_ASSERT(0 == nghttp2_hd_inflate_init(&inflater));
 
-  /* Encode 2 same headers. cookie:alpha is not in the reference set,
-     so first emit literal repr and then 2 emits of indexed repr. */
+  /* Encode 2 same headers.  Emit 1 literal reprs and 1 index repr. */
   rv = nghttp2_hd_deflate_hd_bufs(&deflater, &bufs, nva1, ARRLEN(nva1));
   blocklen = nghttp2_bufs_len(&bufs);
 
@@ -174,13 +173,12 @@ void test_nghttp2_hd_deflate_same_indexed_repr(void)
   nva_out_reset(&out);
   nghttp2_bufs_reset(&bufs);
 
-  /* Encode 3 same headers. This time, cookie:alpha is in the
-     reference set, so the encoder emits indexed repr 6 times */
+  /* Encode 3 same headers.  This time, emits 3 index reprs. */
   rv = nghttp2_hd_deflate_hd_bufs(&deflater, &bufs, nva2, ARRLEN(nva2));
   blocklen = nghttp2_bufs_len(&bufs);
 
   CU_ASSERT(0 == rv);
-  CU_ASSERT(blocklen == 6);
+  CU_ASSERT(blocklen == 3);
   CU_ASSERT(blocklen == inflate_hd(&inflater, &out, &bufs, 0));
 
   CU_ASSERT(3 == out.nvlen);
@@ -195,120 +193,6 @@ void test_nghttp2_hd_deflate_same_indexed_repr(void)
   nghttp2_hd_deflate_free(&deflater);
 }
 
-void test_nghttp2_hd_deflate_common_header_eviction(void)
-{
-  nghttp2_hd_deflater deflater;
-  nghttp2_hd_inflater inflater;
-  nghttp2_nv nva[] = {MAKE_NV("h1", ""),
-                      MAKE_NV("h2", "")};
-  nghttp2_bufs bufs;
-  ssize_t blocklen;
-  /* Default header table capacity is 4096. Adding 2 byte header name
-     and 4060 byte value, which is 4094 bytes including overhead, to
-     the table evicts first entry. */
-  uint8_t value[3038];
-  nva_out out;
-  size_t i;
-  int rv;
-
-  frame_pack_bufs_init(&bufs);
-
-  nva_out_init(&out);
-  memset(value, '0', sizeof(value));
-  for(i = 0; i < 2; ++i) {
-    nva[i].value = value;
-    nva[i].valuelen = sizeof(value);
-  }
-
-  nghttp2_hd_deflate_init(&deflater);
-  nghttp2_hd_inflate_init(&inflater);
-
-  /* First emit "h1: ..." to put it in the reference set (index
-     = 0). */
-  rv = nghttp2_hd_deflate_hd_bufs(&deflater, &bufs, nva, 1);
-  blocklen = nghttp2_bufs_len(&bufs);
-
-  CU_ASSERT(0 == rv);
-  CU_ASSERT(blocklen > 0);
-  CU_ASSERT(blocklen == inflate_hd(&inflater, &out, &bufs, 0));
-
-  CU_ASSERT(1 == out.nvlen);
-  nghttp2_nv_array_sort(nva, 1);
-  assert_nv_equal(nva, out.nva, 1);
-
-  nva_out_reset(&out);
-  nghttp2_bufs_reset(&bufs);
-
-  /* Encode with second header */
-
-  rv = nghttp2_hd_deflate_hd_bufs(&deflater, &bufs, nva, 2);
-  blocklen = nghttp2_bufs_len(&bufs);
-
-  CU_ASSERT(0 == rv);
-  CU_ASSERT(blocklen > 0);
-
-  /* Check common header "h1: ...:, which is removed from the
-     header table because of eviction, is still emitted by the
-     inflater */
-  CU_ASSERT(blocklen == inflate_hd(&inflater, &out, &bufs, 0));
-
-  CU_ASSERT(2 == out.nvlen);
-  nghttp2_nv_array_sort(nva, 2);
-  assert_nv_equal(nva, out.nva, 2);
-
-  nva_out_reset(&out);
-  nghttp2_bufs_reset(&bufs);
-
-  CU_ASSERT(1 == deflater.ctx.hd_table.len);
-  CU_ASSERT(1 == inflater.ctx.hd_table.len);
-
-  nghttp2_bufs_free(&bufs);
-  nghttp2_hd_inflate_free(&inflater);
-  nghttp2_hd_deflate_free(&deflater);
-}
-
-void test_nghttp2_hd_deflate_clear_refset(void)
-{
-  nghttp2_hd_deflater deflater;
-  nghttp2_hd_inflater inflater;
-  nghttp2_bufs bufs;
-  ssize_t blocklen;
-  nghttp2_nv nv[] = {
-    MAKE_NV(":path", "/"),
-    MAKE_NV(":scheme", "http")
-  };
-  size_t i;
-  nva_out out;
-  int rv;
-
-  frame_pack_bufs_init(&bufs);
-
-  nva_out_init(&out);
-  nghttp2_hd_deflate_init2(&deflater,
-                           NGHTTP2_HD_DEFAULT_MAX_DEFLATE_BUFFER_SIZE);
-  nghttp2_hd_deflate_set_no_refset(&deflater, 1);
-  nghttp2_hd_inflate_init(&inflater);
-
-  for(i = 0; i < 2; ++i) {
-    rv = nghttp2_hd_deflate_hd_bufs(&deflater, &bufs, nv, ARRLEN(nv));
-    blocklen = nghttp2_bufs_len(&bufs);
-
-    CU_ASSERT(0 == rv);
-    CU_ASSERT(blocklen > 1);
-    CU_ASSERT(blocklen == inflate_hd(&inflater, &out, &bufs, 0));
-
-    CU_ASSERT(ARRLEN(nv) == out.nvlen);
-    assert_nv_equal(nv, out.nva, ARRLEN(nv));
-
-    nva_out_reset(&out);
-    nghttp2_bufs_reset(&bufs);
-  }
-
-  nghttp2_bufs_free(&bufs);
-  nghttp2_hd_inflate_free(&inflater);
-  nghttp2_hd_deflate_free(&deflater);
-}
-
 void test_nghttp2_hd_inflate_indexed(void)
 {
   nghttp2_hd_inflater inflater;
@@ -410,9 +294,11 @@ void test_nghttp2_hd_inflate_indname_inc(void)
   CU_ASSERT(1 == out.nvlen);
   assert_nv_equal(&nv, out.nva, 1);
   CU_ASSERT(1 == inflater.ctx.hd_table.len);
-  assert_nv_equal(&nv,
-                  &GET_TABLE_ENT(&inflater.ctx,
-                                 inflater.ctx.hd_table.len-1)->nv, 1);
+  assert_nv_equal
+    (&nv,
+     &GET_TABLE_ENT
+     (&inflater.ctx,
+      NGHTTP2_STATIC_TABLE_LENGTH + inflater.ctx.hd_table.len-1)->nv, 1);
 
   nva_out_reset(&out);
   nghttp2_bufs_free(&bufs);
@@ -459,7 +345,6 @@ void test_nghttp2_hd_inflate_indname_inc_eviction(void)
   nghttp2_bufs_reset(&bufs);
 
   CU_ASSERT(3 == inflater.ctx.hd_table.len);
-  CU_ASSERT(GET_TABLE_ENT(&inflater.ctx, 0)->flags & NGHTTP2_HD_FLAG_REFSET);
 
   nghttp2_bufs_free(&bufs);
   nghttp2_hd_inflate_free(&inflater);
@@ -530,9 +415,11 @@ void test_nghttp2_hd_inflate_newname_inc(void)
   CU_ASSERT(1 == out.nvlen);
   assert_nv_equal(&nv, out.nva, 1);
   CU_ASSERT(1 == inflater.ctx.hd_table.len);
-  assert_nv_equal(&nv,
-                  &GET_TABLE_ENT(&inflater.ctx,
-                                 inflater.ctx.hd_table.len-1)->nv, 1);
+  assert_nv_equal
+    (&nv,
+     &GET_TABLE_ENT
+     (&inflater.ctx,
+      NGHTTP2_STATIC_TABLE_LENGTH + inflater.ctx.hd_table.len-1)->nv, 1);
 
   nva_out_reset(&out);
   nghttp2_bufs_free(&bufs);
@@ -641,12 +528,64 @@ void test_nghttp2_hd_inflate_zero_length_huffman(void)
   nghttp2_hd_inflate_free(&inflater);
 }
 
+void test_nghttp2_hd_ringbuf_reserve(void)
+{
+  nghttp2_hd_deflater deflater;
+  nghttp2_hd_inflater inflater;
+  nghttp2_nv nv;
+  nghttp2_bufs bufs;
+  nva_out out;
+  int i;
+  ssize_t rv;
+  ssize_t blocklen;
+
+  frame_pack_bufs_init(&bufs);
+  nva_out_init(&out);
+
+  nv.flags = NGHTTP2_NV_FLAG_NONE;
+  nv.name = (uint8_t*)"a";
+  nv.namelen = strlen((const char*)nv.name);
+  nv.valuelen = 4;
+  nv.value = malloc(nv.valuelen);
+  memset(nv.value, 0, nv.valuelen);
+
+  nghttp2_hd_deflate_init2(&deflater, 8000);
+  nghttp2_hd_inflate_init(&inflater);
+
+  nghttp2_hd_inflate_change_table_size(&inflater, 8000);
+  nghttp2_hd_deflate_change_table_size(&deflater, 8000);
+
+  for(i = 0; i < 150; ++i) {
+    memcpy(nv.value, &i, sizeof(i));
+    rv = nghttp2_hd_deflate_hd_bufs(&deflater, &bufs, &nv, 1);
+    blocklen = nghttp2_bufs_len(&bufs);
+
+    CU_ASSERT(0 == rv);
+    CU_ASSERT(blocklen > 0);
+
+    CU_ASSERT(blocklen == inflate_hd(&inflater, &out, &bufs, 0));
+
+    CU_ASSERT(1 == out.nvlen);
+    assert_nv_equal(&nv, out.nva, 1);
+
+    nva_out_reset(&out);
+    nghttp2_bufs_reset(&bufs);
+  }
+
+  nghttp2_bufs_free(&bufs);
+  nghttp2_hd_inflate_free(&inflater);
+  nghttp2_hd_deflate_free(&deflater);
+
+  free(nv.value);
+}
+
 void test_nghttp2_hd_change_table_size(void)
 {
   nghttp2_hd_deflater deflater;
   nghttp2_hd_inflater inflater;
-  nghttp2_nv nva[] = { MAKE_NV(":method", "GET"),
-                       MAKE_NV(":path", "/") };
+  nghttp2_nv nva[] = { MAKE_NV("alpha", "bravo"),
+                       MAKE_NV("charlie", "delta") };
+  nghttp2_nv nva2[] = { MAKE_NV(":path", "/") };
   nghttp2_bufs bufs;
   ssize_t rv;
   nva_out out;
@@ -663,10 +602,8 @@ void test_nghttp2_hd_change_table_size(void)
   CU_ASSERT(0 == nghttp2_hd_inflate_change_table_size(&inflater, 8000));
   CU_ASSERT(0 == nghttp2_hd_deflate_change_table_size(&deflater, 8000));
 
-  CU_ASSERT(127 == deflater.ctx.hd_table.mask);
   CU_ASSERT(4096 == deflater.ctx.hd_table_bufsize_max);
 
-  CU_ASSERT(255 == inflater.ctx.hd_table.mask);
   CU_ASSERT(8000 == inflater.ctx.hd_table_bufsize_max);
   CU_ASSERT(8000 == inflater.settings_hd_table_bufsize_max);
 
@@ -691,10 +628,8 @@ void test_nghttp2_hd_change_table_size(void)
   CU_ASSERT(0 == nghttp2_hd_inflate_change_table_size(&inflater, 1024));
   CU_ASSERT(0 == nghttp2_hd_deflate_change_table_size(&deflater, 1024));
 
-  CU_ASSERT(127 == deflater.ctx.hd_table.mask);
   CU_ASSERT(1024 == deflater.ctx.hd_table_bufsize_max);
 
-  CU_ASSERT(255 == inflater.ctx.hd_table.mask);
   CU_ASSERT(1024 == inflater.ctx.hd_table_bufsize_max);
   CU_ASSERT(1024 == inflater.settings_hd_table_bufsize_max);
 
@@ -718,11 +653,9 @@ void test_nghttp2_hd_change_table_size(void)
   CU_ASSERT(0 == nghttp2_hd_inflate_change_table_size(&inflater, 0));
   CU_ASSERT(0 == nghttp2_hd_deflate_change_table_size(&deflater, 0));
 
-  CU_ASSERT(127 == deflater.ctx.hd_table.mask);
   CU_ASSERT(0 == deflater.ctx.hd_table.len);
   CU_ASSERT(0 == deflater.ctx.hd_table_bufsize_max);
 
-  CU_ASSERT(255 == inflater.ctx.hd_table.mask);
   CU_ASSERT(0 == inflater.ctx.hd_table.len);
   CU_ASSERT(0 == inflater.ctx.hd_table_bufsize_max);
   CU_ASSERT(0 == inflater.settings_hd_table_bufsize_max);
@@ -757,10 +690,8 @@ void test_nghttp2_hd_change_table_size(void)
   CU_ASSERT(0 == nghttp2_hd_inflate_change_table_size(&inflater, 8000));
   CU_ASSERT(0 == nghttp2_hd_deflate_change_table_size(&deflater, 8000));
 
-  CU_ASSERT(255 == deflater.ctx.hd_table.mask);
   CU_ASSERT(8000 == deflater.ctx.hd_table_bufsize_max);
 
-  CU_ASSERT(255 == inflater.ctx.hd_table.mask);
   CU_ASSERT(8000 == inflater.ctx.hd_table_bufsize_max);
   CU_ASSERT(8000 == inflater.settings_hd_table_bufsize_max);
 
@@ -783,10 +714,8 @@ void test_nghttp2_hd_change_table_size(void)
   CU_ASSERT(0 == nghttp2_hd_inflate_change_table_size(&inflater, 16383));
   CU_ASSERT(0 == nghttp2_hd_deflate_change_table_size(&deflater, 16383));
 
-  CU_ASSERT(255 == deflater.ctx.hd_table.mask);
   CU_ASSERT(8192 == deflater.ctx.hd_table_bufsize_max);
 
-  CU_ASSERT(511 == inflater.ctx.hd_table.mask);
   CU_ASSERT(16383 == inflater.ctx.hd_table_bufsize_max);
   CU_ASSERT(16383 == inflater.settings_hd_table_bufsize_max);
 
@@ -823,7 +752,6 @@ void test_nghttp2_hd_change_table_size(void)
   nghttp2_hd_deflate_init2(&deflater, 1024);
   nghttp2_hd_inflate_init(&inflater);
 
-  CU_ASSERT(127 == deflater.ctx.hd_table.mask);
   CU_ASSERT(1024 == deflater.ctx.hd_table_bufsize_max);
 
   /* This emits context update with buffer size 1024 */
@@ -843,9 +771,63 @@ void test_nghttp2_hd_change_table_size(void)
   nva_out_reset(&out);
   nghttp2_bufs_reset(&bufs);
 
-  nghttp2_bufs_free(&bufs);
   nghttp2_hd_inflate_free(&inflater);
   nghttp2_hd_deflate_free(&deflater);
+
+  /* Check that table size UINT32_MAX can be received */
+  nghttp2_hd_deflate_init2(&deflater, UINT32_MAX);
+  nghttp2_hd_inflate_init(&inflater);
+
+  CU_ASSERT(0 == nghttp2_hd_inflate_change_table_size(&inflater, UINT32_MAX));
+  CU_ASSERT(0 == nghttp2_hd_deflate_change_table_size(&deflater, UINT32_MAX));
+
+  rv = nghttp2_hd_deflate_hd_bufs(&deflater, &bufs, nva, 2);
+  blocklen = nghttp2_bufs_len(&bufs);
+
+  CU_ASSERT(0 == rv);
+  CU_ASSERT(UINT32_MAX == deflater.ctx.hd_table_bufsize_max);
+
+  CU_ASSERT(blocklen == inflate_hd(&inflater, &out, &bufs, 0));
+  CU_ASSERT(UINT32_MAX == inflater.ctx.hd_table_bufsize_max);
+  CU_ASSERT(UINT32_MAX == inflater.settings_hd_table_bufsize_max);
+
+  nva_out_reset(&out);
+  nghttp2_bufs_reset(&bufs);
+
+  nghttp2_hd_inflate_free(&inflater);
+  nghttp2_hd_deflate_free(&deflater);
+
+  /* Check that context update emitted twice */
+  nghttp2_hd_deflate_init2(&deflater, 4096);
+  nghttp2_hd_inflate_init(&inflater);
+
+  CU_ASSERT(0 == nghttp2_hd_inflate_change_table_size(&inflater, 0));
+  CU_ASSERT(0 == nghttp2_hd_inflate_change_table_size(&inflater, 3000));
+  CU_ASSERT(0 == nghttp2_hd_deflate_change_table_size(&deflater, 0));
+  CU_ASSERT(0 == nghttp2_hd_deflate_change_table_size(&deflater, 3000));
+
+  CU_ASSERT(0 == deflater.min_hd_table_bufsize_max);
+  CU_ASSERT(3000 == deflater.ctx.hd_table_bufsize_max);
+
+  rv = nghttp2_hd_deflate_hd_bufs(&deflater, &bufs, nva2, 1);
+  blocklen = nghttp2_bufs_len(&bufs);
+
+  CU_ASSERT(0 == rv);
+  CU_ASSERT(3 < blocklen);
+  CU_ASSERT(3000 == deflater.ctx.hd_table_bufsize_max);
+  CU_ASSERT(UINT32_MAX == deflater.min_hd_table_bufsize_max);
+
+  CU_ASSERT(blocklen == inflate_hd(&inflater, &out, &bufs, 0));
+  CU_ASSERT(3000 == inflater.ctx.hd_table_bufsize_max);
+  CU_ASSERT(3000 == inflater.settings_hd_table_bufsize_max);
+
+  nva_out_reset(&out);
+  nghttp2_bufs_reset(&bufs);
+
+  nghttp2_hd_inflate_free(&inflater);
+  nghttp2_hd_deflate_free(&deflater);
+
+  nghttp2_bufs_free(&bufs);
 }
 
 static void check_deflate_inflate(nghttp2_hd_deflater *deflater,
@@ -1099,7 +1081,7 @@ void test_nghttp2_hd_deflate_bound(void)
 
   bound = nghttp2_hd_deflate_bound(&deflater, nva, ARRLEN(nva));
 
-  CU_ASSERT(1 + 6 + 6 * 2 * 2 +
+  CU_ASSERT(12 + 6 * 2 * 2 +
             nva[0].namelen + nva[0].valuelen +
             nva[1].namelen + nva[1].valuelen
             == bound);
@@ -1110,7 +1092,7 @@ void test_nghttp2_hd_deflate_bound(void)
 
   bound2 = nghttp2_hd_deflate_bound(&deflater, nva, ARRLEN(nva));
 
-  CU_ASSERT(bound + 2 == bound2);
+  CU_ASSERT(bound == bound2);
 
   nghttp2_bufs_free(&bufs);
   nghttp2_hd_deflate_free(&deflater);
@@ -1159,3 +1141,79 @@ void test_nghttp2_hd_public_api(void)
   nghttp2_hd_deflate_del(deflater);
 }
 
+static size_t encode_length(uint8_t *buf, uint64_t n, size_t prefix)
+{
+  size_t k = (1 << prefix) - 1;
+  size_t len = 0;
+  *buf &= ~k;
+  if(n >= k) {
+    *buf++ |= k;
+    n -= k;
+    ++len;
+  } else {
+    *buf++ |= n;
+    return 1;
+  }
+  do {
+    ++len;
+    if(n >= 128) {
+      *buf++ = (1 << 7) | (n & 0x7f);
+      n >>= 7;
+    } else {
+      *buf++ = (uint8_t)n;
+      break;
+    }
+  } while(n);
+  return len;
+}
+
+void test_nghttp2_hd_decode_length(void)
+{
+  uint32_t out;
+  size_t shift;
+  int final;
+  uint8_t buf[16];
+  uint8_t *bufp;
+  size_t len;
+  ssize_t rv;
+  size_t i;
+
+  memset(buf, 0, sizeof(buf));
+  len = encode_length(buf, UINT32_MAX, 7);
+
+  rv = nghttp2_hd_decode_length(&out, &shift, &final, 0, 0, buf, buf + len, 7);
+
+  CU_ASSERT((ssize_t)len == rv);
+  CU_ASSERT(0 != final);
+  CU_ASSERT(UINT32_MAX == out);
+
+  /* Make sure that we can decode integer if we feed 1 byte at a
+     time */
+  out = 0;
+  shift = 0;
+  final = 0;
+  bufp = buf;
+
+  for(i = 0; i < len; ++i, ++bufp) {
+    rv = nghttp2_hd_decode_length(&out, &shift, &final, out, shift,
+                                  bufp, bufp + 1, 7);
+
+    CU_ASSERT(rv == 1);
+
+    if(final) {
+      break;
+    }
+  }
+
+  CU_ASSERT(i == len - 1);
+  CU_ASSERT(0 != final);
+  CU_ASSERT(UINT32_MAX == out);
+
+  /* Check overflow case */
+  memset(buf, 0, sizeof(buf));
+  len = encode_length(buf, 1ll << 32, 7);
+
+  rv = nghttp2_hd_decode_length(&out, &shift, &final, 0, 0, buf, buf + len, 7);
+
+  CU_ASSERT(-1 == rv);
+}
