@@ -34,6 +34,10 @@ void test_nghttp2_session_recv_continuation(void);
 void test_nghttp2_session_recv_headers_with_priority(void);
 void test_nghttp2_session_recv_premature_headers(void);
 void test_nghttp2_session_recv_altsvc(void);
+void test_nghttp2_session_recv_unknown_frame(void);
+void test_nghttp2_session_recv_unexpected_continuation(void);
+void test_nghttp2_session_recv_settings_header_table_size(void);
+void test_nghttp2_session_recv_too_large_frame_length(void);
 void test_nghttp2_session_continue(void);
 void test_nghttp2_session_add_frame(void);
 void test_nghttp2_session_on_request_headers_received(void);
@@ -50,7 +54,7 @@ void test_nghttp2_session_on_window_update_received(void);
 void test_nghttp2_session_on_data_received(void);
 void test_nghttp2_session_send_headers_start_stream(void);
 void test_nghttp2_session_send_headers_reply(void);
-void test_nghttp2_session_send_headers_header_comp_error(void);
+void test_nghttp2_session_send_headers_frame_size_error(void);
 void test_nghttp2_session_send_headers_push_reply(void);
 void test_nghttp2_session_send_rst_stream(void);
 void test_nghttp2_session_send_push_promise(void);
@@ -58,6 +62,9 @@ void test_nghttp2_session_is_my_stream_id(void);
 void test_nghttp2_session_upgrade(void);
 void test_nghttp2_session_reprioritize_stream(void);
 void test_nghttp2_submit_data(void);
+void test_nghttp2_submit_data_read_length_too_large(void);
+void test_nghttp2_submit_data_read_length_smallest(void);
+void test_nghttp2_submit_data_twice(void);
 void test_nghttp2_submit_request_with_data(void);
 void test_nghttp2_submit_request_without_data(void);
 void test_nghttp2_submit_response_with_data(void);
@@ -95,9 +102,6 @@ void test_nghttp2_session_set_option(void);
 void test_nghttp2_session_data_backoff_by_high_pri_frame(void);
 void test_nghttp2_session_pack_data_with_padding(void);
 void test_nghttp2_session_pack_headers_with_padding(void);
-void test_nghttp2_session_pack_headers_with_padding2(void);
-void test_nghttp2_session_pack_headers_with_padding3(void);
-void test_nghttp2_session_pack_headers_with_padding4(void);
 void test_nghttp2_pack_settings_payload(void);
 void test_nghttp2_session_stream_dep_add(void);
 void test_nghttp2_session_stream_dep_remove(void);
@@ -107,5 +111,9 @@ void test_nghttp2_session_stream_dep_all_your_stream_are_belong_to_us(void);
 void test_nghttp2_session_stream_attach_data(void);
 void test_nghttp2_session_stream_attach_data_subtree(void);
 void test_nghttp2_session_keep_closed_stream(void);
+void test_nghttp2_session_graceful_shutdown(void);
+void test_nghttp2_session_on_header_temporal_failure(void);
+void test_nghttp2_session_recv_client_preface(void);
+void test_nghttp2_session_delete_data_item(void);
 
 #endif /* NGHTTP2_SESSION_TEST_H */
